{-# LANGUAGE
ViewPatterns,
ParallelListComp,
DuplicateRecordFields,
DisambiguateRecordFields,
RecordWildCards,
NamedFieldPuns,
TemplateHaskell,
DeriveDataTypeable,
DeriveGeneric,
ScopedTypeVariables,
TypeFamilies,
RankNTypes,
MultiParamTypeClasses,
FlexibleInstances,
FlexibleContexts
#-}



{- |

Module: BaselMachine.Arena

Description : Data structures representing a collection of interacting colonies of cells, information about those cells, etc.

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)


-}

module BaselMachine.Arena
  (
    NullLocation(..),
    NullSignal(..),
    Source(..),
    Target(..),
    ChannelSpec(..),
    ChannelState(..),
    CellState(..),
    ColonySpec(..),
    ColonyInitialState(..),
    ColonyState(..),
    ArenaState(..),
    BroadcastSpec(..),
    BroadcastState(..),
    SensorSpec(..),
    CellActivationRangeTable(..),
    ColonyActivationRangeTable,
    ArenaActivationRangeTable,
    ResourceSummary(..),
    ColonyResourceSummary(..),
    ArenaRunSummary(..),
    ColonyTraceSummary(..),
    ArenaTraceSummary(..),
    ArenaTraceStep,
    TracePoint(..),
    BadColonySpecException(..),
    runArena,
    runArenaTrace
  ) where

import GHC.Generics

import Control.DeepSeq
import Control.Exception
import Control.Monad
import Control.Monad.ST

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
-- import Data.Foldable
import qualified Data.Graph.Inductive.Graph as G
import Data.Graph.Inductive.Graph (Node, LNode, LEdge)
import Data.Graph.Inductive.PatriciaTree
import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
import Data.IntSet (IntSet)
import qualified Data.IntSet as IntSet
import Data.List
import Data.Maybe
import Data.Map (Map)
import qualified Data.Map as Map
-- import Data.Semigroup
import Data.Sequence (Seq, (|>), ViewR(..), viewr)
import qualified Data.Sequence as Seq
import Data.STRef
import Data.Typeable()
import qualified Data.Vector.Generic as V
import qualified Data.Vector.Generic.Mutable as MV
import qualified Data.Vector.Unboxed as UV

-- import Debug.Trace
-- import Text.Printf

import BaselMachine.AesonUtil
import BaselMachine.CircularBuffer
import BaselMachine.Compiler
import BaselMachine.MVM
import BaselMachine.Signal
import WGM.OrphanInstances ( )
import WGM.Snapshot

-- === General implementation notes ===
--
-- I've tried to make state transitions run with as little
-- allocation as possible.  That means there's a lot of places
-- where I put changed objects into an STRef or something similar
-- rather than allocate a new record.
--
-- The reason for including the Arena in the Basel Machine
-- package is that for speed and memory efficiency, I need Arena
-- code to have access to the mutable virtual machines in each
-- cell, but I don't want anything outside of the Basel Machine
-- package to have that access.

data NullLocation = NullLocation
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData NullLocation
$(deriveJSON aesonOptions ''NullLocation)


data Source =
  Source
  {
    sourceCellIx :: !Int,
    sourcePattern :: !Int,
    sourceThreshold :: !Int,
    sourceActivationDelta :: !Int
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData Source
instance Binary Source
$(deriveJSON aesonOptions ''Source)

instance Monad m => Snapshot m Source



data Target =
  Target
  {
    targetCellIx :: !Int,
    targetPattern :: !Int,
    targetActivationDelta :: !Int
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance Monad m => Snapshot m Target

instance NFData Target
instance Binary Target
$(deriveJSON aesonOptions ''Target)


data ChannelSpec =
  ChannelSpec
  {
    channelSource :: !Source,
    channelTarget :: !Target,
    channelDelay :: !Int
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance Monad m => Snapshot m ChannelSpec

instance NFData ChannelSpec
instance Binary ChannelSpec
$(deriveJSON aesonOptions ''ChannelSpec)


data MChannelState s =
  MChannelState
  {
    channelSpec :: ChannelSpec,
    channelSpikeTrainState :: CircularBuffer s
  }

data ChannelState =
  ChannelState
  {
    channelSpec :: ChannelSpec,
    channelSpikeTrainState :: Seq Bool
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData ChannelState
instance Binary ChannelState
$(deriveJSON aesonOptions ''ChannelState)



instance Snapshot (ST s) (MChannelState s) where
  type Memento (MChannelState s) = ChannelState

  snapshot MChannelState{..} = do
    spikeTrainMemento <- snapshot channelSpikeTrainState
    return ChannelState {
      channelSpikeTrainState = spikeTrainMemento,
      ..}



data MCellState s =
  MCellState
  {
    cellVM :: MVM s,
    cellTotalReactionCount :: STRef s Int,
    cellTotalReactionDeltas :: STRef s Int
  }

data CellState =
  CellState
  {
    cellVM :: UV.Vector Int,
    cellTotalReactionCount :: Int,
    cellTotalReactionDeltas :: Int
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData CellState
instance Binary CellState
$(deriveJSON aesonOptions ''CellState)

instance Snapshot (ST s) (MCellState s) where

  type Memento (MCellState s) = CellState

  snapshot MCellState{..} = do
    sVM <- snapshot cellVM
    trc <- snapshot cellTotalReactionCount
    trd <- snapshot cellTotalReactionDeltas
    return CellState {
      cellVM = sVM,
      cellTotalReactionCount = trc,
      cellTotalReactionDeltas = trd }


-- | The signal should be something immutable that can serve as its own 'Memento'
data BroadcastSpec signal =
  BroadcastSpec
  {
    broadcastSource :: !Source,
    broadcastMaximumAge :: Maybe Int,
    broadcastSignal :: !signal
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData signal => NFData (BroadcastSpec signal)
instance Binary signal => Binary (BroadcastSpec signal)
$(deriveJSON aesonOptions ''BroadcastSpec)

instance (Monad m, Snapshot m signal) =>
   Snapshot m (BroadcastSpec signal)



data MBroadcastState location signal s =
  MBroadcastState
  {
    broadcastMaximumAge :: Maybe Int,
    broadcastSignal :: !signal,
    broadcastFromColonyIx :: Maybe Int,
    broadcastOrigin :: location,
    broadcastAge :: STRef s Int
  }

data BroadcastState location signal =
  BroadcastState
  {
    broadcastMaximumAge :: Maybe Int,
    broadcastSignal :: signal,
    broadcastFromColonyIx :: Maybe Int,
    broadcastOrigin :: location,
    broadcastAge :: Int
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance (NFData location, NFData signal) => NFData (BroadcastState location signal)
instance (Binary location, Binary signal) => Binary (BroadcastState location signal)
$(deriveJSON aesonOptions ''BroadcastState)

instance Snapshot (ST s) (MBroadcastState location signal s) where

  type Memento (MBroadcastState location signal s) =
    BroadcastState location signal

  snapshot MBroadcastState{..} = do
    age <- snapshot broadcastAge
    return BroadcastState{
      broadcastAge = age,
      ..}

instance Snapshot m (BroadcastState location signal)

instance Reconstruct (ST s) (MBroadcastState location signal s) where
  reconstruct BroadcastState{..} = do
    ageRef <- reconstruct broadcastAge
    return MBroadcastState{
      broadcastAge = ageRef,
      ..}



data SensorSpec signal =
  SensorSpec
  {
    sensorSignal :: !signal,
    sensorTarget :: !Target
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData signal => NFData (SensorSpec signal)
instance Binary signal => Binary (SensorSpec signal)
$(deriveJSON aesonOptions ''SensorSpec)



data ColonySpec signal =
  ColonySpec
  {
    colonyCellCount :: Int,
    colonyCellIxToName :: IntMap String,
    colonySynapseSpecs :: [ChannelSpec],
    colonyBroadcastSpecs :: [BroadcastSpec signal],
    colonySensorSpecs :: [SensorSpec signal]
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData signal => NFData (ColonySpec signal)
instance Binary signal => Binary (ColonySpec signal)
$(deriveJSON aesonOptions ''ColonySpec)

-- Complication is that there's no nice way to put ToJSONKey
-- signal into the context for deriveJSON.  So, we can't use
-- Map signal Target within ColonySpec.
-- Hence we build the map and cache it within MColonyState instead


data MColonyState location signal s =
  MColonyState
  {
    colonySpec :: ColonySpec signal,
    colonyName :: String,
    colonyLocation :: STRef s location,
    colonyCellGraph :: STRef s (Gr (MCellState s) (MChannelState s)),
    colonySensors :: Map signal Target
  }


data ColonyState location signal =
  ColonyState
  {
    colonySpec :: ColonySpec signal,
    colonyName :: String,
    colonyLocation :: location,
    colonyCellGraph :: ([LNode CellState], [LEdge ChannelState])
  }
  deriving (Eq, Ord, Read, Show, Generic, Typeable)

-- Binary and other instances for Gr are missing, which is why
-- the cell graph snapshot is in list form.

instance (NFData location, NFData signal) => NFData (ColonyState location signal)
instance (Binary location, Binary signal) => Binary (ColonyState location signal)
$(deriveJSON aesonOptions ''ColonyState)




instance
  Snapshot (ST s) (MColonyState location signal s) where

  type Memento (MColonyState location signal s) =
    ColonyState location signal

  snapshot MColonyState{..} = do
    location <- readSTRef colonyLocation
    cellGraphM <- snapshot colonyCellGraph
    return ColonyState {
      colonyLocation = location,
      colonyCellGraph = cellGraphM,
      ..}



data ColonyInitialState location signal =
  ColonyInitialState {
  initName :: String,
  initColonySpec :: ColonySpec signal,
  initProgram :: (ProgramInformation, ReactionProgram),
  initLocation :: location
  }
  deriving (Generic, Typeable)

-- These don't work because of ReactionProgram
-- instance NFData ColonyInitialState
-- instance Binary ColonyInitialState
-- $(deriveJSON aesonOptions ''ColonyInitialState)



data MArenaState location signal s =
  MArenaState
  {
    arenaColonyIxToState ::
      STRef s (IntMap (MColonyState location signal s)),
    arenaBroadcastStates ::
      STRef s [MBroadcastState location signal s]
  }

data ArenaState location signal =
  ArenaState
  {
    arenaColonyIxToState ::
      IntMap (ColonyState location signal),
    arenaBroadcastStates ::
      [BroadcastState location signal]
  }
  deriving (Eq, Read, Show, Generic, Typeable)

instance (NFData location, NFData signal) => NFData (ArenaState location signal)
instance (Binary location, Binary signal) => Binary (ArenaState location signal)
$(deriveJSON aesonOptions ''ArenaState)



instance
  Snapshot (ST s) (MArenaState location signal s) where

  type Memento (MArenaState location signal s) =
    ArenaState location signal

  snapshot MArenaState{..} = do
    colonyIxToState <- snapshot arenaColonyIxToState
    broadcastStates <- snapshot arenaBroadcastStates
    return ArenaState {
      arenaColonyIxToState = colonyIxToState,
      arenaBroadcastStates = broadcastStates,
      ..}



-- === Utilities ===


mkCheckedLookup ::
  (key -> map -> Bool)
  -> (key -> map -> Maybe value)
  -> key -> map -> value

mkCheckedLookup ck lk k m =
  assert (ck k m) $ fromJust (lk k m)

imGet = mkCheckedLookup IntMap.member IntMap.lookup
imGet :: IntSet.Key -> IntMap value -> value

{-# INLINE imGet #-}



nodeLabel ::
  (G.Graph gr)
  =>
  Node -> gr nLabel eLabel -> nLabel
nodeLabel = mkCheckedLookup G.gelem lk
  where
    lk g n = G.lab n g

{-# INLINE nodeLabel #-}


-- === Phases related to broadcasts ===

-- | Add external broadcasts to the arena
mArenaIncorporateInputs ::
  [BroadcastState location signal]
  -> MArenaState location signal s
  -> ST s ()

mArenaIncorporateInputs [] _ =
  return ()

mArenaIncorporateInputs inputMs MArenaState{..} = do
  externalBroadcasts <- reconstruct inputMs
  modifySTRef' arenaBroadcastStates (externalBroadcasts ++)

{-# INLINE mArenaIncorporateInputs #-}


-- | Send broadcasts to agents
mArenaPhaseBroadcastSend ::
  forall location signal s.
  (BroadcastSignal location signal,
   Ord signal,
   -- for trace
   Show signal
  )
  =>
  MArenaState location signal s -> ST s ()

mArenaPhaseBroadcastSend MArenaState{..} = do
  broadcastStates <- readSTRef arenaBroadcastStates
  colonyIxToState <- readSTRef arenaColonyIxToState
  -- traceM $ printf "mArenaPhaseBroadcastSend: sending %d broadcasts" (length broadcastStates)
  forM_ broadcastStates $ \ MBroadcastState{..} -> do
    -- traceM $ printf "Broadcast %s" (show broadcastSignal)
    forM_ (IntMap.toList colonyIxToState) $ \ (_colIx, MColonyState{..}) -> do
      let ColonySpec{..} = colonySpec
      case Map.lookup broadcastSignal colonySensors of
        Nothing -> do
          -- traceM $ printf "No matching sensor in colony %d" colIx
          return ()
        Just Target{..} -> do
          -- traceM $ printf "Matching sensor in colony %d: %s" colIx (show t)
          loc <- readSTRef colonyLocation
          cellGraph <- readSTRef colonyCellGraph
          age <- readSTRef broadcastAge
          let MCellState{..} = nodeLabel targetCellIx cellGraph
              str = (strengthAt
                     broadcastSignal
                     broadcastOrigin
                     age
                     loc)
              delta = floor (str * fromIntegral targetActivationDelta)
          -- traceM $ printf "Colony %d: A[%d] += %d" colIx targetPattern delta
          mvmApplyDelta cellVM targetPattern delta

{-# INLINE mArenaPhaseBroadcastSend #-}


-- | Add 1 to the age of all broadcasts and get rid of any that
-- have passed their maximum age.
mArenaPhaseBroadcastIncrementAge ::
  (Show signal) =>
  MArenaState location signal s -> ST s ()

mArenaPhaseBroadcastIncrementAge MArenaState{..} = do
  broadcastStates <- readSTRef arenaBroadcastStates
  forM_ broadcastStates $ \ MBroadcastState{..} -> do
    modifySTRef' broadcastAge (1 +)
  survivingBroadcasts <- filterM notTooOld broadcastStates
  -- traceM ("mArenaPhaseBroadcastIncrementAge: surviving broadcasts: " ++ show [broadcastSignal | MBroadcastState{..} <- survivingBroadcasts])
  writeSTRef arenaBroadcastStates survivingBroadcasts
  where
    -- Return whether a broadcast can be aged another step before
    -- expiring
    notTooOld MBroadcastState{..} =
      case broadcastMaximumAge of
        Nothing ->
          return True
        Just m ->  do
          age <- readSTRef broadcastAge
          return (age < m)

{-# INLINE mArenaPhaseBroadcastIncrementAge #-}


-- | Check all agents to see whether they are emitting a
-- broadcast.
mArenaPhaseBroadcastEmit ::
  forall location signal s.
  (BroadcastSignal location signal,
  Show signal)
  =>
  MArenaState location signal s -> ST s ()

mArenaPhaseBroadcastEmit MArenaState{..} = do
  colonyIxToState <- readSTRef arenaColonyIxToState
  forM_ (IntMap.toList colonyIxToState) $ \ (colonyIx, colonyState) -> do
    -- traceM $ printf "mArenaPhaseBroadcastEmit: checking colony %d" colonyIx
    let MColonyState{..} = colonyState
        ColonySpec{..} = colonySpec
        broadcastFromColonyIx = Just colonyIx
    cellGraph <- readSTRef colonyCellGraph
    newEmissions <- forM colonyBroadcastSpecs $ \ broadcastSpec -> do
      -- traceM ("mArenaPhaseBroadcastEmit: checking " ++ show broadcastSpec)
      let BroadcastSpec{..} = broadcastSpec
          Source{..} = broadcastSource
          MCellState{..} = nodeLabel sourceCellIx cellGraph
      a <- mvmReadActivation cellVM sourcePattern
      if a >= sourceThreshold
        then do
        -- traceM ("mArenaPhaseBroadcastEmit: emitting broadcast " ++ show broadcastSignal)
        mvmApplyDelta cellVM sourcePattern sourceActivationDelta
        broadcastOrigin <- readSTRef colonyLocation
        broadcastAge <- newSTRef 0
        return (Just MBroadcastState{..})
        else do
        -- traceM "mArenaPhaseBroadcastEmit: no broadcast"
        return Nothing
    oldEmissions <- readSTRef arenaBroadcastStates
    let allEmissions = catMaybes newEmissions ++ oldEmissions
    writeSTRef arenaBroadcastStates allEmissions

{-# INLINE mArenaPhaseBroadcastEmit #-}


-- === Synapse phase ===

-- | Check all cells to see whether they send a spike down a
-- synapse.
mColonyRunSynapseEmitters ::
  MColonyState location signal s
  -> ST s ()

mColonyRunSynapseEmitters MColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  -- run each pre-synaptic emitter
  forM_ (G.labEdges cellGraph) $ \ (_senderIx, _receiverIx, channelState) -> do
    -- assert senderIx == sourceCellIx
    let MChannelState{..} = channelState
        ChannelSpec{..} = channelSpec
        Source{..} = channelSource
        MCellState{..} = nodeLabel sourceCellIx cellGraph
    a <- mvmReadActivation cellVM sourcePattern
    let spike = a >= sourceThreshold
    when spike $ do
      mvmApplyDelta cellVM sourcePattern sourceActivationDelta
    pushBackBool channelSpikeTrainState spike

{-# INLINE mColonyRunSynapseEmitters #-}


-- | Checke all synapsese to see whether a spike arrived at the
-- end and must be received by a cell.
mColonyRunSynapseReceptors ::
  MColonyState location signal s
  -> ST s ()

mColonyRunSynapseReceptors MColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  -- run each post-synaptic receptor
  forM_ (G.labEdges cellGraph) $ \ (_senderIx, _receiverIx, channelState) -> do
    -- assert receiverIx == targetCellIx
    let MChannelState{..} = channelState
        ChannelSpec{..} = channelSpec
        Target{..} = channelTarget
        MCellState{..} = nodeLabel targetCellIx cellGraph
    justArrived <- popFrontBool channelSpikeTrainState
    when justArrived $ do
      mvmApplyDelta cellVM targetPattern targetActivationDelta

{-# INLINE mColonyRunSynapseReceptors #-}


-- | Run both the emission and reception parts of synapses.
mArenaPhaseSynapses ::
  MArenaState location signal s -> ST s ()

mArenaPhaseSynapses MArenaState{..} = do
  colonyIxToState <- readSTRef arenaColonyIxToState
  forM_ (IntMap.toList colonyIxToState) $ \ (_colonyIx, colonyState) -> do
    mColonyRunSynapseEmitters colonyState
    mColonyRunSynapseReceptors colonyState

{-# INLINE mArenaPhaseSynapses #-}


-- === Computation phase ===

-- | Run all reactions in a 'Colony'.  Accumulate the total count
-- of reactions that were run and the total absolute value of all
-- deltas applied.
mColonyRunReactions ::
  MColonyState location signal s
  -> ST s ()

mColonyRunReactions MColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  forM_ (G.labNodes cellGraph) $ \ (_node, cellState) -> do
    let MCellState{..} = cellState
    (reactionCount, reactionDeltas) <- mvmRunReactions cellVM
    -- traceM $ printf "mColonyRunReactions: Running cell %d" node
    -- traceM $ printf "Add reaction count %d" reactionCount
    -- traceM $ printf "Add reaction deltas %d" reactionDeltas
    modifySTRef' cellTotalReactionCount (reactionCount +)
    modifySTRef' cellTotalReactionDeltas (reactionDeltas +)

{-# INLINE mColonyRunReactions #-}


-- | Run all reactions in every 'Colony'.
mArenaPhaseReactions ::
  MArenaState location signal s -> ST s ()

mArenaPhaseReactions MArenaState{..} = do
  colonyIxToState <- readSTRef arenaColonyIxToState
  forM_ (IntMap.toList colonyIxToState) $ \ (_colonyIx, colonyState) -> do
    -- traceM $ printf "mArenaPhaseReactions: running colony %d" colonyIx
    mColonyRunReactions colonyState

{-# INLINE mArenaPhaseReactions #-}

-- === Monitor functions ===

type ArenaPhaseMonitor location signal s =
  MArenaState location signal s
  -> ST s ()

class ArenaStepMonitor monitor where
  monitorInit ::
    monitor location signal s -> ArenaPhaseMonitor location signal s
  monitorInit _ _ = return ()

  monitorAfterBroadcastSend ::
    monitor location signal s -> ArenaPhaseMonitor location signal s
  monitorAfterBroadcastSend _ _ = return ()

  monitorAfterBroadcastIncrementAge ::
    monitor location signal s -> ArenaPhaseMonitor location signal s
  monitorAfterBroadcastIncrementAge _ _ = return ()

  monitorAfterSynapses ::
    monitor location signal s -> ArenaPhaseMonitor location signal s
  monitorAfterSynapses _ _ = return ()

  monitorAfterReactions ::
    monitor location signal s -> ArenaPhaseMonitor location signal s
  monitorAfterReactions _ _ = return ()

  monitorAfterBroadcastEmit ::
    monitor location signal s -> ArenaPhaseMonitor location signal s
  monitorAfterBroadcastEmit _ _ = return ()


{-
-- | A step monitor that does nothing.
data NullStepMonitor location signal s = NullStepMonitor

-- | This instance does nothing.
instance ArenaStepMonitor NullStepMonitor
-}

logSnapshot ::
  (Snapshot (ST s) t)
  =>
  STRef s (Seq (Memento t))
  -> t
  -> ST s ()

logSnapshot archive x = logSnapshotSx archive id x


logSnapshotSx ::
  (Snapshot (ST s) t)
  =>
  STRef s (Seq (Memento t))
  -> (u -> t)
  -> u
  -> ST s ()

logSnapshotSx archive view x = do
  memento <- snapshot (view x)
  modifySTRef' archive (|> memento)



newtype BroadcastLogMonitor location signal s =
  BroadcastLogMonitor {
  broadcastArchive :: STRef s (Seq [BroadcastState location signal])
  }

instance ArenaStepMonitor BroadcastLogMonitor where

  monitorAfterBroadcastEmit BroadcastLogMonitor{..} MArenaState{..} =
    logSnapshot broadcastArchive arenaBroadcastStates



-- === Construction ===

data BadColonySpecException signal =
  BadColonySpec {
    badCellNames :: [(Int, String)],
    badBroadcastSpecs :: [BroadcastSpec signal],
    badSensorSpecs :: [SensorSpec signal],
    badSynapseSpecs :: [ChannelSpec]
    } deriving (Show, Eq, Typeable, Data, Generic)

instance (Typeable signal, Show signal, Eq signal) =>
  Exception (BadColonySpecException signal)

checkColonySpec :: (Monad m, Typeable signal, Show signal, Eq signal) => ColonySpec signal -> m ()
checkColonySpec ColonySpec{..} =
  if null badCellNames
     && null badSynapseSpecs
     && null badBroadcastSpecs
  then
    return ()
  else
    throw BadColonySpec{..}
  where
    cellIxBad j =
      j < 0 || j >= colonyCellCount

    cellNameBad (j, _) =
      cellIxBad j

    badCellNames =
      filter cellNameBad (IntMap.toList colonyCellIxToName)

    synapseBad ChannelSpec{..} =
      let
        Source{..} = channelSource
        Target{..} = channelTarget
      in
        cellIxBad sourceCellIx || cellIxBad targetCellIx

    badSynapseSpecs =
      filter synapseBad colonySynapseSpecs

    broadcastBad BroadcastSpec{..} =
      let
        Source{..} = broadcastSource
      in
        cellIxBad sourceCellIx

    badBroadcastSpecs =
      filter broadcastBad colonyBroadcastSpecs

    sensorBad SensorSpec{..} =
      let
        Target{..} = sensorTarget
      in
        cellIxBad targetCellIx

    badSensorSpecs =
      filter sensorBad colonySensorSpecs



-- | Return the highest pattern mentioned by any instruction from
-- the program that was compiled, or any of the synapses, inputs,
-- or emitters.

maxPattern :: ColonySpec signal -> ProgramInformation -> Int

maxPattern ColonySpec{..} ProgramInformation{..} =
  overallMax
  where
    maxRefReactions = patternsReferenced

    maxRefSyn = maximum (0:[
      max targetPattern sourcePattern
      | ChannelSpec{..} <- colonySynapseSpecs,
        let Source{..} = channelSource
            Target{..} = channelTarget])

    maxRefIn =
      (maximum
       (concat [
           [0],
           [sourcePattern
           | BroadcastSpec{..} <- colonyBroadcastSpecs,
             let Source{..} = broadcastSource],
           [targetPattern
           | SensorSpec{..} <- colonySensorSpecs,
             let Target{..} = sensorTarget] ]))

    overallMax = maximum [
      IntSet.findMax maxRefReactions,
      maxRefSyn,
      maxRefIn
      ]



-- | Make a mutable colony in an initial state

makeMutableColony ::
  (Eq signal,
   Ord signal,
   Show signal,
   Typeable signal)
  =>
  String  -- ^ name
  -> location -- ^ location
  -> ColonySpec signal -- ^ specification
  -> (ProgramInformation, ReactionProgram) -- ^ genetic program for the cells
  -> ST s (MColonyState location signal s)

makeMutableColony colonyName location colonySpec@ColonySpec{..} (progInfo, reactions) = do
  -- Throw an exception if something is off
  checkColonySpec colonySpec
  -- construct vitrual machines
  let vmStateSize = 1 + maxPattern colonySpec progInfo
  annotatedNodes <- forM [0 .. colonyCellCount - 1] $ \ j -> do
    cellVM <- makeZeroMVM vmStateSize reactions
    cellTotalReactionCount <- newSTRef 0
    cellTotalReactionDeltas <- newSTRef 0
    return (j, MCellState{..})
  annotatedEdges <- forM colonySynapseSpecs $ \ channelSpec -> do
    let ChannelSpec{..} = channelSpec
        Source{..} = channelSource
        Target{..} = channelTarget
    channelSpikeTrainState <- synapseWire channelDelay
    return (sourceCellIx, targetCellIx, MChannelState{..})
  let cellGraph = G.mkGraph annotatedNodes annotatedEdges
  colonyCellGraph <- newSTRef cellGraph
  colonyLocation <- newSTRef location
  let colonySensors = Map.fromList [(sensorSignal, sensorTarget) | SensorSpec{..} <- colonySensorSpecs]
  return MColonyState{..}



-- | Make a wire with @n@ zeros in it, and enough room for at
-- least @n+1@ items.

synapseWire :: Int -> ST s (CircularBuffer s)

synapseWire numSteps =
  let
    -- This is kind of overkill, but there's no point in skimping
    -- here.  There has to be room for at least two more bytes,
    -- so we can push a new byte in one end and pop one out the
    -- other, and at all times have room in the buffer so that
    -- the two pointers don't hit each other: numSteps + 1 here
    -- would theoretically be enough, except that it allows the
    -- two pointers to hit each other after the push and before
    -- the pop, and the resulting memento would indicate that the
    -- buffer is empty.
    q = quot (numSteps + 2) 16
    bSize = 16 * (q + 1)
  in do
     b <- makeCircularBuffer bSize
     replicateM_ numSteps (pushFront b 0)
     return b



-- | Make a mutable arena

makeMutableArena ::
  (Eq signal,
   Show signal,
   Ord signal,
   Typeable signal)
  =>
  [ColonyInitialState location signal]
  -> [BroadcastState location signal]
  -> ST s (MArenaState location signal s)

makeMutableArena colonyInits broadcastInits = do
  mColonyStates <- forM colonyInits $ \ ColonyInitialState{..} -> do
    makeMutableColony initName initLocation initColonySpec initProgram
  let colonyIxToState = IntMap.fromList (zip [0..] mColonyStates)
  arenaColonyIxToState <- newSTRef colonyIxToState
  mBroadcastInits <- mapM reconstruct broadcastInits
  arenaBroadcastStates <- newSTRef mBroadcastInits
  return MArenaState{..}




-- === Activation range tables ===

data MCellActivationRangeTable s =
  MCellActivationRangeTable {
  activationMin :: UV.MVector s Int,
  activationMax :: UV.MVector s Int
  }


data CellActivationRangeTable =
  CellActivationRangeTable {
  activationMin :: UV.Vector Int,
  activationMax :: UV.Vector Int
  }

instance
  Snapshot (ST s) (MCellActivationRangeTable s) where
  type Memento (MCellActivationRangeTable s) = CellActivationRangeTable

  snapshot MCellActivationRangeTable{..} = do
    frozenMin <- V.freeze activationMin
    frozenMax <- V.freeze activationMax
    return CellActivationRangeTable {
      activationMin = frozenMin,
      activationMax = frozenMax }


makeMCellActivationRangeTable ::
  Int -- ^ number of entries
  -> ST s (MCellActivationRangeTable s)

makeMCellActivationRangeTable numEntries = do
  activationMin <- MV.replicate numEntries 0
  activationMax <- MV.replicate numEntries 0
  return MCellActivationRangeTable{..}

-- | This has to include an 'STRef' because the number of cells
-- in a colony will eventually be allowed to change.

type MColonyActivationRangeTable s =
  STRef s (IntMap (MCellActivationRangeTable s))

type ColonyActivationRangeTable =
  IntMap CellActivationRangeTable


makeMColonyActivationRangeTable ::
  ColonyInitialState location signal
  -> ST s (MColonyActivationRangeTable s)

makeMColonyActivationRangeTable ColonyInitialState{..}  = do
  let ColonySpec{..} = initColonySpec
      (progInfo, _) = initProgram
      numEntries :: Int
      numEntries = 1 + maxPattern initColonySpec progInfo
  mCellARTS <- replicateM colonyCellCount (makeMCellActivationRangeTable numEntries)
  newSTRef (IntMap.fromList (zip [0..] mCellARTS))


-- | This has to include an 'STRef' because the number of
-- colonies in the arena will eventually be allowed to change.

type MArenaActivationRangeTable s =
  STRef s (IntMap (MColonyActivationRangeTable s))

type ArenaActivationRangeTable =
  IntMap ColonyActivationRangeTable


makeMArenaActivationRangeTable ::
  [ColonyInitialState location signal]
  -> ST s (MArenaActivationRangeTable s)

makeMArenaActivationRangeTable colonyInits = do
  colonyARTs <- mapM makeMColonyActivationRangeTable colonyInits
  newSTRef (IntMap.fromList (zip [0..] colonyARTs))



updateCellActivationRangeTable ::
  MCellActivationRangeTable s
  -> MCellState s
  -> ST s ()

updateCellActivationRangeTable MCellActivationRangeTable{..} MCellState{..} = do
  forM_ [0.. mvmNumPatterns cellVM - 1] $ \j -> do
    a <- mvmReadActivation cellVM j
    MV.modify activationMin (min a) j
    MV.modify activationMax (max a) j



updateColonyActivationRangeTables ::
  MColonyActivationRangeTable s
  -> MColonyState location signal s
  -> ST s ()

updateColonyActivationRangeTables colonyARTRef MColonyState{..} = do
  colonyART <- readSTRef colonyARTRef
  cellGraph <- readSTRef colonyCellGraph
  forM_ (G.labNodes cellGraph) $ \ (j, cellState) -> do
    let cellART = imGet j colonyART
    updateCellActivationRangeTable cellART cellState



updateArenaActivationRangeTables ::
  MArenaActivationRangeTable s
  -> MArenaState location signal s
  -> ST s ()

updateArenaActivationRangeTables arenaARTRef MArenaState{..} = do
  arenaART <- readSTRef arenaARTRef
  colonyIxToState <- readSTRef arenaColonyIxToState
  forM_ (IntMap.toList colonyIxToState) $ \(j, colonyState) -> do
    let colonyART = imGet j arenaART
    updateColonyActivationRangeTables colonyART colonyState


mergeCellActivationRangeTables ::
   CellActivationRangeTable
   -> CellActivationRangeTable
   -> CellActivationRangeTable

mergeCellActivationRangeTables aART bART = do
  CellActivationRangeTable{
    activationMin = V.zipWith min aMin bMin,
    activationMax = V.zipWith max aMax bMax
    }
  where
    CellActivationRangeTable {
      activationMin = aMin,
      activationMax = aMax } = aART
    CellActivationRangeTable {
      activationMin = bMin,
      activationMax = bMax } = bART



mergeColonyActivationRangeTable ::
  ColonyActivationRangeTable
  -> CellActivationRangeTable

mergeColonyActivationRangeTable cellIxToART =
  foldl1' mergeCellActivationRangeTables (IntMap.elems cellIxToART)

-- === Tracing ===

-- | A 'TracePoint' is a time during a time step when only part
-- of the computation of an 'Arena' has been completed.
data TracePoint =
  Initial
  | AfterBroadcasts
  | AfterSynapses
  | AfterReactions
  | AfterEmitters
    deriving (Read, Show, Eq, Ord, Enum, Data, Typeable, Generic)

instance NFData TracePoint
instance Binary TracePoint
$(deriveJSON aesonOptions ''TracePoint)

instance ToJSONKey TracePoint
instance FromJSONKey TracePoint

-- | Data from various phases of a single time step
type ArenaTraceStep location signal = Map TracePoint (ArenaState location signal)

data TraceMonitor location signal s =
  TraceMonitor {
    arenaStateArchive :: STRef s (Seq (ArenaTraceStep location signal)),
    arenaActivationRangeTable :: MArenaActivationRangeTable s
    }
  deriving (Typeable, Generic)



makeTraceMonitor ::
  [ColonyInitialState location signal]
  -> ST s (TraceMonitor location signal s)

makeTraceMonitor colonyInits = do
  arenaStateArchive <- newSTRef Seq.empty
  arenaActivationRangeTable <- makeMArenaActivationRangeTable colonyInits
  return TraceMonitor{..}


traceMonitorAdd ::
  TracePoint
  -> TraceMonitor location signal s
  -> MArenaState location signal s
  -> ST s ()

traceMonitorAdd tracePoint TraceMonitor{..} arenaState = do
  memento <- snapshot arenaState
  modifySTRef' arenaStateArchive (insert tracePoint memento)
  updateArenaActivationRangeTables arenaActivationRangeTable arenaState
  where
    insert tp m history =
      case viewr history of
        EmptyR ->
          error "BaselMachine.Arena.traceMontiorAdd: empty history"
        past :> current ->
          past |> Map.insert tp m current


instance ArenaStepMonitor TraceMonitor where

  -- Begin a new history item
  monitorInit tm@TraceMonitor{..} mArena = do
    modifySTRef' arenaStateArchive (|> Map.empty)
    traceMonitorAdd Initial tm mArena
  monitorAfterBroadcastSend = traceMonitorAdd AfterBroadcasts
  -- monitorAfterBroadcastIncrementAge
  monitorAfterSynapses = traceMonitorAdd AfterSynapses
  monitorAfterReactions = traceMonitorAdd AfterReactions
  monitorAfterBroadcastEmit = traceMonitorAdd AfterEmitters



data ColonyTraceSummary =
  ColonyTraceSummary {

  trPerCellOutputPatterns :: IntMap IntSet,
  -- ^ indexed by cell index, a set of all patterns ever read
  -- from that cell for any form of output

  trPerCellPresynapticPatterns :: IntMap IntSet,
  -- ^ indexed by cell index, a set of all patterns ever read
  -- from that cell by a synapse

  trPerCellActivationRanges :: IntMap (UV.Vector Int, UV.Vector Int),
  -- ^ indexed by cell index, each pair consists of arrays of the
  -- (minimum, maximum) activation levels of each pattern in that
  -- cell over the entire trace

  trOverallActivationRanges :: (UV.Vector Int, UV.Vector Int)
  -- ^ arrays of the (minimum, maximum) activation levels of each
  -- pattern in all cells over the entire trace
  }
  deriving (Eq, Ord, Read, Show, Data, Typeable, Generic)

instance NFData ColonyTraceSummary
instance Binary ColonyTraceSummary
$(deriveJSON aesonOptions ''ColonyTraceSummary)


-- === Running an arena ===

mArenaStep ::
  (BroadcastSignal location signal,
   Ord signal,
   ArenaStepMonitor monitor,
   Show signal)
  =>
  monitor location signal s
  -> [BroadcastState location signal]
  -> MArenaState location signal s
  -> ST s ()

mArenaStep monitor inputs a@MArenaState{..} = do

  monitorInit monitor a

  mArenaIncorporateInputs inputs a
  mArenaPhaseBroadcastSend a
  monitorAfterBroadcastSend monitor a

  mArenaPhaseBroadcastIncrementAge a
  monitorAfterBroadcastIncrementAge monitor a

  mArenaPhaseSynapses a
  monitorAfterSynapses monitor a

  mArenaPhaseReactions a
  monitorAfterReactions monitor a

  mArenaPhaseBroadcastEmit a
  monitorAfterBroadcastEmit monitor a

{-# SPECIALIZE mArenaStep :: (Ord signal, Show signal, BroadcastSignal location signal) => BroadcastLogMonitor location signal s -> [BroadcastState location signal] -> MArenaState location signal s -> ST s () #-}

{-# SPECIALIZE mArenaStep :: (Ord signal, Show signal, BroadcastSignal location signal) => TraceMonitor location signal s -> [BroadcastState location signal] -> MArenaState location signal s -> ST s () #-}

{-# INLINE mArenaStep #-}

mArenaRun ::
  (BroadcastSignal location signal,
   Ord signal,
   ArenaStepMonitor monitor,
   Show signal)
  =>
  monitor location signal s
  -> (MArenaState location signal s -> ST s Bool)
  -> MArenaState location signal s
  -> Int
  -- ^ maximum number of steps to run
  -> [[BroadcastState location signal]]
  -- ^ for each time step, a list of inputs to feed into the arena
  -> ST s ()

mArenaRun _ _ _ 0 _ = return ()

mArenaRun monitor stopCheckAxn arenaState numSteps [] =
  mArenaRun monitor stopCheckAxn arenaState numSteps [[]]

mArenaRun monitor stopCheckAxn arenaState numSteps (input:moreInputs) = do
  mArenaStep monitor input arenaState
  stopSignalled <- stopCheckAxn arenaState
  unless stopSignalled $ do
    mArenaRun monitor stopCheckAxn arenaState (numSteps - 1) moreInputs

{-# SPECIALIZE mArenaRun :: (Ord signal, Show signal, BroadcastSignal location signal) => BroadcastLogMonitor location signal s -> (MArenaState location signal s -> ST s Bool) -> MArenaState location signal s -> Int -> [[BroadcastState location signal]] -> ST s () #-}

{-# SPECIALIZE mArenaRun :: (Ord signal, Show signal, BroadcastSignal location signal) => TraceMonitor location signal s -> (MArenaState location signal s -> ST s Bool) -> MArenaState location signal s -> Int -> [[BroadcastState location signal]] -> ST s () #-}

{-# INLINE mArenaRun #-}


gRunArena ::
  forall location signal monitor s.
  (Eq signal,
   Ord signal,
   Show signal,
   Typeable signal,
   ArenaStepMonitor monitor,
   BroadcastSignal location signal)
  =>
  monitor location signal s
  -> [ColonyInitialState location signal]
  -> Maybe signal
  -- ^ stop when this signal is broadcast
  -> Int
  -- ^ maximum number of steps to run
  -> [[BroadcastState location signal]]
  -- ^ for each time step, a list of inputs to feed into the arena
  -> ST s (MArenaState location signal s)

gRunArena monitor colonyInits mStopSignal numSteps arenaInputs = do
  mArena <- makeMutableArena colonyInits []
  mArenaRun monitor stopCheckAxn mArena numSteps arenaInputs
  return mArena
  where
    stopCheckAxn :: MArenaState location signal s -> ST s Bool
    stopCheckAxn MArenaState{..} =
      case mStopSignal of
        Nothing -> return False -- never stop
        Just stopSignal -> do -- stop if particular signal emitted
          broadcastStates <- readSTRef arenaBroadcastStates
          let checks = [
                stopSignal == broadcastSignal
                | MBroadcastState{..} <- broadcastStates]
          return (or checks)


data ResourceSummary =
  ResourceSummary {
  totalReactionCount :: Int,
  totalReactionDeltas :: Int
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData ResourceSummary
instance Binary ResourceSummary
$(deriveJSON aesonOptions ''ResourceSummary)

instance Semigroup ResourceSummary where
  x <> y =
    ResourceSummary {
    totalReactionCount = totalReactionCount x + totalReactionCount y,
    totalReactionDeltas = totalReactionDeltas x + totalReactionDeltas y
    }


instance Monoid ResourceSummary where
  mempty = ResourceSummary {
    totalReactionCount = 0,
    totalReactionDeltas =  0
    }

  mappend = (<>)

mCellResourceSummary ::
  MCellState s -> ST s ResourceSummary

mCellResourceSummary MCellState{..} = do
  totalReactionCount <- readSTRef cellTotalReactionCount
  totalReactionDeltas <- readSTRef cellTotalReactionDeltas
  return ResourceSummary{..}



data ColonyResourceSummary =
  ColonyResourceSummary {
  colonyResourcesUsed ::
      ResourceSummary,
  colonyCellIxToResourcesUsed ::
      IntMap ResourceSummary
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData ColonyResourceSummary
instance Binary ColonyResourceSummary
$(deriveJSON aesonOptions ''ColonyResourceSummary)


mColonyResourceSummary ::
  MColonyState location signal s -> ST s ColonyResourceSummary

mColonyResourceSummary MColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  nodeAndSummaries <- forM (G.labNodes cellGraph) $ \(node, mCell) -> do
    summary <- mCellResourceSummary mCell
    return (node, summary)
  let colonyCellIxToResourcesUsed = IntMap.fromList nodeAndSummaries
      colonyResourcesUsed = mconcat (map snd nodeAndSummaries)
  return ColonyResourceSummary{..}



data ArenaRunSummary location signal =
  ArenaRunSummary {
  arenaInputs :: [[BroadcastState location signal]],
  arenaRunBroadcasts ::
      Seq [BroadcastState location signal],
    -- ^ snapshots of all arena broadcasts
  arenaColonyIxToResourcesUsed ::
      IntMap ColonyResourceSummary
  }
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance (NFData location, NFData signal) => NFData (ArenaRunSummary location signal)
instance (Binary location, Binary signal) => Binary (ArenaRunSummary location signal)
$(deriveJSON aesonOptions ''ArenaRunSummary)



runArena ::
  forall location signal.
  (Eq signal,
   Ord signal,
   Show signal,
   Typeable signal,
   BroadcastSignal location signal)
  =>
  [ColonyInitialState location signal]
  -> Maybe signal
  -- ^ stop when this signal is broadcast
  -> Int
  -- ^ maximum number of steps to run
  -> [[BroadcastState location signal]]
  -- ^ for each time step, a list of inputs to feed into the arena
  -> ArenaRunSummary location signal

runArena colonyInits mStopSignal numSteps arenaInputs = runST $ do
  broadcastArchive <- newSTRef Seq.empty
  let stepMonitor = BroadcastLogMonitor{..}
  mArena <- gRunArena stepMonitor colonyInits mStopSignal numSteps arenaInputs
  let MArenaState{..} = mArena
  arenaRunBroadcasts <- readSTRef broadcastArchive
  colonyIxToState <- readSTRef arenaColonyIxToState
  ixAndSummary <- forM (IntMap.toList colonyIxToState) $ \(n, mColonyState) -> do
    sum <- mColonyResourceSummary mColonyState
    return (n, sum)
  let arenaColonyIxToResourcesUsed = IntMap.fromList ixAndSummary
  return ArenaRunSummary{..}



data ArenaTraceSummary location signal =
  ArenaTraceSummary {
  trArenaInputs :: [[BroadcastState location signal]],
  trArenaSteps :: Seq (ArenaTraceStep location signal),
  trColonyIxToSummary :: IntMap ColonyTraceSummary,
  trArenaRunSummary :: ArenaRunSummary location signal
  }
  deriving (Eq, Read, Show, Generic, Typeable)

-- Ord and Data are not possible because of ArenaTraceStep

instance (NFData location, NFData signal) => NFData (ArenaTraceSummary location signal)
instance (Binary location, Binary signal) => Binary (ArenaTraceSummary location signal)
$(deriveJSON aesonOptions ''ArenaTraceSummary)


runArenaTrace ::
  forall location signal.
  (Eq signal,
   Show signal,
   Ord signal,
   Typeable signal,
   BroadcastSignal location signal)
  =>
  [ColonyInitialState location signal]
  -> Maybe signal -- ^ stop when this signal is broadcast
  -> Int -- ^ maximum number of steps to run
  -> [[BroadcastState location signal]] -- ^ inputs to the arena
  -> ArenaTraceSummary location signal

runArenaTrace colonyInits mStopSignal numSteps arenaInputs = runST $ do
  stepMonitor@TraceMonitor{..} <- makeTraceMonitor colonyInits
  mArena <- gRunArena stepMonitor colonyInits mStopSignal numSteps arenaInputs
  let MArenaState{..} = mArena
  colonyIxPerCellActivationRanges <- snapshot arenaActivationRangeTable
  colonyIxToState <- readSTRef arenaColonyIxToState
  trArenaSteps <- readSTRef arenaStateArchive
  let colonyIxs = IntMap.keys colonyIxPerCellActivationRanges
  let trColonyIxToSummary =
        (IntMap.fromList
         [(colonyIx, ColonyTraceSummary{..})
         | colonyIx <- colonyIxs,
           let trPerCellActivationRanges =
                 fmap zap $ imGet colonyIx colonyIxPerCellActivationRanges
               trOverallActivationRanges =
                 (zap
                  $ mergeColonyActivationRangeTable
                  $ imGet colonyIx colonyIxPerCellActivationRanges)
               MColonyState{..} = imGet colonyIx colonyIxToState
               trPerCellOutputPatterns = perCellOutputPatterns colonySpec
               trPerCellPresynapticPatterns = perCellPresynapticPatterns colonySpec])
  colonyIxToState <- readSTRef arenaColonyIxToState
  ixAndSummary <- forM (IntMap.toList colonyIxToState) $ \(n, mColonyState) -> do
    sum <- mColonyResourceSummary mColonyState
    return (n, sum)
  let arenaColonyIxToResourcesUsed = IntMap.fromList ixAndSummary
      arenaRunBroadcasts = extractBroadcastSummaries trArenaSteps
      trArenaInputs = arenaInputs
      trArenaRunSummary = ArenaRunSummary{..}
  return ArenaTraceSummary{..}


zap CellActivationRangeTable{..} = (activationMin, activationMax)
zap :: CellActivationRangeTable -> (UV.Vector Int, UV.Vector Int)

extractBroadcastSummary ::
  ArenaTraceStep location signal
  -> [BroadcastState location signal]

extractBroadcastSummary traceStep =
  arenaBroadcastStates
  where
    ArenaState{..} = traceStep Map.! AfterEmitters



extractBroadcastSummaries ::
  Seq (ArenaTraceStep location signal)
  -> Seq [BroadcastState location signal]

extractBroadcastSummaries traceSteps =
  fmap extractBroadcastSummary traceSteps



-- | Return an 'IntMap' from cell index to an 'IntSet' of all
-- patterns read from that cell to generate output.

perCellOutputPatterns :: ColonySpec signal -> IntMap IntSet

perCellOutputPatterns ColonySpec{..} =
  (IntMap.fromListWith IntSet.union
    [ (sourceCellIx, IntSet.singleton sourcePattern)
      | BroadcastSpec{..} <- colonyBroadcastSpecs,
        let Source{..} = broadcastSource])



-- | Return an 'IntMap' from cell index to an 'IntSet' of all
-- patterns read from that cell by a synapse.

perCellPresynapticPatterns :: ColonySpec signal -> IntMap IntSet

perCellPresynapticPatterns ColonySpec{..} =
  (IntMap.fromListWith IntSet.union
   [ (sourceCellIx, IntSet.singleton sourcePattern)
   | ChannelSpec{..} <- colonySynapseSpecs,
     let Source{..} = channelSource])
