{-# LANGUAGE
BangPatterns,
MagicHash,
UnboxedTuples,
NoMonomorphismRestriction,
TypeFamilies,
FlexibleInstances,
MultiParamTypeClasses,
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
RankNTypes,
StandaloneDeriving,
DeriveDataTypeable,
DeriveGeneric,
TemplateHaskell,
OverloadedStrings,
FlexibleContexts
#-}

{-|

Module: BaselMachine.CircularBuffer

Description: Low-level circular buffer

Copyright: © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC)

Double-ended queue of finite capacity, just bytes or bits.  This
is meant to be as fast as possible, and is really meant just to
store 'Bool' values, so there is no error checking at all,
although the mod calculations ensure that all the memory
references are at least not going to cause a crash.  A quirk: A
full buffer looks just like an empty buffer, so don't let it fill
up!

Note that because of how GHC primitive operations are
implemented, an odd byte value is considered 'True' and an even
byte value is considered 'False'. Apparently 'isTrue#'
effectively checks the low order bit.

-}

module BaselMachine.CircularBuffer (
  CircularBuffer,
  makeCircularBuffer,
  pushFront,
  popFront,
  pushBack,
  popBack,
  pushFrontBool,
  popFrontBool,
  pushBackBool,
  popBackBool,
  getAllBytes,
  getAllBools
  )
       where

import GHC.Exts
import GHC.ST

import Control.Monad
import Control.Monad.ST ()

import Data.Sequence (Seq)
import qualified Data.Sequence as Seq

import WGM.Snapshot

-- Since there's no way to store an unboxed mutable number
-- outside an array, I'm using this ptrs array to store two
-- Int#s, 0:begin and 1:end.  Since I want all of this to work on
-- 32 bit and 64 bit architectures, it's simplest to just
-- allocate 16 bytes and read the begin two Int# items, no matter
-- what.  The convention is that the content of the buffer is
-- indexed by [begin, end), STL style, so that the buffer can be
-- empty, as in begin == end.


-- | Double-ended queue of finite capacity, just bytes.
data CircularBuffer s =
  CircularBuffer {
    numBytes :: Int#,
    ptrs :: MutableByteArray# s,
    contents :: MutableByteArray# s
    }

instance Snapshot (ST s) (CircularBuffer s) where
  type Memento (CircularBuffer s) = Seq Bool

  snapshot b = do
    bools <- getAllBools b
    return (Seq.fromList bools)


-- *grr* --
-- rem (-1) 5 = -1

-- I need mod.

-- This doesn't handle every case, but it gets the ones I need.
-- A more robust solution is ((k % m) + m) % m, but I'm trying to
-- save time here...
mod# :: Int# -> Int# -> Int#
mod# k m =
  remInt# (k +# m) m

{-# INLINE mod# #-}

makeCircularBuffer :: Int -> ST s (CircularBuffer s)
makeCircularBuffer (I# numBytes) =
  ST axn
  where
    axn s0 =
      let
        !(# sC, contents #) = newByteArray# numBytes s0
        !(# sP, ptrs #) = newByteArray# 16# sC
        -- fill with zeros
        sC0 = writeIntArray# ptrs 0# 0# sP
        sC1 = writeIntArray# ptrs 1# 0# sC0
      in
       (# sC1, CircularBuffer{..} #)

{-# INLINE makeCircularBuffer #-}


-- | Push a byte onto the front of the buffer.
pushFront :: CircularBuffer s -> Int -> ST s ()
pushFront b (I#  n) =
  ST (pushFront# b n)

{-# INLINE pushFront #-}


pushFront# :: CircularBuffer d -> Int# -> State# d -> (# State# d, () #)
pushFront# CircularBuffer{..} n s0 =
  let
    !(# s1, beginVal0 #) = readIntArray# ptrs 0# s0
    beginVal1 = mod# (beginVal0 -# 1#) numBytes
    s2 = writeInt8Array# contents beginVal1 n s1
    s3 = writeIntArray# ptrs 0# beginVal1 s2
  in
   (# s3, () #)

{-# INLINE pushFront# #-}


-- | Pop a byte off the front of the buffer.
popFront :: CircularBuffer s -> ST s Int
popFront b =
  ST axn
  where
    axn s0 =
      let
        !(# s1, n #) = popFront# b s0
      in
       (# s1, I# n #)

{-# INLINE popFront #-}


popFront# :: CircularBuffer d -> State# d -> (# State# d, Int# #)
popFront# CircularBuffer{..} s0 =
  let
    !(# s1, beginVal0 #) = readIntArray# ptrs 0# s0
    !(# s2, n #) = readInt8Array# contents beginVal0 s1
    beginVal1 = mod# (beginVal0 +# 1#) numBytes
    s3 = writeIntArray# ptrs 0# beginVal1 s2
  in
   (# s3, n #)

{-# INLINE popFront# #-}


-- | Push a byte onto the back of the buffer.
pushBack :: CircularBuffer s -> Int -> ST s ()
pushBack b (I#  n) =
  ST (pushBack# b n)

{-# INLINE pushBack #-}


pushBack# :: CircularBuffer d -> Int# -> State# d -> (# State# d, () #)
pushBack# CircularBuffer{..} n s0 =
  let
    !(# s1, endVal0 #) = readIntArray# ptrs 1# s0
    endVal1 = mod# (endVal0 +# 1#) numBytes
    s2 = writeInt8Array# contents endVal0 n s1
    s3 = writeIntArray# ptrs 1# endVal1 s2
  in
   (# s3, () #)

{-# INLINE pushBack# #-}


-- | Pop a byte off the back of the buffer.
popBack :: CircularBuffer s -> ST s Int
popBack b =
  ST axn
  where
    axn s0 =
      let
        !(# s1, n #) = popBack# b s0
      in
       (# s1, I# n #)

{-# INLINE popBack #-}


popBack# :: CircularBuffer d -> State# d -> (# State# d, Int# #)
popBack# CircularBuffer{..} s0 =
  let
    !(# s1, endVal0 #) = readIntArray# ptrs 1# s0
    endVal1 = mod# (endVal0 -# 1#) numBytes
    !(# s2, n #) = readInt8Array# contents endVal1 s1
    s3 = writeIntArray# ptrs 1# endVal1 s2
  in
   (# s3, n #)

{-# INLINE popBack# #-}


pushFrontBool :: CircularBuffer s -> Bool -> ST s ()
pushFrontBool b True = ST (pushFront# b 1#)
pushFrontBool b False = ST (pushFront# b 0#)
{-# INLINE pushFrontBool #-}

pushBackBool :: CircularBuffer s -> Bool -> ST s ()
pushBackBool b True = ST (pushBack# b 1#)
pushBackBool b False = ST (pushBack# b 0#)
{-# INLINE pushBackBool #-}


popBackBool :: CircularBuffer s -> ST s Bool
popBackBool b =
  ST axn
  where
    axn s0 =
      let
        !(# s1, n #) = popBack# b s0
      in
       (# s1, isTrue# n #)

{-# INLINE popBackBool #-}


popFrontBool :: CircularBuffer s -> ST s Bool
popFrontBool b =
  ST axn
  where
    axn s0 =
      let
        !(# s1, n #) = popFront# b s0
      in
       (# s1, isTrue# n #)

{-# INLINE popFrontBool #-}


-- | Fetch all bytes, from back to front.
getAllBytes :: CircularBuffer s -> ST s [Int]
getAllBytes CircularBuffer{..} =
  ST axn
  where
    axn s0 =
      let
        !(# s1, beginVal #) = readIntArray# ptrs 0# s0
        !(# s2, endVal #) = readIntArray# ptrs 1# s1

        loop# j s0 =
          if isTrue# (j ==# endVal)
          then (# s0, [] #)
          else
            let
              !(# s1, b #) = readInt8Array# contents j s0
              !(# s2, bs #) = loop# (mod# (j +# 1#) numBytes) s1
            in
             (# s2, I# b : bs #)

      in
       loop# beginVal s0

{-# INLINE getAllBytes #-}


-- | Fetch all bytes as booleans, from back to front.
getAllBools :: CircularBuffer s -> ST s [Bool]
getAllBools CircularBuffer{..} =
  ST axn
  where
    axn s0 =
      let
        !(# s1, beginVal #) = readIntArray# ptrs 0# s0
        !(# s2, endVal #) = readIntArray# ptrs 1# s1

        loop# j s0 =
          if isTrue# (j ==# endVal)
          then (# s0, [] #)
          else
            let
              !(# s1, b #) = readInt8Array# contents j s0
              !(# s2, bs #) = loop# (mod# (j +# 1#) numBytes) s1
            in
             (# s2, isTrue# b : bs #)

      in
       loop# beginVal s0

{-# INLINE getAllBools #-}

_test1 :: Int
_test1 = runST $ do
  b <- makeCircularBuffer 5
  pushBack b 1
  popBack b

_test2 :: Int
_test2 = runST $ do
  b <- makeCircularBuffer 5
  pushFront b 2
  popFront b

_test3 :: [Int]
_test3 = runST $ do
  b <- makeCircularBuffer 5
  pushBack b 1
  pushFront b 2
  pushBack b 3
  pushFront b 4
  getAllBytes b

_test4 :: ([Int], [Int], Seq Bool, Seq Bool)
_test4 = runST $ do
  b <- makeCircularBuffer 5
  _ <- forM [0 .. 3] (pushFront b)
  bytes1 <- getAllBytes b
  bools1 <- snapshot b
  _ <- replicateM 4 (popBack b)
  _ <- forM [0 .. 3] (pushFront b)
  _ <- replicateM 2 (popBack b)
  bytes2 <- getAllBytes b
  bools2 <- snapshot b
  return (bytes1, bytes2, bools1, bools2)
