{-# LANGUAGE RecordWildCards, RankNTypes, DeriveDataTypeable, DeriveGeneric #-}

{-|

Module      : BaselMachine.Compiler

Description : Functions that compile instructions to the internal representation

Copyright   :  © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module BaselMachine.Compiler (
  -- * Main interface
  ProgramInformation(..),
  ReactionProgram,
  compileUnchecked
  ) where

import GHC.Generics

import Control.DeepSeq
import Data.Aeson
import Data.Binary
import Data.Data
import qualified Data.IntSet as IntSet

import BaselMachine.Instruction
import BaselMachine.VMViaPrim


-- | Information assembled during compilation

data ProgramInformation = ProgramInformation {
  instructions :: [Instruction],
  -- ^ instructions used to build this machine
  reactions :: [Reaction],
  -- ^ reactions extracted from the instructions
  patternsReferenced :: IntSet.IntSet
  -- ^ set of patterns referenced by the program
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData ProgramInformation
instance Binary ProgramInformation
instance ToJSON ProgramInformation
instance FromJSON ProgramInformation



-- | Compile the given list of 'Instructions', without checking
-- for mistakes.

compileUnchecked ::
  [Instruction] -- ^ the program
  -> (ProgramInformation, ReactionProgram)

compileUnchecked instructions =
  let
    reactions =
      concatMap splitInstruction (zip [0..] instructions)

    basicReactions = [(sPat, thr, tPat, delta)
                      | Reaction{..} <- reactions]

    compiledProgram = primCompile basicReactions

    patternsReferenced =
      IntSet.fromList $ concat [[sPat, tPat] | Reaction{..} <- reactions]
  in
   (ProgramInformation{..}, compiledProgram)
