{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, DeriveDataTypeable, DeriveGeneric, NoMonomorphismRestriction, ParallelListComp, TemplateHaskell #-}

{-|

Module: BaselMachine.EfficaciousAnalysis

Description: Tools for analyzing UM programs

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : POSIX

-}


module BaselMachine.EfficaciousAnalysis (
  EfficaciousAnalysis(..),
  CellEfficaciousAnalysis(..),
  efficaciousAnalysis,
  ) where

import GHC.Generics

import Control.DeepSeq
import Data.Aeson
import Data.Aeson.TH
--import Data.Binary
import Data.Data
import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
import Data.IntSet (IntSet)
import qualified Data.IntSet as IntSet
import Data.Map.Strict (Map)
import qualified Data.Map.Strict as Map
import Data.Set (Set)
import qualified Data.Set as Set
--import Debug.Trace
--import Data.Typeable
import qualified Data.Vector.Unboxed as UV
import Data.Vector.Unboxed (Vector, (!))

import BaselMachine.Arena
import BaselMachine.Compiler
import BaselMachine.Instruction




-- | (Private) Starting with the given patterns, repeatedly add
-- reactions from the given list of instructions that activate or
-- inhibit a pattern of interest.  Go until a closed network of
-- co-referential reactions has been built.  Return
-- the reactions of interest, and target patterns found by this
-- process.

efficaciousConnections ::
  IntSet
  -- ^ interesting patterns to start with
  -> Set Reaction
  -- ^ each instruction fragment to consider
  -> (Set Reaction, IntSet)

efficaciousConnections startingPatterns startingRxns =
  let
    -- closureLoop :: Set Reaction -> Set Reacion -> Set Int -> ...
    closureLoop remainingRxns includedRxns includedTPats =
      let
        -- Split remaining reactions into:
        --
        -- moreRxns = those that target a member of includedTPats
        --
        -- and
        --
        -- leftoverRxns = those that don't

        (moreRxns, leftoverRxns) =
          (Set.partition
           (\rxn -> tPat rxn `Set.member` includedTPats)
           remainingRxns)

        -- The switch patterns of moreFrgs are now of interest as
        -- targets of other instructions
        moreTPats = Set.map sPat moreRxns
      in
       if Set.null moreRxns
       then -- no more reactions of interest
         (includedRxns, includedTPats)
       else let -- more links to go; prep to loop

         -- We've found more fragments to include in the interesting list
         nextRxns = moreRxns `Set.union` includedRxns

         -- And more target patterns of interest
         nextTPats = moreTPats `Set.union` includedTPats
         in
          -- trace ("closureLoop:\n includedTPats="
          --       ++ show includedTPats
          --       ++ "\n moreTPats="
          --       ++ show moreTPats
          --       ++ "\n nextTpats="
          --       ++ show nextTPats ++ "\n") $
          closureLoop leftoverRxns nextRxns nextTPats


    -- Run the main loop to get
    --
    -- effRxnsSet = set of efficacious fragments
    --
    -- effTPatsSet = set of patterns targetted by efficacious fragments
    --  (Has to be a Set rather than an IntSet because of an fmap)
    (effRxnsSet, effTPatsSet) =
      closureLoop startingRxns Set.empty (Set.fromDistinctAscList (IntSet.toAscList startingPatterns))

  in
   (effRxnsSet, IntSet.fromDistinctAscList (Set.toAscList effTPatsSet))


-- | This data type collects a bunch of pieces of information
-- that are produced together by 'efficaciousAnalysis' for
-- efficiency.  This information is for a single cell.

data CellEfficaciousAnalysis =
  CellEfficaciousAnalysis {

  cellEfficaciousInstructionIxs :: IntSet,
  -- ^ the set of indices of efficacious instructions

  cellEfficaciousReactionSet :: Set Reaction,
  -- ^ set of efficacious 'Reaction's

  cellEfficaciousTargetPatterns :: IntSet
  -- ^ set of all target patterns in all efficacious 'Reaction's
  } deriving (Eq, Ord, Read, Show, Data, Typeable, Generic)

instance NFData CellEfficaciousAnalysis
$(deriveJSON defaultOptions ''CellEfficaciousAnalysis)


-- | This data type collects a bunch of pieces of information
-- that are produced together by 'efficaciousAnalysis' for
-- efficiency.  This record is for an entire colony.

data EfficaciousAnalysis = EfficaciousAnalysis {
  {-
  -- Captured for debugging purposes
  perCellOutputPatterns :: IntMap IntSet,
  perCellPresynapticPatterns :: IntMap IntSet,
  perCellInterestingPatterns :: IntMap IntSet,
  perCellActivationRanges :: IntMap (Vector Int, Vector Int),
  -}


  perCellEfficacious :: IntMap CellEfficaciousAnalysis,
  -- ^ map from cell indices to record of efficatious
  -- instructions and reactions within that cell

  allEfficaciousInstructionIxs :: IntSet,
  -- ^ indices of all efficacious 'Instruction's across all cells

  allEfficaciousReactions :: Set Reaction,
  -- ^ set of all efficacious 'Reaction's across all cells

  allEfficaciousReactionCounts :: [(Reaction, Int)]
  -- ^ list of efficacious 'Reaction's and how many copies of
  -- that reaction there are in the genome; the 'Reaction'
  -- objects are modified by setting their 'index' field to 0.
  -- (I would like to make this a Map object but that causes
  -- problems deriving instances related to JSON.)

  } deriving (Eq, Ord, Read, Show, Data, Typeable, Generic)

instance NFData EfficaciousAnalysis
$(deriveJSON defaultOptions ''EfficaciousAnalysis)


-- | Find the efficacious instructions and reactions.

efficaciousAnalysis ::
  ProgramInformation
  -> [ColonyTraceSummary] -- ^ Trace results from all possible inputs
  -> EfficaciousAnalysis

efficaciousAnalysis ProgramInformation{..} traceResults =
  EfficaciousAnalysis {..}
  where
    reactionSet :: Set Reaction
    reactionSet = Set.fromList reactions

    reactionCountMap :: Map Reaction Int
    reactionCountMap = Map.unionsWith (+) [
      Map.singleton r{index=0} 1
      | r <- reactions]

    -- Interesting patterns come from these sources:
    -- * Output of the colony via emitters
    -- * Presynaptic patterns

    perCellOutputPatterns :: IntMap IntSet
    perCellOutputPatterns =
      (IntMap.unionsWith IntSet.union
       (map trPerCellOutputPatterns traceResults))


    perCellPresynapticPatterns :: IntMap IntSet
    perCellPresynapticPatterns =
      (IntMap.unionsWith IntSet.union
       (map trPerCellPresynapticPatterns traceResults))

    perCellInterestingPatterns :: IntMap IntSet
    perCellInterestingPatterns =
      IntMap.unionWith IntSet.union perCellOutputPatterns perCellPresynapticPatterns

    perCellActivationRanges :: IntMap (Vector Int, Vector Int)
    perCellActivationRanges =
      (IntMap.unionsWith mergeVectorMinMax
       (map trPerCellActivationRanges traceResults))


    effForCell :: Int -> IntSet -> CellEfficaciousAnalysis
    effForCell cellIx interestingPatterns =
      let
        (_minActivations, maxActivations) =
          perCellActivationRanges IntMap.! cellIx

        -- all reactions that are ever active in any of the traced runs
        activeRxns = Set.filter isEverActive reactionSet

        isEverActive Reaction{..} =
          maxActivations ! sPat >= thr

        (effRxns, effPats) =
          efficaciousConnections interestingPatterns activeRxns

        effIxs = IntSet.fromAscList (map index (Set.toAscList effRxns))

      in
        CellEfficaciousAnalysis {
        cellEfficaciousInstructionIxs = effIxs,
        cellEfficaciousReactionSet = effRxns,
        cellEfficaciousTargetPatterns = effPats
        }

    -- This is the main calculation:
    perCellEfficacious =
      IntMap.mapWithKey effForCell perCellInterestingPatterns

    allEfficaciousInstructionIxs =
      IntSet.unions [cellEfficaciousInstructionIxs
                    | CellEfficaciousAnalysis{..} <- IntMap.elems perCellEfficacious]

    allEfficaciousReactions =
      Set.unions [cellEfficaciousReactionSet
                 | CellEfficaciousAnalysis{..} <- IntMap.elems perCellEfficacious]

    allEfficaciousReactionCounts =
      [(r0, reactionCountMap Map.! r0)
       | r <- Set.toList allEfficaciousReactions,
         let r0 = r{index=0}]


-- | Given (minV1, maxV1) and (minV2, maxV2) of equal length
-- vectors, produce vectors (minV3, maxV3), where each element of
-- minV3 is the minimum of the corresponding elements of minV1
-- and minV2, and each element of maxV3 is the maximum of the
-- corresponding elements of maxV1 and maxV2.

mergeVectorMinMax ::
  (Vector Int, Vector Int)
  -> (Vector Int, Vector Int)
  -> (Vector Int, Vector Int)

mergeVectorMinMax (min1, max1) (min2, max2) =
  (UV.zipWith min min1 min2,
   UV.zipWith max max1 max2)
