{-# LANGUAGE RecordWildCards, TypeFamilies, RecordWildCards, NamedFieldPuns, DeriveDataTypeable, DeriveGeneric #-}

{-|

Module      : BaselMachine.Instruction

Description : Virtual machine instructions

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module BaselMachine.Instruction
       (Instruction(..),
        Reaction(..),
        printNumberedInstruction,
        printInstruction,
        printReaction,
        splitInstruction
        )
       where

import GHC.Generics

import Control.DeepSeq
-- import Control.Monad

import Data.Aeson
import Data.Binary
-- import Data.Bits
import Data.Data
-- import Data.Default
import Data.List
-- import Data.String
import Data.Typeable

import Text.Printf

import WGM.ShortShow


-- | One genetically encoded instruction, consisting of a switch
-- pattern and threshold, and a list of reactions.

data Instruction = Instruction {
  switchPattern :: Int,
  switchThreshold :: Int,
  targetsAndDeltas :: [(Int, Int)] -- ^ list of (target pattern, delta)
  } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance NFData Instruction
instance Binary Instruction

-- I'm putting in explicit JSON serialization instances that don't
-- include all the key strings, just so things don't get so huge
instance ToJSON Instruction where
  toJSON Instruction{..} =
    toJSON (switchPattern, switchThreshold, targetsAndDeltas)

instance FromJSON Instruction where
  parseJSON obj = do
    (switchPattern, switchThreshold, targetsAndDeltas) <- parseJSON obj
    return Instruction{..}





-- | This uses fixed field widths, so it's appropriate for a
-- table in a fixed-width font.
printNumberedInstruction :: (Int, Instruction) -> String
printNumberedInstruction (j, instruction) =
  (printf "%5d: %s" j (printInstruction instruction))

-- | This uses fixed field widths, so it's appropriate for a
-- table in a fixed-width font.
printInstruction :: Instruction -> String
printInstruction Instruction{..} =
  (printf "A[%2d]≥%2d ⇒ %s"
   switchPattern switchThreshold
   reactions)
  where
    reactions = intercalate " " [
      printf "A[%2d]+=%2d" target delta
      | (target, delta) <- targetsAndDeltas ]

-- | For finer-grained program analysis, a 'Reaction' consists of
-- a switch pattern, a threshold, and a delta.  The
-- interpretation is that if @A[sPat] >= thr@ add @delta@ to
-- @A[tPat]@.  Each genetically encoded instruction can encode
-- several reactions.
data Reaction =
  Reaction {
    index :: Int, -- ^ location in list of instructions generated from a genome
    sPat :: Int, -- ^ switch pattern
    thr :: Int, -- ^ threshold
    tPat :: Int,  -- ^ target pattern
    delta :: Int -- ^ amount to add to @A[tPat]@ when active
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData Reaction
instance ToJSON Reaction
instance FromJSON Reaction
instance Binary Reaction

-- | This uses fixed field widths, so it's appropriate for a
-- table in a fixed-width font.
printReaction :: Reaction -> String
printReaction Reaction{..} =
  (printf "%5d: A[%2d]≥%2d ⇒ A[%2d]+=%2d"
   index sPat thr tPat delta)

-- | Split an indexed instruction into reactions

splitInstruction ::
  (Int, Instruction) -> [Reaction]
splitInstruction (j, Instruction {..}) =
  [Reaction j switchPattern switchThreshold target delta
   | (target, delta) <- targetsAndDeltas]

-- I don't know why, but GHC complains that 'printInstruction' is
-- not in scope if this isn't at the bottom... !?!?
instance ShortShow Instruction where
  shortShow = printInstruction
