{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables,
DeriveDataTypeable, DeriveGeneric, FlexibleContexts,
OverloadedStrings,
GADTs, StandaloneDeriving,
DataKinds #-}

{-|
Module      : Digiostelium.Report

Description : Generating reports from agent queries

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Some implementation notes: The madness of ByteStrings:

* Aeson and Pandoc use Data.ByteString.Lazy.

* Yaml uses Data.ByteString.

So I'm going to do everything in terms of
'Data.ByteString.Buidler' and deal with that, since it'll be more
uniform and less crazy.

-}

module Digiostelium.OldReport (
  writeReports
  )
where

import GHC.Generics()

import Control.Exception
import Control.Monad

import Data.Aeson (ToJSON(..), FromJSON(..), object, (.=))
import qualified Data.Aeson as J
import Data.ByteString.Builder
--import qualified Data.ByteString as BS
import qualified Data.ByteString.UTF8 as U
import qualified Data.ByteString.Lazy as BSL
--import qualified Data.ByteString.Lazy.UTF8 as LU
import Data.Data
import Data.Default
import Data.List
import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
--import Data.IntSet (IntSet)
--import qualified Data.IntSet as IntSet
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set
import Data.Monoid
--import Data.String
import Data.Text (Text)
import qualified Data.Text as Text
import Data.Text.Encoding (encodeUtf8Builder)
-- import qualified Data.Text.IO as TextIO
import Data.Typeable()
import qualified Data.Yaml as Yaml
import System.Directory
import System.FilePath
import System.IO
import Text.Pandoc
import Text.Pandoc.Builder
import Text.Printf
import Text.XML.Light

--import Digiostelium.Agent
--import Digiostelium.ColonyToStyledNetwork
--import Digiostelium.Genome
--import Digiostelium.GraphML
import BaselMachine.Instruction
import Digiostelium.Logging
--import Digiostelium.Problem
import Digiostelium.ProblemInventory
import Digiostelium.QueryResult as QR
import Digiostelium.ReportFormatTypes
import Digiostelium.Snip


-- | Not every combination of report, output format, etc. is
-- sensible, so ...
data ReportMismatchException where
  ResultFormatMismatchException ::
    forall gs problem.
    (UsefulGenomeSpec gs, UsefulProblem problem) =>
    gs -> problem -> AgentResultX gs problem  -> OutputFormat -> ReportMismatchException

  FormatDestinationMismatchException ::
    OutputFormat -> OutputDestination -> ReportMismatchException

  EnsembleNotAvailableForOutputFormatException ::
    OutputFormat -> ReportMismatchException

deriving instance Typeable ReportMismatchException
deriving instance Show ReportMismatchException

instance Exception ReportMismatchException



writeReports  ::
  (UsefulGenomeSpec gs,
   UsefulProblem prob)
  =>
  gs
  -> prob
  -> OutputFormat
  -> OutputDestination
  -> Map Integer (AgentResultX gs prob)
  -> IO ()

writeReports genomeSpec problem outputFormat AllToStdout rMap = do
  hPutBuilder stdout b
  hPutStr stdout "\n"
  where
    b = buildEnsembleReport genomeSpec problem outputFormat rMap

writeReports genomeSpec problem outputFormat
  (ToFilePerAgentId overwrite pathTemplate)
  rMap = do
  ensureDirectoryForFile pathTemplate
  forM_ (Map.toList rMap) $ \(n, r) -> do
    let reportBSBuilder = buildReport genomeSpec problem outputFormat r
        p = printf pathTemplate n
    fileExists <- doesFileExist p
    if (not fileExists || overwrite)
      then do
      msgInfo (printf "Writing report to %s" p)
      withBinaryFile p WriteMode $ \h -> do
        hSetBuffering h (BlockBuffering Nothing)
        hPutBuilder h reportBSBuilder
      else do
      msgInfo (printf "File %s exists, won't overwrite" p)



buildReport  ::
  (UsefulGenomeSpec gs,
   UsefulProblem prob)
  =>
  gs
  -> prob
  -> OutputFormat
  -> AgentResultX gs prob
  -> Builder

buildReport genomeSpec problem JSONOutput r =
  lazyByteString
  $ J.encode r -- gives a Data.Lazy.ByteString in UTF8

buildReport genomeSpec problem YAMLOutput r =
  byteString
  $ Yaml.encode r -- gives a Data.ByteString in UTF8

buildReport genomeSpec problem PlainTextOutput r =
  encodeUtf8Builder
  $ runPurePandoc
  $ writePlainNoWrap
  $ setTitle (reportTitle r)
  $ doc
  $ buildReportBlocks genomeSpec problem r

buildReport genomeSpec problem MarkdownOutput r =
  encodeUtf8Builder
  $ runPurePandoc
  $ writeMarkdown def
  $ setTitle (reportTitle r)
  $ doc
  $ buildReportBlocks genomeSpec problem r

{-
buildReport genomeSpec problem GraphMLOutput QR.ReactionNetwork{..} =
  -- do various conversions here... --
  $ showTopElement
  $ networkToGraphMLDocument reactionNetwork
-}

buildReport genomeSpec problem outputFormat r =
  throw (ResultFormatMismatchException genomeSpec problem r outputFormat)

buildEnsembleReport  ::
  (UsefulGenomeSpec gs,
   UsefulProblem prob)
  =>
  gs
  -> prob
  -> OutputFormat
  -> Map Integer (AgentResultX gs prob)
  -> Builder

buildEnsembleReport genomeSpec problem JSONOutput rMap =
  lazyByteString $ J.encode rMap

buildEnsembleReport genomeSpec problem YAMLOutput rMap =
  byteString $ Yaml.encode rMap

buildEnsembleReport genomeSpec problem PlainTextOutput rMap =
  encodeUtf8Builder
  $ runPurePandoc
  $ writePlainNoWrap
  $ buildMultiDoc genomeSpec problem rMap

buildEnsembleReport genomeSpec problem MarkdownOutput rMap =
  encodeUtf8Builder
  $ runPurePandoc
  $ writeMarkdown def
  $ buildMultiDoc genomeSpec problem rMap

buildEnsembleReport genomeSpec problem GraphMLOutput rMap =
  throw (EnsembleNotAvailableForOutputFormatException GraphMLOutput)



-- | Given a path to a file, create the directory it would be in
-- if necessary.
ensureDirectoryForFile :: FilePath -> IO ()
ensureDirectoryForFile filePath =
  createDirectoryIfMissing True (dropFileName filePath)


-- This can maybe be generalized but this is what I have in mind:

-- | How to convert query results to Pandoc documents
buildReportBlocks ::
  (UsefulProblem prob,
   UsefulGenomeSpec gs) =>
  gs
  -> prob
  -> AgentResultX gs prob
  -> Blocks





buildReportBlocks genomeSpec problem Efficacious{..} =
  (header 1 $ text $ printf "Agent %d" id)
  <>
  (para
   ("Rating:" <> space <> texts rating <> linebreak <>
    "Lifespan:" <> space <> texts lifespan <> linebreak <>
    "Parents:" <> space <> texts parentIds <> linebreak <>
    "Genome length:" <> space <> texts genomeLength <> linebreak))
  <>
  cellReports
  where
    cellReports =
      (mconcat
       (map cellReport (IntMap.keys efficaciousReactions)))

    cellReport cellIx =
      (header 2 $ text $ printf "Cell %d" cellIx)
      <>
      (header 3 "Efficacious reactions")
      <>
      reactionBlock (efficaciousReactions IntMap.! cellIx)
      <>
      (header 3 "Efficacious instructions")
      <>
      instructionBlock (efficaciousInstructions IntMap.! cellIx)

buildReportBlocks genomeSpec problem RatingInputs{..} =
  (header 1 $ text $ printf "Agent %d" id)
  <>
  (para ("Rating:" <> space <> texts rating))
  <>
  mconcat inputReports

  where
    inputReports =
      map inputReport (zip [(0::Int) ..] allInputs)

    inputReport (j, (input, output, totalReactionCount, totalReactionDeltas, rating)) =
      (para
       ("Rating input #" <> space <> texts j <> ":" <> linebreak <>
        "Input :" <> space <> toInlines problem input <> linebreak <>
        "Output:" <> space <> toInlines problem output <> linebreak <>
        "Rxn ct:" <> space <> texts totalReactionCount <> linebreak <>
        "Tot dl:" <> space <> texts totalReactionDeltas <> linebreak <>
        "Rating:" <> space <> texts rating <> linebreak))

-- General case: Block of YAML code
buildReportBlocks genomeSpec problem r =
  codeBlock $ U.toString $ Yaml.encode r

-- Tools for the big Efficacious report


-- Tools for building document elements --
-- I haven't used these in a while, so assume these no longer work as written.

{-
divSectionLike :: String -> Inlines -> Blocks -> Blocks
divSectionLike cls title content =
  (divWith ("", [cls], [])
   (divTitle title
    <>
    divContent content))

divSection :: Inlines -> Blocks -> Blocks
divSection = divSectionLike "section"

divTitle :: Inlines -> Blocks
divTitle = divWith ("", ["title"], []) . plain

divContent :: Blocks -> Blocks
divContent = divWith ("", ["content"], [])


divProgram :: String -> Blocks
divProgram = divWith ("", ["program"], []) . codeBlock

divTraceBlock :: Blocks -> Blocks
divTraceBlock =
  divWith ("", ["trace"], [])
-}

-- Tools for making links --
{-

-- | Return @Just "fromLeftParent"@ or @Just "fromRightParent"@
-- to indicate which parent an instruction came from, or
-- 'Nothing' if it's from no source or multiple sources.

styleByParent :: IsString str => GeneDetails t -> Maybe str

styleByParent (GeneDetails {..}) =
  let
    parentIxs = IntSet.fromList
                [parentIx
                | (span, FromChromosome {..}) <- geneSource]
  in
   if IntSet.size parentIxs /= 1
   then Nothing
   else
     if 0 `IntSet.member` parentIxs
     then Just "fromLeftParent"
     else Just "fromRightParent"


viewAgentLink :: Integer -> Inlines -> Inlines
viewAgentLink agentId label =
  link url title label
  where
    url =
      printf "javascript:document.doViewAgent(%d);void(0);" agentId
    title =
      printf "View agent %d" agentId

viewAgentLinkAuto :: Integer -> Inlines
viewAgentLinkAuto agentId =
  viewAgentLink agentId (text (printf "%d" agentId))

viewTraceLink ::
  (ToJSON input)
  =>
  Integer -> input -> String -> Inlines

viewTraceLink agentId input label =
  link url title (str label)
  where
    inputAsBSL = J.encode input
    inputAsStr = LU.toString inputAsBSL
    url =
      (printf "javascript:document.doViewTrace(%d,%s);void(0);"
       agentId
       inputAsStr)
    title =
      "trace this input"

viewTraceLinkAuto ::
  (ToJSON input)
  =>
  Integer -> input  -> Inlines

viewTraceLinkAuto agentId input =
  viewTraceLink agentId input "Trace"

viewRecordingLink ::
  (ToJSON input)
  =>
  Integer -> input -> String -> Inlines

viewRecordingLink agentId input label =
  link url title (str label)
  where
    inputAsBSL = J.encode input
    inputAsStr = LU.toString inputAsBSL
    url =
      (printf "javascript:document.doViewRecording(%d,%s);void(0);"
       agentId
       inputAsStr)
    title =
      "get recording"

viewRecordingLinkAuto ::
  (ToJSON input)
  =>
  Integer -> input  -> Inlines

viewRecordingLinkAuto agentId input =
  viewRecordingLink agentId input "Recording"


loadQueryLink ::
  Show q =>
  q -> Inlines -> Inlines
loadQueryLink query label =
  link url title label
  where
    urlBSL =
      (BSL.concat
       ["javascript:document.doLoadAgents(",
        J.encode $ object [
          "query" .= show query ],
        ");void(0);"])
    url = LU.toString urlBSL
    title =
      printf "Load %s" (show query)

loadGenerationLink :: Integer -> Inlines -> Inlines
loadGenerationLink gen label =
  "<link>"
  -- loadQueryLink (Generation gen) label

loadGenerationLinkAuto :: Integer -> Inlines
loadGenerationLinkAuto gen =
  loadGenerationLink gen (fromString $ show gen)

viewGenomeDiagramLink :: Integer -> Inlines -> Inlines
viewGenomeDiagramLink agentId label =
  link url title label
  where
    url =
      printf "javascript:document.doViewGenomeDiagram(%d);void(0);" agentId
    title =
      printf "View genome diagram for agent %d" agentId

viewGenomeDiagramLinkAuto :: Integer -> Inlines
viewGenomeDiagramLinkAuto agentId =
  viewGenomeDiagramLink agentId "genome diagram"

-}
