{-# LANGUAGE TypeFamilies, FlexibleContexts, RecordWildCards, NamedFieldPuns, TemplateHaskell, DeriveDataTypeable, DeriveGeneric #-}

{-|

Module: Digiostelium.PairProblem

Description : Abstractions for representing a problem to be solved by pairs of interacting agents.

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.PairProblem where

import GHC.Generics

import Control.DeepSeq

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
import Data.Sequence (Seq)
--import qualified Data.Sequence as Seq
import Data.Typeable

import BaselMachine.Arena
import BaselMachine.Compiler (ProgramInformation, ReactionProgram)
import BaselMachine.Instruction()
import BaselMachine.Signal
import WGM.ParUtil

import Digiostelium.InputOutputProblem

data PairLocation = Here | There
  deriving (Enum, Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData PairLocation
instance Binary PairLocation
$(deriveJSON defaultOptions ''PairLocation)

instance ToJSONKey PairLocation
instance FromJSONKey PairLocation

type Pair t = (t, t)



pairProblemInitialArenaSpec problem (senderInfoProg, receiverInfoProg) =
  [senderInitialState, receiverInitialState]
  where
    -- I'm sure I'll want to generalize this
    senderColonySpec = ColonySpec {
      colonyCellCount = 1,
      colonyCellIxToName = IntMap.fromList [(0, "Cell")],
      colonySynapseSpecs = [],
      colonyBroadcastSpecs = [
                             ],
      colonySensorSpecs = [
                          ]
      }

    senderInitialState = ColonyInitialState {
      initName = "Sender",
      initColonySpec = senderColonySpec,
      initProgram = senderInfoProg,
      initLocation = Here
      }

    receiverColonySpec = ColonySpec {
      colonyCellCount = 1,
      colonyCellIxToName = IntMap.fromList [(0, "Cell")],
      colonySynapseSpecs = [],
      colonyBroadcastSpecs = [
                             ],
      colonySensorSpecs = [
                          ]
      }

    receiverInitialState = ColonyInitialState {
      initName = "Receiver",
      initColonySpec = receiverColonySpec,
      initLocation = There
      }

{-
class (
  InputOutputProblem problem,
  ArenaInitializationSpec problem ~ Pair (ProgramInformation, ReactionProgram)
  ) =>
  PairProblem problem where

  -- | Build the problem's initial colony state from the result
  -- of compiling the genome.
  initialArenaSpec ::
    problem
    -> Pair (ProgramInformation, ReactionProgram)
    -> [ColonyInitialState location signal]
-}


{-

-- | Trace a colony on a single input

traceColonyOn ::
  (SingletonProblem problem,
  NFData (Location problem),
  Eq (Signal problem),
  Show (Signal problem),
  Typeable (Signal problem))
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> Input problem
  -> (Input problem,
     Output problem,
     ArenaTraceSummary (Location problem) (Signal problem))

traceColonyOn problem progAndInfo input =
  (input, output, trSummary)
  where
    colonyInput = problemInputToArenaInput problem input
    trSummary@ArenaTraceSummary{..} =
      (runArenaTrace
       [initialColonyState problem progAndInfo]
       (stopSignal problem)
       (length colonyInput)
       colonyInput)

    ArenaRunSummary{..} = trArenaRunSummary
    output = arenaOutputToProblemOutput problem arenaRunBroadcasts



-- | Trace a colony on all inputs used for rating it, including
-- all the trace mementos, which can take up a lot of memory.

traceColonyOnAll ::
  (SingletonProblem problem)
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> [(Input problem,
      Output problem,
      ArenaTraceSummary (Location problem) (Signal problem))]

traceColonyOnAll problem progAndInfo =
  parListDeep
  $ map (traceColonyOn problem progAndInfo)
  $ inputsForRating problem



runColonyOn ::
  (SingletonProblem problem)
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> Input problem
  -> (Output problem, Int, Int)
  -- ^ problem output, total reaction count, total reaction deltas

runColonyOn problem progAndInfo input =
  (problemOutput, totalReactionCount, totalReactionDeltas)
  where
    colonyInput = problemInputToArenaInput problem input
    ArenaRunSummary{..} =
      (runArena
       [initialColonyState problem progAndInfo]
       (stopSignal problem)
       (length colonyInput)
       colonyInput)
    problemOutput =
      arenaOutputToProblemOutput problem arenaRunBroadcasts
    ColonyResourceSummary{..} = arenaColonyIxToResourcesUsed IntMap.! 0
    ResourceSummary{..} = colonyResourcesUsed


-- | Run a colony on all inputs used for rating it.
runColonyOnAll ::
  (SingletonProblem problem)
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> [(Input problem, Output problem, Int, Int, Rating problem)]

runColonyOnAll problem progAndInfo =
  parListDeep $
  [(input, output, totalReactionCount, totalReactionDeltas,
    rateResult problem input output totalReactionCount totalReactionDeltas)
  | input <- inputsForRating problem,
    let (output, totalReactionCount, totalReactionDeltas) =
          runColonyOn problem progAndInfo input]


-- | Rate the colony on each possible input and add the ratings
-- together.
rateColonyAdd ::
  (Num (Rating problem),
   SingletonProblem problem) =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> Rating problem

rateColonyAdd problem progAndInfo =
  sum [r | (_, _, _, _, r) <- runColonyOnAll problem progAndInfo]

-}
