{-# LANGUAGE BangPatterns, RecordWildCards, NamedFieldPuns, MultiParamTypeClasses, FlexibleInstances, FlexibleContexts, ScopedTypeVariables, UndecidableInstances, StandaloneDeriving, DeriveDataTypeable, ParallelListComp, DeriveGeneric, TemplateHaskell #-}

module Digiostelium.Agent (
  Egg(..),
  Agent(..),
  Lifespan,
  birthDate,
  deathDate,
  summarizeAgent
  ) where

import GHC.Generics

import Control.DeepSeq

import Data.Aeson.TH
import Data.Binary
import Data.Data
-- import Data.Typeable
import qualified Data.Sequence as Seq


import Digiostelium.AesonUtil
import Digiostelium.Genome

import System.Random.RandomST

{- | Either just a bith time, or a birth time and death time -}
type Lifespan = Either Integer (Integer, Integer)

data Egg genome =
  Egg {
    parentIds :: [Integer], -- ^ list of parent indices
    genome :: genome
    }
  deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance NFData genome => NFData (Egg genome)
instance Binary genome => Binary (Egg genome)
$(deriveJSON myJSONOpts ''Egg)

instance HasGenome (Egg genome) genome where
  getGenome = genome

data Agent genome rating =
  Agent {
    identifier :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    egg :: Egg genome
    }
  deriving (Read, Show, Eq, Typeable, Data, Generic) -- no Ord, see below

instance (NFData genome, NFData rating) => NFData (Agent genome rating)
instance (Binary genome, Binary rating) => Binary (Agent genome rating)
$(deriveJSON myJSONOpts ''Agent)

instance HasGenome (Agent genome rating) genome where
  getGenome = genome . egg


-- | 'Agent' ordering is that agent 1 comes before (= 'LT') agent
-- 2 if its rating is higher, or if their ratings are equal and
-- agent 1 has a higher identifier (= younger).  Since sorting and
-- comparison functions give ascending order, some of the lines in
-- this implementation have to be sort of backward so that ascending
-- order of 'Agent' is descending rating.
--
-- The reason for secondary ordering by identifier: Older
-- highly-rated genomes should be replaced by younger equally-rated
-- genomes as the simulation continues.  This seems to be necessary
-- to avoid stagnation.

instance (Ord rating, Eq genome) => Ord (Agent genome rating) where
  compare a1 a2 =
    let
      cR = compare (rating a2) (rating a1)
      cI = compare (identifier a2) (identifier a1)
    in
     case cR of
       EQ -> cI
       _ -> cR

birthDate :: Agent genome rating -> Integer
birthDate (Agent {..}) =
  case lifespan of
    Left b -> b
    Right (b, _) -> b

deathDate :: Agent genome rating -> Maybe Integer
deathDate (Agent {..}) =
  case lifespan of
    Left _ -> Nothing
    Right (_, d) -> Just d

-- | Pull a sample child of two parents 'Agent's

instance
  (Sample (gs, genome, genome) genome)
  =>
  (Sample
   (gs,
    Agent genome rating,
    Agent genome rating)
   (Egg genome)) where
  pullSample (gs, lParent, rParent) = do
    let lGenome = genome (egg lParent)
    let rGenome = genome (egg rParent)
    let parentIds = [identifier lParent, identifier rParent]
    genome <- pullSample (gs, lGenome, rGenome)
    return (Egg {..})


-- | Pull two 'Agents' at random uniformly, then do mutation and
-- crossover.

instance
  (Sample (gs, genome, genome) genome)
  =>
  (Sample
   (gs, Seq.Seq (Agent genome rating))
   (Egg genome)) where
  pullSample (gs, genomesToBreed) = do
    pLeft :: Agent genome rating <- pullSample genomesToBreed
    pRight :: Agent genome rating <- pullSample genomesToBreed
    pullSample (gs, pLeft, pRight)


summarizeAgent ::
  (Show rating) =>
  Agent genome rating -> String
summarizeAgent (Agent {egg = Egg {..}, ..}) =
  unlines [
    "identifier: " ++ show identifier,
    "  rating:   " ++ show rating,
    "  lifespan: " ++ show lifespan,
    "  parents: " ++ show parentIds ]
