{-# LANGUAGE BangPatterns, RecordWildCards, NamedFieldPuns, DisambiguateRecordFields, FlexibleInstances, MultiParamTypeClasses, ScopedTypeVariables, DeriveDataTypeable, TemplateHaskell, DeriveGeneric #-}

{-|
Module      : Digiostelium.Chromosome

Description : A chromosome (as bit string) type and associated functions

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

The 'Chromosome' type is a 'BitString' but generally interpreted
to consist of genes (blocks of bits of some uniform size).  This
module contains functions for performing mutations, crossover,
and alignment on 'Chromosome's.

-}

module Digiostelium.Chromosome (
  Chromosome,
  CrossoverType(..),
  SetCrossoverType(..),
  CrossoverSpec(..),
  BlockAlignmentSpec(..),
  Crossover(..),
  SingleChromosomeMutation(..),
  SingleChromosomeMutationSpec(..),
  chromosomeLength,
  chromosomeSubstring,
  chromosomeTake,
  chromosomeDrop,
  chromosomeJoin,
  mutationInterval,
  shortShowMutation,
  mutateChromosome,
  allPossibleSingleChromosomeMutations,
  makeBlockCrossoverSpec,
  makeDefaultBlockAlignmentSpec,
  blockAlignSmithWaterman,
  blockSelfAlignmentMatrix,
  blockAlignmentMatrix
  ) where

import GHC.Generics

import Control.DeepSeq
-- import Control.Exception

-- import Debug.Trace

import Data.Aeson.TH
import Data.Array
import Data.Binary
import Data.Data
import Data.Default
import Data.List
import Data.Typeable

import Text.Printf

import Digiostelium.AesonUtil
import Digiostelium.BlockSize
import Data.BitString
import WGM.ParUtil
import System.Random.RandomST
import Digiostelium.SmithWaterman


type Chromosome = BitString

chromosomeLength :: Chromosome -> Int
chromosomeLength c =
  bsNumBits c

chromosomeSubstring :: Int -> Int -> Chromosome -> BitString
chromosomeSubstring begin end c =
  bsSubstring begin end c

chromosomeTake :: Int -> Chromosome -> Chromosome
chromosomeTake n c =
  bsTake n c

chromosomeDrop :: Int -> Chromosome -> Chromosome
chromosomeDrop n c =
  bsDrop n c

chromosomeJoin :: Chromosome -> Chromosome -> Chromosome
chromosomeJoin cLeft cRight =
  bsJoin cLeft cRight

data CrossoverType =
      NoCrossover
      -- ^ always return the left parent with no crossover
      | SinglePointKeepTail
        -- ^ crossover at a single point, preserving any tail of
        -- the longer genome whenever possible; leads to growth
        -- of genome length
      | SinglePointLengthNeutral
        -- ^ crossover at a single point, preserving the mean
        -- length of genomes but losing the tail of the longer
        -- genome sometimes
      | SinglePointAligned BlockAlignmentSpec
        -- ^ do Smith-Waterman alignment and pick a point on the
        -- resulting alignment path
      deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance BlockSize CrossoverType where

  getBlockSize (SinglePointAligned bas) =
    getBlockSize bas

  getBlockSize _ = 0

  setBlockSize blockSize (SinglePointAligned bas) =
    SinglePointAligned (setBlockSize blockSize bas)

  setBlockSize _ x = x

instance Default CrossoverType where

  def = SinglePointAligned def



data BlockAlignmentSpec = BlockAlignmentSpec {
      matchWeight :: Int, -- ^ should be positive; points per matching bit
      mismatchWeight :: Int, -- ^ should be negative; points per mismatching bit
      gapWeight :: Int, -- ^ should be negative; points per gap bit
      alignmentBlockSize :: Int -- ^ bit length of a block
      } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance BlockSize BlockAlignmentSpec where

  getBlockSize BlockAlignmentSpec{..} = alignmentBlockSize

  setBlockSize alignmentBlockSize bas =
    bas {alignmentBlockSize}

instance Default BlockAlignmentSpec where

  def = makeDefaultBlockAlignmentSpec 1



-- | Make a reasonable default 'BlockAlignmentSpec'
makeDefaultBlockAlignmentSpec ::
  Int -- ^ alignment block size
  -> BlockAlignmentSpec

makeDefaultBlockAlignmentSpec alignmentBlockSize =
  BlockAlignmentSpec {
    matchWeight = 2,
    mismatchWeight = -1,
    gapWeight = -1,
    ..}



class SetCrossoverType t where
  -- | Set the crossover type of an object, setting its block
  -- size to the block size of the rest of the object
  setCrossoverType :: CrossoverType -> t -> t


data CrossoverSpec = CrossoverSpec {
  crossoverType :: !CrossoverType,
  crossoverBlockSize :: !Int
  } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance BlockSize CrossoverSpec where

  getBlockSize CrossoverSpec{..} =
    crossoverBlockSize

  setBlockSize crossoverBlockSize cSpec =
    makeBlockCrossoverSpec (crossoverType cSpec) crossoverBlockSize

instance SetCrossoverType CrossoverSpec where

  setCrossoverType ct cs =
    cs {
    crossoverType = setBlockSize (getBlockSize cs) ct
    }

makeBlockCrossoverSpec :: CrossoverType -> Int -> CrossoverSpec
makeBlockCrossoverSpec crossoverType crossoverBlockSize =
  CrossoverSpec {
  crossoverType = setBlockSize crossoverBlockSize crossoverType,
  crossoverBlockSize
  }


-- | A crossover record means that a crossover takes [0..leftEnd)
-- from the left chromosome and combines it with [rightBegin.. )
-- from the right chromosome

data Crossover = Crossover {
  leftEnd :: !Int,
  rightBegin :: !Int
  } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)


-- | A 'SingleChromosomeMutation' is a bit flip, duplication, or
-- deletion that affects one chromosome only.  Flips happen
-- first, then duplications, then deletions.

data SingleChromosomeMutation =
  Flip !Int
  | DuplicateInterval !Int !Int -- [begin, end) bit indices
  | DeleteInterval !Int !Int -- [begin, end) bit indices
  deriving (Read, Show, Eq, Typeable, Data, Generic)

instance Ord SingleChromosomeMutation where
  compare (Flip m) (Flip n) = compare m n
  compare (Flip _) _ = LT
  compare _ (Flip _) = GT
  compare (DuplicateInterval mBegin mEnd) (DuplicateInterval nBegin nEnd) =
    case compare mBegin nBegin of
      EQ ->
        compare mEnd nEnd
      other ->
        other
  compare (DeleteInterval mBegin mEnd) (DeleteInterval nBegin nEnd) =
    case compare mBegin nBegin of
      EQ ->
        compare mEnd nEnd
      other ->
        other
  compare (DuplicateInterval mBegin mEnd) (DeleteInterval nBegin nEnd) =
    case compare mBegin nBegin of
      EQ ->
        case compare mEnd nEnd of
          EQ ->
            LT -- duplications first
          other ->
            other
      other ->
        other
  compare (DeleteInterval mBegin mEnd) (DuplicateInterval nBegin nEnd) =
    case compare mBegin nBegin of
      EQ ->
        case compare mEnd nEnd of
          EQ ->
            GT -- duplications first
          other ->
            other
      other ->
        other


-- | Return @[begin, end)@ for bits affected by this mutation
mutationInterval :: SingleChromosomeMutation -> (Int, Int)
mutationInterval (Flip n) = (n, n+1)
mutationInterval (DuplicateInterval begin end) = (begin, end)
mutationInterval (DeleteInterval begin end) = (begin, end)

shortShowMutation :: SingleChromosomeMutation -> String

shortShowMutation (Flip n) =
  "flp " ++ show n

shortShowMutation (DuplicateInterval begin end) =
  "dup " ++ show begin ++ " " ++ show end

shortShowMutation (DeleteInterval begin end) =
  "del " ++ show begin ++ " " ++ show end

-- | Apply a sequence of 'SingleChromosomeMutation's to a
-- 'Chromosome', resulting in a huge amount of additional
-- information

mutateChromosome ::
  Chromosome
  -> [SingleChromosomeMutation] -- ^ mutations to be applied
  -> (Chromosome,
      [((Int,Int), (Int,Int), [Int], BitString)])
  -- ^ The resulting 'Chromosome', plus a list of chunks
  -- annotated with the resulting interval in the child, the
  -- corresponding interval in the parent, a list of flipped bits
  -- relative to the beginning of the chunk, and the resulting
  -- 'BitString'


mutateChromosome pChromosome mutations =
  let
    flipped = [bitIx | Flip bitIx <- mutations]
    annChunks = mutationLoop pChromosome (sort mutations) 0 0
    rChromosome = bsRepack $ bsConcat [chunk | (_,_,chunk) <- annChunks]
    annPlusChunks =
      [(rInt, pInt, flips, chunk)
       | (rInt, pInt@(pBegin, pEnd), chunk) <- annChunks,
         let flips =
               [bitIx - pBegin
               | bitIx <- flipped, pBegin <= bitIx && bitIx <= pEnd]]
  in
   -- trace (show mutations) $
   (rChromosome, annPlusChunks)

-- | (Internal function) I need something that will go through
-- the mutations and give me lists of which bit ranges in the
-- mutated result map from which bit ranges in the original

mutationLoop ::
  Chromosome -- ^ pChromosome : the parent chromosome
  -> [SingleChromosomeMutation]
  -> Int -- ^ rBegin : where we are in the child chromosome
  -> Int -- ^ pBegin : where we are in the parent chromosome
  -> [((Int,Int), (Int,Int), BitString)] -- ^ all range mappings

mutationLoop
  pChromosome
  allMutations@(DeleteInterval begin end : moreMutations)
  rBegin
  pBegin
  | (pBegin < begin) =
    -- r: [===---)
    -- p: [===[xxx)---)
    let
      keptPrefixLength = begin - pBegin
      rEnd = rBegin + keptPrefixLength
      pEnd = pBegin + keptPrefixLength
    in
     ((rBegin, rEnd),
      (pBegin, pEnd),
      bsSubstring pBegin pEnd pChromosome) :
     mutationLoop
     pChromosome
     allMutations
     rEnd
     pEnd
     -- should reduce to this next case

  | (pBegin == begin) =
    -- r:  [---)
    -- p: [[xxx)---)
    let
      deletedLength = end - begin
      pEnd = pBegin + deletedLength
    in
     mutationLoop
     pChromosome
     moreMutations
     rBegin
     pEnd
  | otherwise =
      (error
       (printf "mutationLoop: unexpected case: DeleteInterval %d %d, %d %d"
        begin end rBegin pBegin))

mutationLoop
  pChromosome
  allMutations@(DuplicateInterval begin end : moreMutations)
  rBegin
  pBegin
  | (pBegin < begin) =
    -- r: [===[+++)+++)
    -- p: [===[+++))
    let
      keptPrefixLength = begin - pBegin
      rEnd = rBegin + keptPrefixLength
      pEnd = pBegin + keptPrefixLength
    in
     ((rBegin, rEnd),
      (pBegin, pEnd),
      bsSubstring pBegin pEnd pChromosome) :
     mutationLoop
     pChromosome
     allMutations
     rEnd
     pEnd
     -- should reduce to this next case
  | (pBegin == begin) =
    -- [[+++)[+++)---)
    -- [[+++)---)
    let
      dupLength = end - begin
      rEnd = rBegin + dupLength
      pEnd = pBegin + dupLength
    in
     ((rBegin, rEnd),
      (pBegin, pEnd),
      bsSubstring pBegin pEnd pChromosome) :
     mutationLoop
     pChromosome
     moreMutations
     rEnd
     pBegin

  | otherwise =
      (error
       (printf "mutationLoop: unexpected case: DuplicateInterval %d %d, %d %d"
        begin end rBegin pBegin))


-- all of these should be first in the list
mutationLoop
  pChromosome
  (Flip n : moreMutations)
  rBegin
  pBegin =
    mutationLoop
    (bsToggleBit n pChromosome)
    moreMutations
    rBegin
    pBegin

mutationLoop
  pChromosome
  []
  rBegin
  pBegin =
    let
      pEnd = bsNumBits pChromosome
      tailLength = pEnd - pBegin
      rEnd = rBegin + tailLength
    in
     if tailLength > 0
     then
       [((rBegin, rEnd),
         (pBegin, pEnd),
         bsSubstring pBegin pEnd pChromosome)]
     else
       []


data SingleChromosomeMutationSpec =
  SingleChromosomeMutationSpec {
    mutationBlockSize :: !Int,
    pToggle :: !Double, -- ^ probability of toggling a bit
    pDuplicate :: !Double, -- ^ probability of duplicating a block
    pDelete :: !Double, -- ^ probability of deleting a block
    minimumBlocks :: !Int -- ^ all chromosomes must have at least this many blocks
  } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance BlockSize SingleChromosomeMutationSpec where

  getBlockSize SingleChromosomeMutationSpec{..} =
    mutationBlockSize

  setBlockSize mutationBlockSize spec =
    spec {mutationBlockSize}


instance Default SingleChromosomeMutationSpec where
  def = SingleChromosomeMutationSpec {
    mutationBlockSize = 1,
    pToggle = 0.005,
    pDuplicate = 0.001,
    pDelete = 0.001,
    minimumBlocks = 4
    }

allPossibleSingleChromosomeMutations ::
  SingleChromosomeMutationSpec
  -> Int
  -> [SingleChromosomeMutation]

allPossibleSingleChromosomeMutations
  (SingleChromosomeMutationSpec {..})
  genomeLength =
    let
      nBlocks = genomeLength `div` mutationBlockSize
      eachBitToggled = map Flip [0 .. genomeLength-1]
      eachBlockDuplicated =
        [DuplicateInterval begin end
        | j <- [0 .. nBlocks - 1],
          let begin = mutationBlockSize * j,
          let end = begin + mutationBlockSize]
      eachBlockDeleted =
        [DeleteInterval begin end
        | j <- [0 .. nBlocks - 1],
          let begin = mutationBlockSize * j,
          let end = begin + mutationBlockSize]
    in
     concat [eachBitToggled, eachBlockDuplicated, eachBlockDeleted]

-- | Pull a list of 'Mutation's applicable to a chromosome of the
-- given length, sorted (flips, then duplications, then
-- deletions, in order of start index in between).

instance Sample (SingleChromosomeMutationSpec, Int)
         [SingleChromosomeMutation] where

  pullSample (SingleChromosomeMutationSpec {..}, genomeLength) =
    let
      nBlocks = (genomeLength `div` mutationBlockSize) - 1
    in do
      bitsToToggle <-
        pullSample (DiscretizedPoissonProcess pToggle (genomeLength - 1))
      blocksToDuplicate <-
        pullSample (DiscretizedPoissonProcess pDuplicate nBlocks)
      blocksToDelete <-
        pullSample (DiscretizedPoissonProcess pDelete nBlocks)
      -- never delete so many that the result is shorter than minimumBlocks
      let maxDeleteCount = max 0 (nBlocks + length blocksToDuplicate - minimumBlocks)
      return (sort (concat [[Flip c
                            | c <- bitsToToggle],
                            [DuplicateInterval begin end
                            | b <- blocksToDuplicate,
                              let begin = b * mutationBlockSize,
                              let end = begin + mutationBlockSize],
                            [DeleteInterval begin end
                            | b <- take maxDeleteCount blocksToDelete,
                              let begin = b * mutationBlockSize,
                              let end = begin + mutationBlockSize]]))


-- Generate crossover
instance Sample
         (CrossoverSpec, Chromosome, Chromosome)
         Crossover where

  pullSample (CrossoverSpec {..}, leftChromosome, rightChromosome) =
    let
      nBitsLeft = chromosomeLength leftChromosome
      nBitsRight = chromosomeLength rightChromosome
    in
     case crossoverType of
       NoCrossover ->
         return (Crossover nBitsLeft nBitsRight)
       SinglePointKeepTail -> do
         let nBlocksLeft = nBitsLeft `div` crossoverBlockSize
         crossoverPointBlocks <- pullSample (UniformDistribution (1, nBlocksLeft - 1))
         let crossoverIndex = crossoverPointBlocks * crossoverBlockSize
         return (Crossover crossoverIndex crossoverIndex)

       SinglePointLengthNeutral -> do
         let nBits = nBitsLeft `min` nBitsRight
         let nBlocks = nBits `div` crossoverBlockSize
         crossoverPointBlocks <- pullSample (UniformDistribution (1, nBlocks - 1))
         let crossoverIndex = crossoverPointBlocks * crossoverBlockSize
         return (Crossover crossoverIndex crossoverIndex)

       SinglePointAligned as@(BlockAlignmentSpec {..}) ->
         let
           alignmentResult = blockAlignSmithWaterman as leftChromosome rightChromosome
         in do
           -- pull one alignment point out of the best alignment path,
           -- so left block i aligns more or less with right block j
           -- combine [0..i) from the left with [j..end) from the right
           (i, j, step :: Step BitString) <- pullSample (bestPath alignmentResult)
           let iInBits = i*alignmentBlockSize
               jInBits = j*alignmentBlockSize
           {- (trace
            (concat [
                "Chromosome.pullSample[CrossoverSpec]:\nalignmentResult:\n",
                show alignmentResult,
                "\ni: ", show i,
                "\nj: ", show j,
                "\nstep: ", show step,
                "\n"])) -}
           return (Crossover iInBits jInBits)


blockAlignSmithWaterman ::
  BlockAlignmentSpec -> BitString -> BitString -> SmithWatermanResult BitString

blockAlignSmithWaterman as@(BlockAlignmentSpec {..}) leftChromosome rightChromosome =
 (alignSmithWaterman
  (gapWeight * alignmentBlockSize)
  (w as)
  (bsChunks alignmentBlockSize leftChromosome)
  (bsChunks alignmentBlockSize rightChromosome))

-- Weight of a block
w :: BlockAlignmentSpec -> BitString -> BitString -> Int
w (BlockAlignmentSpec {..}) b1 b2 =
  let
    d = bsHammingDistance b1 b2
    e = alignmentBlockSize - d
  in
   matchWeight * e + mismatchWeight * d

-- | Do all alignments for a lot of chromosomes with each other.
-- This function caches the break-down of the chromosomes into
-- blocks and ought to be faster than calling
-- 'blockAlignSmithWaterman' on each pair separatedly.

blockSelfAlignmentMatrix :: BlockAlignmentSpec -> Array Int Chromosome -> Array (Int, Int) (SmithWatermanResult BitString)
blockSelfAlignmentMatrix as@(BlockAlignmentSpec {..}) chromosomeAr =
    let
      chunkedChromosomeAr = fmap (bsChunks alignmentBlockSize) chromosomeAr
      bds@(a,b) = bounds chromosomeAr
      align b1 b2 =
        (alignSmithWaterman
         (gapWeight * alignmentBlockSize)
         (w as)
         b1 b2)

      -- go down the diagonal
      rDiag = [
        ((j, j), align c c)
        | j <- range bds,
          let c = chunkedChromosomeAr ! j]

      -- go through the upper triangle (north east)
      rT1 = [
        ((j, k), align cj ck)
        | j <- range bds,
          k <- range ((j+1), b),
          let cj = chunkedChromosomeAr ! j,
          let ck = chunkedChromosomeAr ! k]

      -- map the upper triangle to the lower triangle (south west)
      rT2 = [
        ((k, j), a)
        | ((j, k), a) <- rT1]

      results = concat [rDiag, rT1, rT2]
    in
     array ((a, a), (b, b)) (parListDeep results)


-- | Do all alignments for a lot of chromosomes from one source
-- with chromosomes from another.  This function caches the
-- break-down of the chromosomes into blocks and ought to be
-- faster than calling 'blockAlignSmithWaterman' on each pair
-- separatedly.

blockAlignmentMatrix ::
  BlockAlignmentSpec
  -> Array Int Chromosome
  -> Array Int Chromosome
  -> Array (Int, Int) (SmithWatermanResult BitString)

blockAlignmentMatrix as@(BlockAlignmentSpec {..}) xAr yAr =
    let
      chunkedXAr = fmap (bsChunks alignmentBlockSize) xAr
      chunkedYAr = fmap (bsChunks alignmentBlockSize) yAr

      (xFirst, xLast) = bounds chunkedXAr
      (yFirst, yLast) = bounds chunkedYAr

      align b1 b2 =
        (alignSmithWaterman
         (gapWeight * alignmentBlockSize)
         (w as)
         b1 b2)

      results = [
        ((j, k), align cj ck)
        | j <- range (bounds chunkedXAr),
          k <- range (bounds chunkedYAr),
          let cj = chunkedXAr ! j,
          let ck = chunkedYAr ! k]

    in
     array ((xFirst, yFirst), (xLast, yLast)) (parListDeep results)


-- Some test cases

testMutations = [DeleteInterval 5 15, Flip 20, DuplicateInterval  28 32, DuplicateInterval 34 36, DeleteInterval 34 36]

testString = (read "\"101100111000111100001111100000111101\"") :: BitString

testResult = mutateChromosome testString testMutations

{-
should give
.0        .5        .10       .15       .20       .25       .30       .35
 1 0 1 1 0 0 1 1 1 0 0 0 1 1 1 1 0 0 0 0 1 1 1 1 1 0 0 0 0 0 1 1 1 1 0 1
           x x x x x x x x x x           ~               * * * *

 1 0 1 1 0                     1 0 0 0 0 0 1 1 1 1 0 0 0  0 0 1 1 0 0 1 1   1 1 0 1
-}

testResultExpected = (read "\"1 0 1 1 0  1 0 0 0 0 0 1 1 1 1 0 0 0  0 0 1 1  0 0 1 1   1 1 0 1\"") :: BitString



instance NFData BlockAlignmentSpec
instance Binary BlockAlignmentSpec
$(deriveJSON myJSONOpts ''BlockAlignmentSpec)

instance NFData CrossoverType
instance Binary CrossoverType
$(deriveJSON myJSONOpts ''CrossoverType)

instance NFData CrossoverSpec
instance Binary CrossoverSpec
$(deriveJSON myJSONOpts ''CrossoverSpec)

instance NFData Crossover
instance Binary Crossover
$(deriveJSON myJSONOpts ''Crossover)

instance NFData SingleChromosomeMutation
instance Binary SingleChromosomeMutation
$(deriveJSON myJSONOpts ''SingleChromosomeMutation)

instance NFData SingleChromosomeMutationSpec
instance Binary SingleChromosomeMutationSpec
$(deriveJSON myJSONOpts ''SingleChromosomeMutationSpec)
