{-# LANGUAGE
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
FlexibleContexts,
ParallelListComp,
TypeFamilies,
DeriveDataTypeable,
DeriveGeneric,
OverloadedStrings,
BangPatterns,
TemplateHaskell,
MultiParamTypeClasses,
FlexibleInstances,
UndecidableInstances
#-}

{-|

Module      : Digiostelium.CoincidenceDetectorProblem

Description : Detecting two spikes in close succession

Copyright   :  © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Code for the problem of responding to two action potentials in
close succession

-}

module Digiostelium.CoincidenceDetectorProblem where

import GHC.Generics

import Control.DeepSeq

import Data.Aeson(ToJSON(..))
import Data.Aeson.TH
import Data.Data
import Data.Default
import qualified Data.IntMap as IntMap
-- import Data.IntMap (IntMap)
import Data.List
-- import Data.Map (Map)
import qualified Data.Map as Map
import Data.Semigroup
import Data.String
import Data.Typeable ()
import Text.Pandoc()
import Text.Pandoc.Builder

import WGM.ShortShow

import BaselMachine.Arena
-- import BaselMachine.Signal

import Digiostelium.AesonUtil
import Digiostelium.GeneralName
import Digiostelium.Genome
-- import Digiostelium.GraphML
import Digiostelium.InputOutputProblem
import Digiostelium.NeuronProblem
import Digiostelium.PandocUtil
import Digiostelium.Problem
import Digiostelium.Report
import Digiostelium.SingletonProblem



data CoincidenceDetectorProblem =
  CoincidenceDetectorProblem {
    neuronProblem :: NeuronProblem,
    -- ^ underlying 'NeuronProblem'
    inputActivation :: Int,
    -- ^ how much activation to add when the neuron receives input
    correctOutputPoints :: Double,
    -- ^ points earned for correct output during the window of an
    -- 'ActionPotentialTasklet'.  For a tasklet that requires a
    -- spike, these points are given if at least once during the
    -- output window the colony sets the output channel.  For a
    -- tasklet that requires silence, these points are given if
    -- the colony never sets the output channel.
    excessOutputPoints :: Double,
    -- ^ points lost for each step during the output window that
    -- the colony sets the output channel in excess of 1; usually
    -- of small magnitude
    falsePositivePoints :: Double,
    -- ^ points lost if an 'ActionPotentialTasklet' specifies no
    -- output should be generated, but the colony sets the output
    -- channel at any time during the output window; usually of
    -- medium magnitude.
    falseNegativePoints :: Double,
    -- ^ points lost if an 'ActionPotentialTasklet' specifies
    -- that a spike should be generated, but the colony never
    -- sets the output channel during the output window; usually
    -- of medium magnitude.
    quietPoints :: Double,
    -- ^ points earned per time step for generating no output
    -- when there is no input
    inputFailurePoints :: Double,
    -- ^ points lost per time step for generating output during
    -- a time when it's supposed to be quietly accumulating input
    trainingTasks :: TrainingSet,
    -- ^ inventory of tasks for training
    reactionCostFactor :: Double,
    -- ^ multiply the number of reactions used by this and
    -- subtract it from the total score
    deltaCostFactor :: Double
    -- ^ multiply the total of all reaction deltas applied less
    -- by this and subtract it from the total score
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)


-- | How to specify which task set is used for training.
-- Unless otherwise notes, these are stage-weighted sets of tasks.
-- That is, easier tasks are weighted more in the overall rating.
-- Learning tasks are for spike-timing-dependent plasticity experiments.
data TrainingSet =
  Medium -- ^ short list of medium-long spike trains
  | Long -- ^ everything for 'Medium' plus a long list of long spike trains
  | LearnMedium -- ^ everything for 'Medium' plus some learning tasks
  | LearnLong -- ^ everything for 'Long' plus many learning tasks
  | LearnLongEqWeight -- ^ same as 'LearnLong' but equally weighted
  | LearnLongDifficultyWeight -- ^ same as 'LearnLong' but difficult tasks are weighted more
  | Custom [CoincidenceDetectorTask]
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance GeneralName CoincidenceDetectorProblem where
  generalName _ = "CoincidenceDetector"

-- | Real number with semigroup operation of addition (<>) = (+)
type instance Rating CoincidenceDetectorProblem = Sum Double

type instance Location CoincidenceDetectorProblem = NullLocation

-- | There's only a signal or not, no structure
type instance Signal CoincidenceDetectorProblem = NullSignal

instance ArenaProblem CoincidenceDetectorProblem where
  -- stopSignal _ = Nothing
  minLandmarkRating _ = 0


-- | score factor and tasklet sequence
data CoincidenceDetectorTask =
  CoincidenceDetectorTask Double [ActionPotentialTasklet]
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData CoincidenceDetectorTask


data ActionPotentialTasklet =
  ActionPotentialTasklet {
    numLeadingOff :: Int,
    -- ^ begin with this many steps of no input
    actionPotentials :: ActionPotentialStream,
    -- ^ then send these spikes
    numTrailingOff :: Int,
    -- ^ end with this many steps of no input
    numOutputSteps :: Int,
    -- ^ neuron should spike during this time window from the end
    -- of the input or not at all
    shouldSpike :: Bool
    -- ^ should a spike during the output window
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData ActionPotentialTasklet

-- | list of (pulse width, time steps until next pulse begins)
type ActionPotentialStream = [(Int, Int)]


-- | Convert an 'ActionPotentialTasklet' to drawing characters.
-- It will consist of @'╌'@ at times when no input spike happens
-- and no output spike is expected, @'┸'@ at times when an input
-- spike happens and no output is expected,
-- @'■'@ at times after an input spike when an output spike is expected,
-- and @'□'@ at times after an input spike when no output spike is expected.
taskletToDrawing :: ActionPotentialTasklet -> String
taskletToDrawing ActionPotentialTasklet{..} =
  concat [
    replicate numLeadingOff '╌',
    actionPotentialStreamToDrawing actionPotentials,
    replicate numOutputSteps (if shouldSpike then '■' else '□'),
    replicate (numTrailingOff - numOutputSteps) '╌']

-- | Convert an 'ActionPotentialStream' to drawing characters
actionPotentialStreamToDrawing :: ActionPotentialStream -> String
actionPotentialStreamToDrawing =
  concatMap actionPotentialToDrawing

-- | Convert an action potential (pulse width, time steps until next pulse) to drawing characters
actionPotentialToDrawing :: (Int, Int) -> String
actionPotentialToDrawing (width, gap) =
  replicate width '┸' ++ replicate gap '╌'


taskletDuration :: ActionPotentialTasklet -> Int

taskletDuration t@ActionPotentialTasklet{..} =
  (numLeadingOff
   + taskletAPDuration t
   + numTrailingOff)


taskletAPDuration :: ActionPotentialTasklet -> Int

taskletAPDuration ActionPotentialTasklet{..} =
  sum [width + gap
      | (width, gap) <- actionPotentials]


data TaskletTally =
  TaskletTally {
    quietTimeSteps :: Int,
    preludeFailures :: Int,
    inputFailures :: Int,
    correctOutput :: Int,
    excessOutputs :: Int,
    falsePositives :: Int,
    falseNegatives :: Int,
    postludeFailures :: Int
    } deriving (Read, Show, Eq, Ord, Data, Typeable)



tallyScore :: CoincidenceDetectorProblem -> TaskletTally -> Double

tallyScore CoincidenceDetectorProblem{..} TaskletTally{..} =
    (fromIntegral correctOutput * correctOutputPoints
     - fromIntegral excessOutputs * excessOutputPoints
     - fromIntegral falsePositives * falsePositivePoints
     - fromIntegral falseNegatives * falseNegativePoints
     - fromIntegral inputFailures * inputFailurePoints
     + (fromIntegral
        (quietTimeSteps
         - preludeFailures
         - postludeFailures)) * quietPoints)

taskletTally :: ActionPotentialTasklet -> [Bool] -> TaskletTally
taskletTally t@ActionPotentialTasklet{..} output =
--  trace
--  ("CoincidenceDetectorProblem.taskletTally:\n"
--   ++ show t
--   ++ "\n"
--   ++ boolsTo01s output
--    ++ "\n"
--    ++ show tally)
  tally
  where
    quietTimeSteps = taskletDuration t - numOutputSteps

    (duringPrelude, afterPrelude) = splitAt numLeadingOff output

    -- count steps with high output during initial off prelude
    preludeFailures = sum (map fromEnum duringPrelude)

    apDuration = taskletAPDuration t

    (duringAPs, afterAPs) =
      splitAt apDuration afterPrelude

    --  high output during input action potentials, basically a
    --  false positive
    inputFailures = maximum (0 : map fromEnum duringAPs)

    (duringWindow, afterWindow) =
      splitAt numOutputSteps afterAPs

    -- count the number of on steps during output window
    apDuringOutputWindow = sum (map fromEnum duringWindow)

    correctOutput =
      if numOutputSteps == 0
      then 0
      else
        if shouldSpike
        then min 1 apDuringOutputWindow
        else
          if apDuringOutputWindow == 0
          then 1
          else 0

    excessOutputs =
      if shouldSpike && apDuringOutputWindow >= 1
      then apDuringOutputWindow - 1
      else 0

    falsePositives =
      if not shouldSpike && apDuringOutputWindow >= 1
      then 1
      else 0

    falseNegatives =
      if shouldSpike && apDuringOutputWindow < 1
      then 1
      else 0

    -- count steps with high output after output window
    postludeFailures = sum (map fromEnum afterWindow)

    tally = TaskletTally {..}



cdTaskTally :: [ActionPotentialTasklet] -> [Bool] -> [TaskletTally]

cdTaskTally [] _ = []

cdTaskTally (t:ts) output =
  taskletTally t thisOutput : cdTaskTally ts moreOutput
  where
    (thisOutput, moreOutput) =
      splitAt (taskletDuration t) output




--- Training sets ---

-- | Make all lists consisting of 'n' tasklets in all possible
-- orders.  There's probably some slick way to do this, but this
-- will do:

buildCDTasks ::
  Int -- ^ 'n'
  -> [a] -- ^ 'tasklets'
  -> [[a]]

buildCDTasks 1 tasklets =
  [[t]
  | t <- tasklets]

buildCDTasks numTaskletsPerTask tasklets =
  let
    taskTails = buildCDTasks (numTaskletsPerTask - 1) tasklets
  in
   [tasklet : taskTail
    | tasklet <- tasklets,
      taskTail <- taskTails]


-- | Medium length set of tasks
tsMedium :: [CoincidenceDetectorTask]
tsMedium =
  concat [
    [CoincidenceDetectorTask 10000 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 100 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2]
    ]


-- | Long set of tasks
tsLong :: [CoincidenceDetectorTask]
tsLong =
  concat
  [
    [CoincidenceDetectorTask 10000 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 100 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 3 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 3 cdTasklets2]
  ]


-- | Medium-length set of tasks, along with tasks for learning:
-- Some begin with a long period of inactivity, which should
-- weaken the input synapse and cause the neuron to spike less
-- easily.  Others begin with a period of very high activity,
-- which should strengthen the input synapse and cause the neuron
-- to spike more easily.
tsLearn :: [CoincidenceDetectorTask]
tsLearn =
  concat
  [
    [CoincidenceDetectorTask 10000 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 100 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2],

    -- There aren't many of these, so make them worth more each
    [CoincidenceDetectorTask 10 [tasklet]
    | tasklet <- cdTaskletsWeaken],

    [CoincidenceDetectorTask 1
     (highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen]
  ]

-- | Long set of tasks, along with tasks for learning.
tsLearnLong :: [CoincidenceDetectorTask]
tsLearnLong =
  concat
  [
    [CoincidenceDetectorTask 10000 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 100 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2],

    -- There aren't many of these, so make them worth more each
    [CoincidenceDetectorTask 10 [tasklet]
    | tasklet <- cdTaskletsWeaken],

    [CoincidenceDetectorTask 1
     (highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen],

    -- Weaken then strengthen
    [CoincidenceDetectorTask 1
     (cdTaskletsWeaken ++ highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen],

    -- Strengthen then weaken
    [CoincidenceDetectorTask 1
     (highActTasklets ++ [afterStrTasklet] ++ cdTaskletsWeaken)
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen]
  ]


-- | Long set of tasks, along with tasks for learning, but
-- all tasks are equally weighted.
tsLearnLongEqWeight :: [CoincidenceDetectorTask]
tsLearnLongEqWeight =
  concat
  [
    [CoincidenceDetectorTask 1 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 1 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2],

    [CoincidenceDetectorTask 1 [tasklet]
    | tasklet <- cdTaskletsWeaken],

    [CoincidenceDetectorTask 1
     (highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen],

    -- Weaken then strengthen
    [CoincidenceDetectorTask 1
     (cdTaskletsWeaken ++ highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen],

    -- Strengthen then weaken
    [CoincidenceDetectorTask 1
     (highActTasklets ++ [afterStrTasklet] ++ cdTaskletsWeaken)
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen]
  ]

-- | Long set of tasks, along with tasks for learning, but
-- all tasks are weighted in increasing order of difficulty.
tsLearnLongDifficultyWeight :: [CoincidenceDetectorTask]
tsLearnLongDifficultyWeight =
  concat
  [
    [CoincidenceDetectorTask 1 [tasklet]
    | tasklet <- cdTasklets1],
    [CoincidenceDetectorTask 10 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets1],
    [CoincidenceDetectorTask 10000 tasklets
    | tasklets <- buildCDTasks 2 cdTasklets2],

    [CoincidenceDetectorTask 100 [tasklet]
    | tasklet <- cdTaskletsWeaken],

    [CoincidenceDetectorTask 10000
     (highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen],

    -- Weaken then strengthen
    [CoincidenceDetectorTask 10000
     (cdTaskletsWeaken ++ highActTasklets ++ [afterStrTasklet])
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen],

    -- Strengthen then weaken
    [CoincidenceDetectorTask 10000
     (highActTasklets ++ [afterStrTasklet] ++ cdTaskletsWeaken)
     | highActTasklets <- buildCDTasks 3 cdTaskletsHighActivity,
       afterStrTasklet <- cdTaskletsAfterStrengthen]
  ]


--- Tasklets ---

-- | A list of the most basic coincidence detection tasklets:
-- various intervals with one or two spikes, some of which are
-- close enough that the receiving neuron should respond with an
-- action potential  of its own.

cdTasklets1 :: [ActionPotentialTasklet]

cdTasklets1 = [
  -- totally quiet, shouldn't trigger
  ActionPotentialTasklet {
     numLeadingOff = 0,
     actionPotentials = [],
     numTrailingOff = 20,
     numOutputSteps = 0,
     shouldSpike = False },
  -- isloated spike, shouldn't trigger
  ActionPotentialTasklet {
     numLeadingOff = 4,
     actionPotentials = [(1,0)],
     numTrailingOff = 15,
     numOutputSteps = 4,
     shouldSpike = False },
  -- back-to-back spikes, should trigger
  ActionPotentialTasklet {
     numLeadingOff = 4,
     actionPotentials = [(1,0),(1,0)],
     numTrailingOff = 14,
     numOutputSteps = 4,
     shouldSpike = True },
  -- two very close spikes, should trigger
  ActionPotentialTasklet {
     numLeadingOff = 4,
     actionPotentials = [(1,1),(1,0)],
     numTrailingOff = 13,
     numOutputSteps = 4,
     shouldSpike = True },
  -- two close spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 4,
    actionPotentials = [(1,2),(1,0)],
    numTrailingOff = 12,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two closeish spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 4,
    actionPotentials = [(1,3),(1,0)],
    numTrailingOff = 12,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two separated spikes, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 4,
    actionPotentials = [(1,6),(1,0)],
    numTrailingOff = 8,
    numOutputSteps = 4,
    shouldSpike = False },
  -- two widely separated spikes, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 4,
    actionPotentials = [(1,8),(1,0)],
    numTrailingOff = 6,
    numOutputSteps = 4,
    shouldSpike = False }
  ]


-- | A list of more difficult coincidence detection tasklets of
-- various lengths: various intervals with several spikes, some
-- of which are close enough that the receiving neuron should
-- respond with an action potential of its own.

cdTasklets2 :: [ActionPotentialTasklet]

cdTasklets2 = [
  -- totally quiet, shouldn't trigger
  ActionPotentialTasklet {
     numLeadingOff = 0,
     actionPotentials = [],
     numTrailingOff = 20,
     numOutputSteps = 0,
     shouldSpike = False },
  -- one spike, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 10,
    actionPotentials = [(1,0)],
    numTrailingOff = 10,
    numOutputSteps = 4,
    shouldSpike = False },
  -- back-to-back spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 5,
    actionPotentials = [(1,0),(1,0)],
    numTrailingOff = 9,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two close spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 8,
    actionPotentials = [(1,2),(1,0)],
    numTrailingOff = 20,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two very close spikes, should trigger, but with an odd spike beforehand
  ActionPotentialTasklet {
    numLeadingOff = 10,
    actionPotentials = [(1,8),(1,1),(1,0)],
    numTrailingOff = 6,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two closeish spikes, should trigger
  ActionPotentialTasklet {
    numLeadingOff = 10,
    actionPotentials = [(1,3),(1,0)],
    numTrailingOff = 20,
    numOutputSteps = 4,
    shouldSpike = True },
  -- two separated spikes, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 5,
    actionPotentials = [(1,6),(1,0)],
    numTrailingOff = 8,
    numOutputSteps = 4,
    shouldSpike = False },
  -- two widely separated spikes, shouldn't trigger
  ActionPotentialTasklet {
    numLeadingOff = 7,
    actionPotentials = [(1,8),(1,0)],
    numTrailingOff = 10,
    numOutputSteps = 4,
    shouldSpike = False }
  ]


-- | Tasklets in which there's a very long period of inactivity
-- during which the synaptic connection should "weaken".
-- Afterward, the cell should only respond to back-to-back
-- inputs.

cdTaskletsWeaken :: [ActionPotentialTasklet]

cdTaskletsWeaken = [
  -- very long wait followed by back-to-back spikes, should trigger
  ActionPotentialTasklet {
     numLeadingOff = 100,
     actionPotentials = [(1,0),(1,0)],
     numTrailingOff = 10,
     numOutputSteps = 4,
     shouldSpike = True },
  -- long wait, a loose spike, followed by back-to-back spikes, should trigger
  ActionPotentialTasklet {
     numLeadingOff = 50,
     actionPotentials = [(1,50),(1,0),(1,0)],
     numTrailingOff = 10,
     numOutputSteps = 4,
     shouldSpike = True },
  -- very long wait followed by two close spikes, should NOT trigger
  ActionPotentialTasklet {
    numLeadingOff = 100,
    actionPotentials = [(1,4),(1,0)],
    numTrailingOff = 20,
    numOutputSteps = 4,
    shouldSpike = False },
  -- long wait,a loose spike, followed by two close spikes, should NOT trigger
  ActionPotentialTasklet {
    numLeadingOff = 50,
    actionPotentials = [(1,50),(1,4),(1,0)],
    numTrailingOff = 20,
    numOutputSteps = 4,
    shouldSpike = False }
  ]

-- | Short tasklets for creating a period of high activity during
-- which the synaptic connection should "strengthen".  Afterward,
-- the cell should spike after slightly separated input pulses
-- that it shouldn't otherwise respond to.

cdTaskletsHighActivity :: [ActionPotentialTasklet]

cdTaskletsHighActivity = [
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,0),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,1),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,2),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True }
  {-
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,3),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,4),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,5),(1,0)],
     numTrailingOff = 5,
     numOutputSteps = 3,
     shouldSpike = True } -}
  ]


-- | Tasklets for use after high activity.  These are different
-- from 'tasklets0', because in these tasklets, input spikes
-- separated by 6 should trigger an action potential.

cdTaskletsAfterStrengthen :: [ActionPotentialTasklet]

cdTaskletsAfterStrengthen = [
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,6),(1,0)],
     numTrailingOff = 3,
     numOutputSteps = 3,
     shouldSpike = True },
  ActionPotentialTasklet {
     numLeadingOff = 2,
     actionPotentials = [(1,8),(1,0)],
     numTrailingOff = 3,
     numOutputSteps = 3,
     shouldSpike = False }
  ]







--- Problem templates ---

instance Default CoincidenceDetectorProblem where

  def = CoincidenceDetectorProblem {
    neuronProblem = def,
    inputActivation = 8,
    correctOutputPoints = 64,
    excessOutputPoints = 4,
    falsePositivePoints = 0,
    falseNegativePoints = 32,
    inputFailurePoints = 4,
    quietPoints = 1,
    trainingTasks = Custom [],
    reactionCostFactor = 0,
    deltaCostFactor = 0
    }


problemTemplateCDMedium :: CoincidenceDetectorProblem
problemTemplateCDMedium =
  def {trainingTasks = Medium}

problemTemplateCDLong :: CoincidenceDetectorProblem
problemTemplateCDLong =
  def {trainingTasks = Long}

problemTemplateCDLearnMedium :: CoincidenceDetectorProblem
problemTemplateCDLearnMedium =
  def {trainingTasks = LearnMedium}

problemTemplateCDLearnLong :: CoincidenceDetectorProblem
problemTemplateCDLearnLong =
  def {trainingTasks = LearnLong}

problemTemplateCDLearnLongEqWeight :: CoincidenceDetectorProblem
problemTemplateCDLearnLongEqWeight =
  def {trainingTasks = LearnLongEqWeight}

problemTemplateCDLearnLongDifficultyWeight :: CoincidenceDetectorProblem
problemTemplateCDLearnLongDifficultyWeight =
  def {trainingTasks = LearnLongDifficultyWeight}



instance InputOutputProblem CoincidenceDetectorProblem where

  -- | Activation to add at each time step
  type instance Input CoincidenceDetectorProblem = CoincidenceDetectorTask

  -- | Spike generated at each time step?
  type instance Output CoincidenceDetectorProblem = [Bool]

  problemInputToArenaInput
    CoincidenceDetectorProblem{..}
    (CoincidenceDetectorTask _ tasklets) =
    neuronProblemInputToArenaInput neuronProblem (toBools tasklets)

  arenaOutputToProblemOutput
    CoincidenceDetectorProblem{..}
    outputSeq =
    arenaOutputToNeuronProblemOutput neuronProblem outputSeq


  -- | List all inputs that are used for rating purposes
  inputsForRating CoincidenceDetectorProblem{..} =
    case trainingTasks of
      Custom ts ->
        ts
      Medium ->
        tsMedium
      Long ->
        tsLong
      LearnMedium ->
        tsLearn
      LearnLong ->
        tsLearnLong
      LearnLongEqWeight ->
        tsLearnLongEqWeight
      LearnLongDifficultyWeight ->
        tsLearnLongDifficultyWeight

  -- | Make these available for the 'SingletonProblem.TestSet' query
  inputsForTesting problem@CoincidenceDetectorProblem{..} =
    Map.fromList [
    ("Rating", inputsForRating problem),
    ("Medium", tsMedium),
    ("Long", tsLong),
    ("LearnMedium", tsLearn),
    ("LearnLong", tsLearnLong),
    ("LearnLongEqWeight", tsLearnLongEqWeight),
    ("LearnLongDifficultyWeight", tsLearnLongDifficultyWeight) ]


instance SingletonProblem CoincidenceDetectorProblem where

  rateResult
    problem@CoincidenceDetectorProblem{..}
    (CoincidenceDetectorTask scoreFactor input) output
    totalReactionCount
    totalReactionDeltas =
    Sum (scoreFactor * rawScore)
    where
      tallies = cdTaskTally input output
      tallyTotal = foldl' (\s t -> s + tallyScore problem t) 0 tallies
      reactionCountCost =
        reactionCostFactor * fromIntegral totalReactionCount
      reactionDeltaCost =
        deltaCostFactor * fromIntegral totalReactionDeltas
      reactionPenalty = reactionCountCost + reactionDeltaCost
      rawScore = tallyTotal - reactionPenalty


  initialColonyState CoincidenceDetectorProblem{..} initProgram =
    ColonyInitialState {..}
    where
      NeuronProblem{..} = neuronProblem

      initName = "Judy"
      initLocation = NullLocation
      colonyCellCount = 1
      colonyCellIxToName = IntMap.fromList [(0, "Neuron")]
      colonySynapseSpecs = []

      colonyBroadcastSpecs = [
        BroadcastSpec {
            broadcastSource = Source {
                sourceCellIx = 0,
                sourcePattern = outputPattern,
                sourceThreshold = outputThreshold,
                sourceActivationDelta = -outputCost
                },
            broadcastMaximumAge = Just 1,
            broadcastSignal = NullSignal
            } ]

      colonySensorSpecs = [
        SensorSpec {
            sensorSignal = NullSignal,
            sensorTarget = Target {
                targetCellIx = 0,
                targetPattern = inputPattern,
                targetActivationDelta = inputActivation } } ]

      initColonySpec = ColonySpec{..}







instance RateEggs CoincidenceDetectorProblem where
  rateEggs gs problem _adults eggs =
    singletonProblemRateEggs gs problem eggs

instance ToBools CoincidenceDetectorTask where
  toBools (CoincidenceDetectorTask _ tasklets) =
    toBools tasklets

instance ToBools ActionPotentialTasklet where
  toBools ActionPotentialTasklet{..} =
    (concat
     (concat
      [
        [replicate numLeadingOff False],
        [replicate pulseWidth True
         ++ replicate gap False
        | (pulseWidth, gap) <- actionPotentials],
        [replicate numTrailingOff False]]))


instance Sniplet CoincidenceDetectorProblem CoincidenceDetectorTask where
  toInlines _ (CoincidenceDetectorTask _ tasklets) =
    code $ fromString $ concatMap taskletToDrawing tasklets

instance Sniplet CoincidenceDetectorProblem [Bool] where
  toInlines _ output =
    code $ fromString $ boolsToDrawing output



instance
  (GenomeSpec genomeSpec,
   ShortShow (Interval genomeSpec),
   ToJSON (Genome genomeSpec) ) =>
  Reportable genomeSpec CoincidenceDetectorProblem SingletonProblemReportSpec


-- These have to be at the bottom because of some weirdness
-- involving whether types are in scope at the time the macro is
-- expanded.

$(deriveJSON myJSONOpts ''CoincidenceDetectorProblem)
$(deriveJSON myJSONOpts ''ActionPotentialTasklet)
$(deriveJSON myJSONOpts ''CoincidenceDetectorTask)
$(deriveJSON myJSONOpts ''TaskletTally)
$(deriveJSON myJSONOpts ''TrainingSet)
