{-# LANGUAGE
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
FlexibleContexts,
MultiParamTypeClasses,
FlexibleInstances,
ParallelListComp,
TypeFamilies,
DeriveDataTypeable,
OverloadedStrings,
TemplateHaskell,
DeriveGeneric,
UndecidableInstances
#-}

{-|

Module: Digiostelium.CopyProblem

Description : Problem for copying bits through a narrow synapse

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Here's the traditional set up I've been using, although the
problem is configurable in many ways:

There are two cells. Cell 0 is the sender and gets role input on
@A[1]@.  Cell 1 is the receiver and gets role input on @A[2]@.

Colonies are allowed a maximum number of time steps (normally
100).  The min time steps parameter is normally 20 and is used
only for scoring.  There is no bonus for finishing faster than
that.

The problem includes a beacon feature, in which points can be
awarded for emitting a beacon signal from the receiver.

Normal configuration:

Synapse channels: 3 --> 4, etc.

Output channels:
@
31 --> -2 (beacon)
0 --> -1 (stop signal)
12 --> 0
13 --> 1 etc.
@

Input channels:
@
0 --> 8
1 --> 9 etc.
@

All thresholds are set at 10 for channels.

Scoring: Each colony is presented with each possible input word.
One correctness point is given for each output bit that matches
the corresponding input bit.  The total correctness score is the
number of correctness points multiplied by the max time steps
(normally 100) and the correctness factor (normally 100) (for a
net factor of 10000).  The timing score is /max time steps minus
(actual time - min time steps)/, totaled over all input words.

For each input word, if the beacon signal is emitted, the score
includes 'beaconPoints' extra.  The default value of
'beaconPoints' is 0, which effectively disables this feature.

-}

module Digiostelium.CopyProblem (
  CopyProblem(..),
  CopyProblemSignal(..),
  isInputBit
  )
where

import GHC.Generics

import Control.DeepSeq

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.Default
import qualified Data.IntMap as IM
import Data.Semigroup
-- import qualified Data.Sequence as Seq
-- import Data.Sequence (Seq, ViewR(..), viewr)
-- import Text.Pandoc.Builder
-- import Text.Printf

import WGM.ShortShow

import BaselMachine.Arena

import Digiostelium.AesonUtil
import Digiostelium.CopyProblemBase
import Digiostelium.GeneralName
import Digiostelium.Genome
import Digiostelium.InputOutputProblem
import Digiostelium.PandocUtil
import Digiostelium.Problem
import Digiostelium.Report
import Digiostelium.SingletonProblem


-- | Colony consists of a sender and a receciver.
--
-- Accept input bit @j@ on @A[j+'inputWordIndexToPattern']@ in sender
--
-- Emit stop as -1 when @A[0] ≥ 'stopThreshold'@ in receiver
--
-- Emit beacon as -2 when @A[`beaconPattern`] ≥
-- `beaconThreshold`@ in receiver
--
-- Set output bit j when @A[j+'outputWordIndexToPattern'] ≥
-- 'outputThreshold'@ in receiver
--
-- Synapse connects sender patterns 'synapseInputIndexToPattern'...
-- to receciver patterns 'synapseOutputIndexToPattern'...

data CopyProblem = CopyProblem {
  wordSize :: Int,
  synapseSize :: Int,
  minSteps :: Int,
  maxSteps :: Int,
  correctnessFactor :: Double,
  inputWordIndexToPattern :: Int,
  -- ^ add this to an index into the input word to get the
  -- pattern in the sender that gets incremented
  inputActivation :: Int,
  -- ^ amount added to an input pattern in the sender when the
  -- corresponding bit in the input word is on
  synapseInputIndexToPattern :: Int,
  -- ^ add this to a bit index into the synaptic word to get the
  -- pattern in the sender that gets read
  synapseOutputIndexToPattern :: Int,
  -- ^ add this to a bit index into the synaptic word to get the
  -- pattern in the receiver that gets incremented
  synapseThreshold :: Int,
  -- ^ the synapse fires when the activation of a synapse pattern
  -- in the sender reaches this threshold
  synapseCost :: Int,
  -- ^ the activation subtracted when the synapse spikes
  synapseDelay :: Int,
  -- ^ how many time steps it takes for a spike to reach the
  -- receiver
  synapseActivation :: Int,
  -- ^ the activation added to a pattern in the receiver when the
  -- synapse spikes
  outputWordIndexToPattern :: Int,
  -- ^ add this to an index into the output word to get the
  -- pattern in the receiver that gets read
  outputThreshold :: Int,
  -- ^ an output bit is set on when the activation of an output
  -- pattern in the receiver reaches this threshold
  outputCost :: Int,
  -- ^ the activation subtracted when an output bit is set
  beaconPattern :: Int,
  -- ^ this pattern is read for the beacon signal
  beaconThreshold :: Int,
  -- ^ in the receiver cell, if @A[`beaconPattern`] ≥
  -- `beaconThreshold`@ then the beacon signal is emitted
  beaconCost :: Int,
  -- ^ amount to subtract from @A[`beaconPattern`]@ in the
  -- receiver cell when the beacon signal is emitted
  beaconPoints :: Double,
  -- ^ points awarded for emitting the beacon signal
  minReactions :: Double,
  -- ^ no cost for averaging up to this many reactions per time step
  reactionCostFactor :: Double,
  -- ^ an energy cost; multiply this by ((the number of reactions
  -- used divided by the number of time steps) less
  -- 'minReactions') and subtract that from the total score
  minTotalDeltas :: Double,
  -- ^ no cost for averaging up to this total of reaction deltas
  -- per time step
  deltaCostFactor :: Double,
  -- ^ an energy cost; multiply this by ((the total of all
  -- reaction deltas applied divided by the number of time steps)
  -- less 'minTotalDeltas') and subtract that from the total score
  roleActivation :: Int
  -- ^ how much activation is added to @A[1]@ in the sender and
  -- @A[2]@ in the receiver on each time step to indicate their
  -- roles.
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData CopyProblem
instance Binary CopyProblem
$(deriveJSON myJSONOpts ''CopyProblem)

-- | The default is the @Copy21@ problem with no beacon cost and
-- no energy costs.
instance Default CopyProblem where
  def = CopyProblem {
    wordSize = 2,
    synapseSize = 1,
    minSteps = 20,
    maxSteps = 100,
    correctnessFactor = 100,
    inputWordIndexToPattern = 8,
    inputActivation = 1,
    synapseInputIndexToPattern = 3,
    synapseOutputIndexToPattern = 4,
    synapseThreshold = 10,
    synapseCost = 0,
    synapseDelay = 0,
    synapseActivation = 1,
    outputWordIndexToPattern = 12,
    outputThreshold = 10,
    outputCost = 0,
    beaconPattern = 31,
    beaconThreshold = 10,
    beaconCost = 0,
    beaconPoints = 0,
    minReactions = 0,
    reactionCostFactor = 0,
    minTotalDeltas = 0,
    deltaCostFactor = 0,
    roleActivation = 1
    }

instance GeneralName CopyProblem where
  generalName _ = "Copy"

-- | Real number with semigroup operation of addition (<>) = (+)
type instance Rating CopyProblem = Sum Double

type instance Location CopyProblem = NullLocation

type instance Signal CopyProblem = CopyProblemSignal



instance ArenaProblem CopyProblem where
  stopSignal CopyProblem{..} =
    Just Stop

  minLandmarkRating _ = 0



instance InputOutputProblem CopyProblem where

  type instance Input CopyProblem = CopyProblemInput

  type instance Output CopyProblem = CopyProblemOutput


  problemInputToArenaInput CopyProblem{..} CopyProblemInput{..} =
    (maxSteps, [senderRole : receiverRole : convertAll inputBits])
    where
      senderRole = BroadcastState {
        broadcastSignal = SenderRole,
        broadcastMaximumAge = Nothing,
        broadcastFromColonyIx = Nothing,
        broadcastOrigin = NullLocation,
        broadcastAge = 0 }


      receiverRole = BroadcastState {
        broadcastSignal = ReceiverRole,
        broadcastMaximumAge = Nothing,
        broadcastFromColonyIx = Nothing,
        broadcastOrigin = NullLocation,
        broadcastAge = 0 }

      convert j = BroadcastState {
        broadcastSignal = InputBit j,
        broadcastMaximumAge = Nothing,
        broadcastFromColonyIx = Nothing,
        broadcastOrigin = NullLocation,
        broadcastAge = 0 }

      convertAll = map convert

  arenaOutputToProblemOutput CopyProblem{..} colonyOutputSeq =
    genericArenaOutputToCopyProblemOutput colonyOutputSeq

  inputsForRating CopyProblem{..} =
    allInputsFromWordSize wordSize

instance SingletonProblem CopyProblem where

  rateResult CopyProblem{..}
    CopyProblemInput{..}
    CopyProblemOutput{..}
    totalReactionCount
    totalReactionDeltas =
    let
      correctnessScore =
        countCorrectBits wordSize inputBits dataBits
      timeScore = maxSteps - (minSteps `max` numTimeSteps)
      beaconScore =
        if Beacon `elem` nonDataOutputs
        then beaconPoints
        else 0
      averageReactionCount =
        fromIntegral totalReactionCount / fromIntegral numTimeSteps
      averageTotalDeltas =
        fromIntegral totalReactionDeltas / fromIntegral numTimeSteps
      reactionCountCost =
        max 0 (reactionCostFactor * (averageReactionCount - minReactions))
      reactionDeltaCost =
        max 0 (deltaCostFactor * (averageTotalDeltas - minTotalDeltas))
    in
     Sum (correctnessFactor * fromIntegral (maxSteps * correctnessScore)
          + fromIntegral timeScore
          + beaconScore
          - reactionCountCost
          - reactionDeltaCost)


  initialColonyState CopyProblem{..} initProgram =
    ColonyInitialState{..}
    where
      initName = "Bob"
      initLocation = NullLocation

      colonyCellCount = 2

      colonyCellIxToName = IM.fromList [
        (0, "Sender"),
        (1, "Receiver")]

      colonySynapseSpecs = [
        ChannelSpec {
            channelSource = Source {
                sourceCellIx = 0,
                sourcePattern = synapseInputIndexToPattern + j,
                sourceThreshold = synapseThreshold,
                sourceActivationDelta = -synapseCost },
            channelTarget = Target {
                targetCellIx = 1,
                targetPattern = synapseOutputIndexToPattern + j,
                targetActivationDelta = synapseActivation },
            channelDelay = synapseDelay }
        | j <- [0 .. synapseSize - 1]]

      colonyBroadcastSpecs = [
        -- stop
        BroadcastSpec {
            broadcastSource = Source {
                sourceCellIx = 1,
                sourcePattern = 0,
                sourceThreshold = outputThreshold,
                sourceActivationDelta = -outputCost },
            broadcastMaximumAge = Just 1,
            broadcastSignal = Stop },
        -- beacon
        BroadcastSpec {
            broadcastSource = Source {
                sourceCellIx = 1,
                sourcePattern = beaconPattern,
                sourceThreshold = beaconThreshold,
                sourceActivationDelta = -beaconCost },
            broadcastMaximumAge = Just 1,
            broadcastSignal = Beacon }
        ] ++ [
        -- data bits
        BroadcastSpec {
            broadcastSource = Source {
                sourceCellIx = 1,
                sourcePattern = outputWordIndexToPattern + j,
                sourceThreshold = outputThreshold,
                sourceActivationDelta = -outputCost },
            broadcastMaximumAge = Just 1,
            broadcastSignal = OutputBit j }
        | j <- [0 .. wordSize - 1]]

      dataSensorSpecs = [
        SensorSpec {
            sensorSignal = InputBit j,
            sensorTarget = Target {
                targetCellIx = 0,
                targetPattern = inputWordIndexToPattern + j,
                targetActivationDelta = inputActivation } }
        | j <- [0 .. wordSize - 1]]

      senderRoleSpec = SensorSpec {
        sensorSignal = SenderRole,
        sensorTarget = Target {
            targetCellIx = 0,
            targetPattern = 1,
            targetActivationDelta = roleActivation } }

      receiverRoleSpec = SensorSpec {
        sensorSignal = ReceiverRole,
        sensorTarget = Target {
            targetCellIx = 1,
            targetPattern = 2,
            targetActivationDelta = roleActivation } }

      colonySensorSpecs = senderRoleSpec : receiverRoleSpec : dataSensorSpecs

      initColonySpec = ColonySpec{..}



instance RateEggs CopyProblem where

  rateEggs gs problem _adults eggs =
    singletonProblemRateEggs gs problem eggs

-- | Convert input to string of 0s and 1s.
instance Sniplet CopyProblem CopyProblemInput where
  toInlines CopyProblem{..} pInput =
    toInlines wordSize pInput

-- | Convert output to 0s and 1s, and how many steps.
instance Sniplet CopyProblem CopyProblemOutput where
  toInlines CopyProblem{..} output =
    toInlines wordSize output



-- === Reporting === --

instance
  (GenomeSpec genomeSpec,
   ShortShow (Interval genomeSpec),
   ToJSON (Genome genomeSpec) ) =>
  Reportable genomeSpec CopyProblem SingletonProblemReportSpec
