{-# LANGUAGE
GADTs,
TypeFamilies,
RecordWildCards,
NamedFieldPuns,
RankNTypes,
FlexibleInstances,
MultiParamTypeClasses,
FlexibleContexts,
UndecidableInstances,
ScopedTypeVariables,
DeriveDataTypeable,
DeriveGeneric,
TemplateHaskell
#-}

{-|

Module: Digiostelium.Evolution

Description : Abstractions for representing a deterministic-rating-based evolutionary problem

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)


This is for some general functions for dealing with the
selection-mutation process.

An interesting detail: The 'SelectFromBlocks' are processed in
order, and I generally list the most restrictive selection blocks
first (only affects the "weak" protocol so far).  That means
those agents have lower identifiers, and are considered older
than agents bred from less restrictive blocks.  That means the
high scoring children from the most restrictive block are at a
slight disadvantage as they are less likely to be among survivors
in future generations.  This might exaggerate the
innovation-via-outliers effect.

-}

module Digiostelium.Evolution
  {-
  (
  SelectFromBlock(..),
  SelectionSpec(..),
  EvaluationException(..),
  resultingPopulationSize,
  nextGeneration,
  rateGenome
  )
-}

where

import GHC.Generics

import Control.DeepSeq
-- import Control.Exception
import Control.Monad.IO.Class


-- import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.Default
import Data.Foldable
import Data.Map (Map)
import qualified Data.Map as M
-- import Data.Sequence (Seq, ViewL(..), ViewR(..))
import qualified Data.Sequence as Seq
import Data.Sequence (Seq)
import Data.Time.Clock.System
-- import Debug.Trace

-- import Text.Printf

import Digiostelium.AesonUtil
import Digiostelium.Agent
-- import BaselMachine.Instruction
-- import BaselMachine.Compiler
import Digiostelium.Genome
import Digiostelium.Logging
import Digiostelium.PopulationDatabase
import Digiostelium.Problem
import System.Random.RandomST
-- import WGM.Util


-- | Counts for how many agents from the top of the population
-- list to breed from, and how many offspring to generate.

data SelectFromBlock =
  UniformBlock {
    numOffspring :: !Int,
    -- ^ generate this many offspring
    numToBreed :: !Int
    -- ^ breed from this many of the top rated individuals
    }
  | Tournament {
    numOffspring :: !Int,
    -- ^ generate this many offspring
    pTakeHigher :: !Double
    -- ^ pick pairs and keep the higher-rated individual with this probability
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData SelectFromBlock
instance Binary SelectFromBlock
$(deriveJSON myJSONOpts ''SelectFromBlock)

data SelectionSpec =
  SelectionSpec {
    selectFromBlocks :: [SelectFromBlock],
    -- ^ breed according to these blocks
    numToKeep :: Int
    -- ^ keep this many top rated individuals from the original
    -- population
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

-- | This is weak selection
instance Default SelectionSpec where
  def = SelectionSpec{..}
    where
      selectFromBlocks =
        [UniformBlock {
            numOffspring = 200,
            numToBreed = 100},
         UniformBlock {
            numOffspring = 100,
            numToBreed = 500}]
      numToKeep = 200

instance NFData SelectionSpec
instance Binary SelectionSpec
$(deriveJSON myJSONOpts ''SelectionSpec)

resultingPopulationSize :: SelectionSpec -> Int
resultingPopulationSize SelectionSpec{..} =
  numToKeep + Prelude.sum (map numOffspring selectFromBlocks)

-- Agent.hs defines
-- Sample (gs, genome, genome) => (Sample (gs, Agent, Agent) Egg)

instance
  (Sample (gs, genome, genome) genome,
   Show rating) =>
  Sample
  (SelectFromBlock, gs, Seq (Agent genome rating))
  [Egg genome] where

  pullSample (UniformBlock {..}, gs, agentsSorted) =
    pullSampleList numOffspring (gs, Seq.take numToBreed agentsSorted)

  pullSample (Tournament {..}, gs, agentsSorted) =
    pullSampleList numOffspring (TournamentPair pTakeHigher, gs, agentsSorted)

data TournamentPair = TournamentPair Double

instance Sample (TournamentPair, Seq x) x where

  pullSample (TournamentPair p, xs) = do
    let n = Seq.length xs
    j1 <- pullSample (UniformDistribution (0, n - 1))
    j2 <- pullSample (UniformDistribution (0, n - 1))
    q <- pullSample (BernoulliDistribution p)
    let k = if q then min j1 j2 else max j1 j2
        -- j1 < j2 means rating a1 >= rating a2
    return (Seq.index xs k)

instance
  (Sample (gs, genome, genome) genome,
   Show rating) =>
  Sample
  (TournamentPair, gs, Seq (Agent genome rating))
  (Egg genome) where

    pullSample (t, gs, agentsSorted) = do
      a1 :: Agent genome rating <- pullSample (t, agentsSorted)
      a2 :: Agent genome rating <- pullSample (t, agentsSorted)
      e :: Egg genome <- pullSample (gs, a1, a2)
      return e



-- | Pull a sample of new genomes and such from a parent
-- population, using a deterministic genome rating function.

instance (
  GenomeSpec gs,
  genome ~ Genome gs,
  rating ~ Rating problem,
  Ord rating,
  Show rating,
  Eq genome,
  RateEggs problem,
  Sample (gs, Genome gs, Genome gs) (Genome gs)
  )
  =>
  Sample
  (SelectionSpec,
    gs,
    problem,
    Map Integer (Agent genome rating))
  (PopulationBirthDeathDelta genome rating) where

  pullSample (SelectionSpec {..},
              gs,
              problem,
              cachedLivingAgents) =
    let
      -- new improved version
      agentsSorted = Seq.sort $ Seq.fromList $ M.elems cachedLivingAgents
      toKill = reverse $ map identifier $ toList
               $ Seq.drop numToKeep agentsSorted

    in do
      newEggss <- (mapM
                   (\ sfb ->
                     pullSample (sfb, gs, agentsSorted))
                   selectFromBlocks)
      let newEggs = Prelude.concat newEggss
      newEggsRatings <- rateEggs gs problem cachedLivingAgents newEggs
      return PopulationBirthDeathDelta{..}


-- | This is the main selection-mutation process step.



nextGeneration ::
  (Sample
     (SelectionSpec, gs, problem, Map Integer (Agent genome rating))
     (PopulationBirthDeathDelta genome rating),
   Typeable rating, Typeable genome, Show rating, Show genome,
   Ord rating, NFData rating, NFData genome, Binary rating,
   Binary genome) =>
  SelectionSpec
  -> gs
  -> problem
  -> PopMonad genome rating ()

nextGeneration selectionSpec genomeSpec problem = do
  tStart <- liftIO getSystemTime
  msgDebug "Evolution.nextGeneration begin"
  cachedLivingAgents <- getLivingAgents
  bdDelta <- (withCurrentRandomSeed $
              pullSample (selectionSpec, genomeSpec, problem,
                          cachedLivingAgents))
  msgDebug "Evolution.nextGeneration birth-death-delta built"
  applyBirthDeathDelta bdDelta
  msgDebug "Evolution.nextGeneration birth-death-delta applied"
  nextTime
  msgDebug "Evolution.nextGeneration done"
  tEnd <- liftIO getSystemTime
  recordGenerationClockTime (systemTimeDifference tStart tEnd)


-- | (Private) Compute the different in seconds between two
-- 'SystemTime' objects.
systemTimeDifference ::
  SystemTime -- ^ start time
  -> SystemTime -- ^ end time
  -> Double

systemTimeDifference tStart tEnd =
  (fromIntegral (systemSeconds tEnd - systemSeconds tStart)
   +
   1e-9 * fromIntegral (systemNanoseconds tEnd - systemNanoseconds tStart))
