{-# LANGUAGE TypeFamilies,
RecordWildCards, NamedFieldPuns,
DeriveDataTypeable, ParallelListComp, FlexibleContexts,
OverloadedStrings #-}

{-|

Module: Digiostelium.GenomeDiagram

Description: Routines for drawing genome diagrams

Copyright: &#xA9; 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)


-}


module Digiostelium.GenomeDiagram (
  IntervalCoordinates(..),
  SVGDiagram,
  GeneStyle(..),
  genomeLayoutDiagram,
  diagramToByteString
  ) where

--import Data.Monoid

--import Debug.Trace

import qualified Data.ByteString.Lazy as BS
import Data.Data
import Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import Data.Typeable ()

import Diagrams.Prelude hiding (trace)
import Diagrams.Backend.SVG

--import Lucid.Svg (renderBS)
import Graphics.Svg (Element, renderBS)

import Digiostelium.Genome
-- import WGM.ShortShow


type SVGDiagram = Diagram SVG

-- | Tools for placing genomic information in a Eucliden plane.

class IntervalCoordinates interval where
  -- | Return (x1, x2, y) meaning that the interval corresponds
  --   to the line segment from (x1, y) to (x2, y).  Disjoint
  --   intervals should map to disjoint line segmenets, and the
  --   overall placement of intervals should reflect the
  --   underlying architecture of the genome.  These coordinates
  --   should place roughly one bit or letter per horizontal
  --   unit, and one strand per vertical unit.

  intervalSegmentCoordinates :: interval -> (Double, Double, Double)

data GeneStyle =
  Ordinary -- ^ small white oval, black outline
  | FlagRight -- ^ large red flag pointing to the right
  | FlagLeft -- ^ large blue flag pointing to the left
  | FlagBoth -- ^ large purple flag pointing left and right
    deriving (Read, Show, Eq, Ord, Enum, Data, Typeable)

drawGene :: GeneStyle -> Double -> Double -> SVGDiagram
drawGene Ordinary = ordinaryGene
drawGene FlagRight = flagRightGene
drawGene FlagLeft = flagLeftGene
drawGene FlagBoth = flagBothGene

ordinaryGene :: Double -> Double -> SVGDiagram
ordinaryGene w h =
  roundedRect w (h/2) (h/4)
  # lc black
  # lw 1
  # fc white

flagRightGene :: Double -> Double -> SVGDiagram
flagRightGene w h =
  (fromVertices
   (map p2
    [(0,0),
     (0.75*w,0),
     (w,0.5*h),
     (0.75*w,h),
     (0,h)]))
  # closeTrail
  # strokeTrail
  # centerXY
  # lc black
  # lw 1
  # fc red

flagLeftGene :: Double -> Double -> SVGDiagram
flagLeftGene w h =
  (fromVertices
   (map p2
    [(0,0),
     (0.75*w,0),
     (w,0.5*h),
     (0.75*w,h),
     (0,h)]))
  # closeTrail
  # strokeTrail
  # reflectX
  # centerXY
  # lc black
  # lw 1
  # fc blue

flagBothGene :: Double -> Double -> SVGDiagram
flagBothGene w h =
  (fromVertices
   (map p2
    [(0,0.5*h),
     (0.25*w,0),
     (0.75*w,0),
     (w,0.5*h),
     (0.75*w,h),
     (0.25*w,h)]))
  # closeTrail
  # strokeTrail
  # centerXY
  # lc black
  # lw 1
  # fc purple


genomeLayoutDiagram ::
  (GenomeSpec gs,
   IntervalCoordinates (Interval gs))
  =>
  gs
  -> Genome gs
  -> IntMap GeneStyle
  -> SVGDiagram

genomeLayoutDiagram gs genome geneStyles =
  let
    h = 40
    geneDiaglets :: [SVGDiagram]
    geneDiaglets = [
      -- trace (printf "x1: %f, x2: %f, y: %f" x1 x2 y) $
      (drawGene (IM.findWithDefault Ordinary j geneStyles) w h)
      # translateX (x1 + halfW)
      # translateY (1.5*h*y)
      | GeneDetails {..} <- geneDetails gs genome,
        let (x1, x2, y) = intervalSegmentCoordinates interval,
        let w = x2 - x1,
        let halfW = w/2
      | j <- [0..]]

    rawPicture = (mconcat geneDiaglets)
                 # centerXY

    result =
      (strut (h *^ unitY)
       ===
       (strut (h *^ unitX) ||| rawPicture ||| strut (h *^ unitX))
       ===
       strut (h *^ unitY))
      # scale 0.25

  in
   result

{- most generally:

diagramToByteString ::
  (RealFloat n,
   Show n,
   Typeable n,
   Monoid m,
   Semigroup m) =>
  QDiagram SVG V2 n m
  -> BS.ByteString

But for our purposes the more restrictive following type is fine:

-}


diagramToByteString ::
  SVGDiagram -> BS.ByteString

-- Note from diagrams 1.3:
-- type SvgM = Svg ()
-- supposedly a synonym for HtmlT m (), refering to Lucid.Base

diagramToByteString dg =
  let
    svgOpts :: Options SVG V2 Double
    svgOpts = SVGOptions absolute Nothing "" [] True
              -- width, SVG definitions, id prefix

    svgObj :: Element
    svgObj = renderDia SVG svgOpts dg
  in
    renderBS svgObj
