{-# LANGUAGE
TemplateHaskell,
DeriveGeneric,
DeriveDataTypeable,
StandaloneDeriving,
MultiParamTypeClasses,
UndecidableInstances,
FlexibleInstances,
GeneralizedNewtypeDeriving,
TypeFamilies,
FlexibleContexts,
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
OverloadedStrings
#-}

{-|

Module: Digiostelium.PairCopyProblemReport

Description : Data structures and functions for reports for 'PairCopyProblem'

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.PairCopyProblemReport (
  PairCopyProblemReportSpec(..),
  PairCopyProblemReport(..),
  pairEfficaciousAnalysis,
  pairDetailsReport
  ) where


import GHC.Generics

import Prelude hiding (id)

import Control.Exception
import Control.Lens hiding (para)

import Data.Aeson
import Data.Aeson.TH
import Data.ByteString.Builder (Builder)
import Data.Data
import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
import Data.IntSet (IntSet)
import qualified Data.IntSet as IntSet
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Maybe
import Data.Set (Set)
import Data.String

import Text.Pandoc.Builder
import Text.Printf

import Data.BitString

import BaselMachine.Arena
import BaselMachine.Compiler (ProgramInformation)
import BaselMachine.EfficaciousAnalysis
import BaselMachine.Instruction

import Digiostelium.AesonUtil
import Digiostelium.Agent
import Digiostelium.BasicReport (BasicReport)
import qualified Digiostelium.BasicReport as BasicReport
import Digiostelium.ColonyToStyledNetwork
import Digiostelium.Genome
import qualified Digiostelium.Genome as G
import Digiostelium.CopyProblemBase
import Digiostelium.PairCopyProblem
import Digiostelium.PandocUtil
import Digiostelium.Problem
import Digiostelium.Report
import Digiostelium.ReportException
import Digiostelium.StyledNetwork

import WGM.SeqUtil
import WGM.ShortShow

data PairCopyProblemReportSpec =
  PairRatingInputs
  | PairDetails {idOfInterest :: Integer}
  | PairReactionNetwork {idOfInterest :: Integer}
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''PairCopyProblemReportSpec)


data PairCopyProblemSubreport genome rating =
  PairRatingInputsSubreport {
    here :: BasicReport genome rating,
    there :: BasicReport genome rating,
    interactionResult :: InteractionResult
    }
  deriving (Read, Show, Eq, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''PairCopyProblemSubreport)


data PairCopyProblemReport genome rating =
  PairRatingInputsReport {
    ratingInputsSubreports :: [PairCopyProblemSubreport genome rating]
  }
  | PairDetailsReport {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    programInfo :: ProgramInformation,
    -- ^ instructions, reactions, and patterns referenced
    geneDescriptions ::  [String],
    instructionDetails :: [(Instruction, BitString, BitString, EncodingDetails)],
    efficaciousInstructions :: IntMap [(Int, Instruction)],
    -- ^ map from cell index to indexed efficacious instructions
    allEfficaciousInstructions :: [(Int, Instruction)],
    -- ^ all efficacious instructions
    efficaciousReactions :: IntMap (Set Reaction),
    -- ^ map from cell index to efficacious reactions
    allEfficaciousReactions :: Set Reaction,
    -- ^ all efficacious reactions
    efficaciousGeneDescriptions :: IntMap [String],
    -- ^ map from cell index to gene details of efficacious
    -- instructions
    allEfficaciousGeneDescriptions :: [String]
    }
  deriving (Read, Show, Eq, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''PairCopyProblemReport)



instance GetTitle opt (PairCopyProblemReport genome rating) where

  getTitle _ PairRatingInputsReport{..} =
    text "Ratings on training inputs"

  getTitle _ PairDetailsReport{..} =
    text "Full Details"

  getTitle _ _ =
    text "(Untitled)"



instance
  (GenomeSpec genomeSpec,
  ToJSON (Genome genomeSpec),
  ShortShow (Interval genomeSpec)) =>
  Reportable genomeSpec PairCopyProblem PairCopyProblemReportSpec where

  buildEnsembleReport = pairCopyProblemEnsembleReport

  -- | TODO
  buildManyReports =
    throw (UnsupportedReportType "buildManyReports: No multi-reports are implemented for PairCopyProblem")


interactionSubreportsFromAllInputs ::
  (GenomeSpec genomeSpec)
  =>
  genomeSpec
  -> PairCopyProblem
  -> [Agent (Genome genomeSpec) (Rating PairCopyProblem)]
  -> [PairCopyProblemSubreport (Genome genomeSpec) (Rating PairCopyProblem)]

interactionSubreportsFromAllInputs genomeSpec problem@PairCopyProblem{..} agents =
  [interactionSubreport problem here there problemInput
  | here <- compiledAgents,
    there <- compiledAgents,
    problemInput <- allInputsFromWordSize wordSize]
  where
    compiledAgents = compileAllToCache genomeSpec agents


interactionSubreport ::
  PairCopyProblem
  -> CompilationCache (Agent genome (Rating PairCopyProblem))
  -> CompilationCache (Agent genome (Rating PairCopyProblem))
  -> CopyProblemInput
  -> PairCopyProblemSubreport genome (Rating PairCopyProblem)

interactionSubreport problem here there problemInput =
 PairRatingInputsSubreport {
  here = BasicReport.MinimalAgentReport {
      id = identifier (here ^. thingWithGenome),
      rating = hereRating },
  there = BasicReport.MinimalAgentReport {
      id = identifier (there ^. thingWithGenome),
      rating = thereRating },
    ..}
  where
    interactionResult = runPairInteraction problem here there problemInput
    (hereRating, thereRating) = rateInteractionResult problem interactionResult


-- | Compute the 'EfficaciousAnalysis' resulting from what
-- happens when an 'Agent' of interest interacts with a list of
-- other 'Agent's.

pairEfficaciousAnalysis ::
  PairCopyProblem
  -> CompilationCache thingWithGenome -- ^ 'Agent' of interest
  -> [CompilationCache thingWithGenome] -- ^ other 'Agent's to interact with
  -> EfficaciousAnalysis

pairEfficaciousAnalysis problem a others =
  jointEffAn (pairEfficaciousAnalyses problem a others)


-- | (Private) Compute efficacious analysis for an agent of
-- interest, when communicating with the given list of other
-- agents.  Since an agent does different things when in the
-- 'Here' or 'There' location, both joint and location-specific
-- analyses are conducted.

data PairEfficaciousAnalyses problem =
  PairEfficaciousAnalyses {
    jointEffAn :: EfficaciousAnalysis,
    hereEffAn :: EfficaciousAnalysis,
    hereColonySpec :: ColonySpec problem,
    thereEffAn :: EfficaciousAnalysis,
    thereColonySpec :: ColonySpec problem
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

pairEfficaciousAnalyses ::
  PairCopyProblem
  -> CompilationCache thingWithGenome
  -> [CompilationCache thingWithGenome]
  -> PairEfficaciousAnalyses CopyProblemSignal

pairEfficaciousAnalyses problem a others =
  PairEfficaciousAnalyses{..}
  where
    hereTraceResultsPlus = concatMap (tracePairOnAll problem a) others
    thereTraceResultsPlus = concatMap (\other -> tracePairOnAll problem other a) others

    hereTraceResults =
      [ trColonyIxToSummary IntMap.! 0
      | (_input, _output, ArenaTraceSummary{..}) <- hereTraceResultsPlus]
    thereTraceResults =
      [ trColonyIxToSummary IntMap.! 1
      | (_input, _output, ArenaTraceSummary{..}) <- thereTraceResultsPlus]

    jointTraceResults = hereTraceResults ++ thereTraceResults

    -- Since the celluar structure is static, the colony spec is
    -- static, so just pick out the first one.
    hereColonySpec =
        colonySpec
        $ fromJust -- had better be there
        $ IntMap.lookup 0 -- colony of interest is Here, so at index 0
        $ arenaColonyIxToState -- map to ColonyState
        $ fromJust -- had better be there
        $ Map.lookup AfterEmitters -- ArenaState at the end of the time step
        $ seqLast -- just the last time step
        $ trArenaSteps
        $ (\(_input, _output, arenaTraceSummary) -> arenaTraceSummary) -- fish out ArenaTraceSummary
        $ head -- just case of first input
        $ hereTraceResultsPlus -- list of ArenaTraceSummary, one entry for each input

    thereColonySpec =
        colonySpec
        $ fromJust -- had better be there
        $ IntMap.lookup 1 -- colony of interest is There, so at index 0
        $ arenaColonyIxToState -- map to ColonyState
        $ fromJust -- had better be there
        $ Map.lookup AfterEmitters -- ArenaState at the end of the time step
        $ seqLast -- just the last time step
        $ trArenaSteps
        $ (\(_input, _output, arenaTraceSummary) -> arenaTraceSummary) -- fish out ArenaTraceSummary
        $ head -- just case of first input
        $ thereTraceResultsPlus -- list of ArenaTraceSummary, one entry for each input

    prog = fst $ a ^. progAndInfo

    hereEffAn = efficaciousAnalysis prog hereTraceResults
    thereEffAn = efficaciousAnalysis prog thereTraceResults
    jointEffAn = efficaciousAnalysis prog jointTraceResults


pairDetailsReport ::
  (GenomeSpec genomeSpec,
   ShortShow (Interval genomeSpec))
  => genomeSpec
  -> PairCopyProblem
  -> CompilationCache (Agent (Genome genomeSpec) rating)
  -> [CompilationCache (Agent (Genome genomeSpec) rating)]
  -> PairCopyProblemReport (Genome genomeSpec) rating

pairDetailsReport genomeSpec problem a others =
  PairDetailsReport{..}
  where
    Agent{..} = a ^. thingWithGenome
    Egg{..} = egg
    id = identifier
    genomeLength = G.genomeLength genomeSpec genome
    geneDescriptions = map shortShow geneDetails
    (programInfo, _) = a ^. progAndInfo

    instructionMap = IntMap.fromList (zip [0..] (a ^. instructions))
    instructionDetails =
      [ (instruction, codonBits, nucleotideBits, instructionEncodingDetails)
      | G.GeneDetails{..} <- geneDetails]

    geneDetails = G.geneDetails genomeSpec genome
    geneDetailsMap = IntMap.fromList (zip [0..] geneDetails)

    EfficaciousAnalysis{
      perCellEfficacious,
      allEfficaciousInstructionIxs,
      allEfficaciousReactions
      } = pairEfficaciousAnalysis problem a others

    listInstructions :: IntSet -> [(Int, Instruction)]
    listInstructions ixSet =
      map (\j -> (j, instructionMap IntMap.! j)) (IntSet.toList ixSet)

    listGeneDescriptions :: IntSet -> [String]
    listGeneDescriptions ixSet =
      (map
       (\j -> shortShow $ geneDetailsMap IntMap.! j)
       (IntSet.toList ixSet))

    efficaciousReactions =
      IntMap.map cellEfficaciousReactionSet perCellEfficacious
    efficaciousInstructions =
      IntMap.map (listInstructions . cellEfficaciousInstructionIxs) perCellEfficacious
    efficaciousGeneDescriptions =
      IntMap.map (listGeneDescriptions . cellEfficaciousInstructionIxs) perCellEfficacious
    allEfficaciousInstructions =
      listInstructions allEfficaciousInstructionIxs
    allEfficaciousGeneDescriptions =
      listGeneDescriptions allEfficaciousInstructionIxs


pairCopyProblemEnsembleReport ::
  (GenomeSpec genomeSpec,
  ToJSON (Genome genomeSpec),
  ShortShow (Interval genomeSpec))
  => genomeSpec
  -> PairCopyProblem
  -> PairCopyProblemReportSpec
  -> OutputFormat
  -> Map Integer (Agent (Genome genomeSpec) (Rating PairCopyProblem))
  -> Builder

pairCopyProblemEnsembleReport genomeSpec problem PairDetails{..} outputFormat aMap =
  case outputFormat of
    PlainTextOutput ->
      buildPandocPlainNoWrap $ pairReportPandoc problem report
    MarkdownOutput ->
      buildPandocMarkdown $ pairReportPandoc problem report
    _ ->
      buildJSONOrYAML outputFormat report
  where
    a = compileToCache genomeSpec $ aMap Map.! idOfInterest
    others = compileAllToCache genomeSpec $ Map.elems $ Map.delete idOfInterest aMap
    report = pairDetailsReport genomeSpec problem a (a:others)

pairCopyProblemEnsembleReport genomeSpec problem PairRatingInputs outputFormat aMap =
  case outputFormat of
    PlainTextOutput ->
      buildPandocPlainNoWrap $ pairReportPandoc problem report
    MarkdownOutput ->
      buildPandocMarkdown $ pairReportPandoc problem report
    _ ->
      buildJSONOrYAML outputFormat report
  where
    report = PairRatingInputsReport {
      ratingInputsSubreports = interactionSubreportsFromAllInputs genomeSpec problem (Map.elems aMap)
      }

pairCopyProblemEnsembleReport genomeSpec problem PairReactionNetwork{..} GraphMLOutput aMap =
  buildXMLLight graphML
  where
    a = compileToCache genomeSpec $ aMap Map.! idOfInterest
    others = compileAllToCache genomeSpec $ Map.elems $ Map.delete idOfInterest aMap
    network = pairReactionNetworkReport problem a others
    graphML = networkToGraphMLDocument network

pairCopyProblemEnsembleReport _ _ reportSpec outputFormat _ =
  throw (UnsupportedFormat
         (printf "pairCopyProblemEnsembleReport: Report spec '%s' with output format '%s' is not defined"
          (show reportSpec)
          (show outputFormat)))


pairReportPandoc ::
  Show rating
  => PairCopyProblem
  -> PairCopyProblemReport genome rating
  -> Pandoc

pairReportPandoc problem report@PairRatingInputsReport{..} =
  setTitle (getTitle () report)
  $ doc
  $ mconcat
  $ map fancySubreport (zip [(0 :: Int) ..] ratingInputsSubreports)
  where

    fancySubreport (j, PairRatingInputsSubreport{..}) =
      ((header 1 $ text $ fromString $ printf "Here: %d  There: %d" (BasicReport.id here) (BasicReport.id there))
        <>
        (header 2 ("Interaction #" <> space <> texts j <> ":"))
        <>
        (para
         (
          "Input    :" <> space <> toInlines problem problemInput <> linebreak <>
          "Output   :" <> space <> toInlines problem problemOutput <> linebreak <>
          "Hr Rxn ct:" <> space <> texts hereTotalReactionCount <> linebreak <>
          "Hr Tot dl:" <> space <> texts hereTotalReactionDeltas <> linebreak <>
          "Th Rxn ct:" <> space <> texts thereTotalReactionCount <> linebreak <>
          "Th Tot dl:" <> space <> texts thereTotalReactionDeltas <> linebreak <>
          "Hr rating:" <> space <> texts (BasicReport.rating here) <> linebreak <>
          "Th rating:" <> space <> texts (BasicReport.rating there)))
        )
      where
        InteractionResult{..} = interactionResult




pairReportPandoc _problem report@PairDetailsReport{..} =
  setTitle (getTitle () report)
  $ doc
  $ ((header 1 $ text $ fromString $ printf "Agent %d" id)
  <>
  (para
    ( "Rating:" <> space <> texts rating <> linebreak <>
      "Lifespan:" <> space <> texts lifespan <> linebreak <>
      "Parents:" <> space <> texts parentIds <> linebreak <>
      "Genome length:" <> space <> texts genomeLength <> linebreak))
  <>
  cellReports)
  where
    cellReports =
      (mconcat
        (map cellReport (IntMap.keys efficaciousReactions)))

    cellReport cellIx =
      (header 2 $ text $ fromString $ printf "Cell %d" cellIx)
      <>
      (header 3 "Efficacious reactions")
      <>
      reactionBlock (efficaciousReactions IntMap.! cellIx)
      <>
      (header 3 "Efficacious instructions")
      <>
      instructionBlock (efficaciousInstructions IntMap.! cellIx)



pairReportPandoc _ _ =
  throw (UnsupportedFormat "pairReportPandoc: This report cannot be shown as Pandoc Markdown")



pairReactionNetworkReport ::
  PairCopyProblem
  -> CompilationCache (Agent genome rating)
  -> [CompilationCache (Agent genome rating)]
  -> StyledNetwork

pairReactionNetworkReport problem a others =
  StyledNetwork {
    networkKey = "Arena",
    networkLabel = "Arena",
    nodes = [],
    edges = [],
    subnetworks = [hereReactionNetwork, thereReactionNetwork] }
  where
    Agent{egg=Egg{..}, ..} = a ^. thingWithGenome

    PairEfficaciousAnalyses{..} = pairEfficaciousAnalyses problem a others

    hereReactionNetwork =
      (efficaciousAnalysisToStyledNetwork
       hereEffAn
       (printf "Agent %d" identifier)
       hereColonySpec)

    thereReactionNetwork =
      (efficaciousAnalysisToStyledNetwork
       thereEffAn
       (printf "Agent %d" identifier)
       thereColonySpec)
