{-# LANGUAGE
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
DeriveDataTypeable,
StandaloneDeriving,
FlexibleContexts,
FlexibleInstances,
GADTs,
ExistentialQuantification,
NoMonomorphismRestriction,
DeriveGeneric,
TemplateHaskell,
ImpredicativeTypes,
RankNTypes,
UndecidableInstances,
ConstraintKinds,
MultiParamTypeClasses,
TypeInType,
TypeFamilies,
QuantifiedConstraints
#-}

{-|
Module      : Digiostelium.ProblemInventory

Description : Collection of all available Problems and other configuration

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.ProblemInventory where

import GHC.Generics

import Control.DeepSeq
import Control.Exception

import Data.Aeson hiding (Array)
import Data.Aeson.TH
import Data.Binary (Binary)
import Data.Data
import Data.Default
import Data.Kind
import Data.Map (Map)
import qualified Data.Map as M
import Data.Typeable()

import System.Random.RandomST

import BaselMachine.Signal

import Digiostelium.Agent (Agent(..))
import Digiostelium.BasicReport
import Digiostelium.BlockSize
import Digiostelium.Chromosome
import Digiostelium.GeneralName
import Digiostelium.Genome
import Digiostelium.Problem
import Digiostelium.SingletonProblem
import Digiostelium.Report
import Digiostelium.ReportException

import WGM.ShortShow

-- Genomes
import qualified Digiostelium.HaploidSingletonGenome as Hap
import qualified Digiostelium.PolyploidGenome as Poly

-- Problems
import Digiostelium.CopyProblem
import Digiostelium.CoincidenceDetectorProblem
import Digiostelium.PairCopyProblem
import Digiostelium.PairCopyProblemReport

-- | I've been wanting this for a long time: Box up an object and
-- a class implementation.
data Box (c :: Type -> Constraint) =
  forall t. c t => Box t

data Impl (c :: Type -> Constraint) t = Impl

data ImplWrap (c :: Type -> Constraint) =
  forall t. c t => ImplWrap (Impl c t)

withNamedObject ::
  Map String (ImplWrap c)
  -> (forall t. c t => Impl c t -> result)
  -> result
  -> String
  -> result
withNamedObject inventory axn ifUnknown name =
  case M.lookup name inventory of
    Just (ImplWrap impl) ->
      axn impl
    Nothing ->
      ifUnknown

--- end of weirdness ---

-- | All the type classes required for a problem.  Note that
-- using a type synonym doesn't work, because the compiler
-- insists all applications of synonyms must be fully saturated,
-- but it doesn't require that for classes. *shrug*
--
-- The constraint FromJSON (Output problem) is needed because we
-- have to export genomes and then import them again, and if I
-- don't tie up all the type variables, the compiler doesn't know
-- how to handle imports using 'QueryResult.AgentResult'.

class
  (Typeable problem,
   ArenaProblem problem,
   RateEggs problem,
   ToJSON problem,
   FromJSON problem,
   GeneralName problem,
   Show problem,
   Semigroup (Rating problem),
   Binary (Rating problem),
   FromJSON (Rating problem),
   NFData (Rating problem),
   Ord (Rating problem),
   Read (Rating problem),
   Show (Rating problem),
   ToJSON (Rating problem),
   Typeable (Rating problem),
   BroadcastSignal (Location problem) (Signal problem),
   Show (Signal problem),
   Typeable (Signal problem),
   Ord (Signal problem),
   NFData (Signal problem),
   NFData (Location problem),
   BoxReportSpec problem
  )
  => UsefulProblem problem

instance UsefulProblem CopyProblem
instance UsefulProblem CoincidenceDetectorProblem
instance UsefulProblem PairCopyProblem

problemTemplates :: Map String (Box UsefulProblem)
problemTemplates =
  M.fromList   [
    -- Copy (within an individual)
    ("Copy", Box (def :: CopyProblem)),
    -- Copy (communication between pairs)
    ("PairCopy", Box (def :: PairCopyProblem)),
    -- Coincidence detectors
    ("CDMedium", Box problemTemplateCDMedium),
    ("CDLong", Box problemTemplateCDLong),
    ("CDLearnMedium", Box problemTemplateCDLearnMedium),
    ("CDLearnLong", Box problemTemplateCDLearnLong),
    ("CDLearnLongEqWeight", Box problemTemplateCDLearnLongEqWeight),
    ("CDLearnLongDifficultyWeight", Box problemTemplateCDLearnLongDifficultyWeight)
  ]


problemInventory :: Map String (ImplWrap UsefulProblem)
problemInventory =
  M.fromList $ map addProblemName [
    ImplWrap (Impl :: Impl UsefulProblem CopyProblem),
    ImplWrap (Impl :: Impl UsefulProblem CoincidenceDetectorProblem),
    ImplWrap (Impl :: Impl UsefulProblem PairCopyProblem)
    ]

problemName :: ImplWrap UsefulProblem -> String
problemName (ImplWrap (Impl :: Impl UsefulProblem t)) =
      generalName (undefined :: t)

addProblemName :: ImplWrap UsefulProblem -> (String, ImplWrap UsefulProblem)
addProblemName pw = (problemName pw, pw)

----------

class
  (GenomeSpec gs,
   Typeable gs,
   Show gs,
   ToJSON gs,
   FromJSON gs,
   GeneralName gs,
   BlockSize gs,
   Show (Genome gs),
   Typeable (Genome gs),
   NFData (Genome gs),
   Binary (Genome gs),
   ToJSON (Genome gs),
   FromJSON (Genome gs),
   Ord (Genome gs),
   Sample gs (Genome gs),
   Sample (gs, Genome gs, Genome gs) (Genome gs),
   ToJSON (AlignmentSpec gs),
   FromJSON (AlignmentSpec gs),
   Default (AlignmentSpec gs),
   BlockSize (AlignmentSpec gs),
   ShortShow (Interval gs)) =>
  UsefulGenomeSpec gs

instance UsefulGenomeSpec Hap.HaploidSingletonGenomeSpec
instance UsefulGenomeSpec Poly.PolyploidGenomeSpec

genomeArchitectureInventory :: Map String (ImplWrap UsefulGenomeSpec)
genomeArchitectureInventory =
  M.fromList[
    ("Haploid",
     ImplWrap (Impl :: Impl UsefulGenomeSpec Hap.HaploidSingletonGenomeSpec)),
    ("Polyploid",
     ImplWrap (Impl :: Impl UsefulGenomeSpec Poly.PolyploidGenomeSpec))
    ]

genomeSpecTemplates :: Map String (Box UsefulGenomeSpec)
genomeSpecTemplates =
  M.fromList [
    ("haploid-32-aligned",
     Box (def :: Hap.HaploidSingletonGenomeSpec)),

    ("haploid-32-length-neutral",
      (Box
       (setCrossoverType SinglePointLengthNeutral (def :: Hap.HaploidSingletonGenomeSpec) ))),

    ("haploid-32-aligned-short",
      Box shortHaploidSpec),
    ("haploid-32-length-neutral-short",
      (Box
       (setCrossoverType SinglePointLengthNeutral shortHaploidSpec))),

    ("diploid-2-16-aligned",
     Box (def :: Poly.PolyploidGenomeSpec))
    ]


-- | 'InstructionSpec' like the default but with only 4 pattern
-- bits
shortInstructionSpec :: InstructionSpec
shortInstructionSpec =
  def {
    patternSpec = ReflectedGrayEncoding 4
    }

shortHaploidSpec :: Hap.HaploidSingletonGenomeSpec
shortHaploidSpec =
  Hap.makeDefaultHaploidSingletonGenomeSpec 3 32 shortInstructionSpec


-- === Reporting === --

data ReportBox where
  ReportBox ::
    forall genomeSpec problem reportSpec.
    (Reportable genomeSpec problem reportSpec) =>
    genomeSpec
    -> problem
    -> reportSpec
    -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
    -> ReportBox

writeReportsBox ::
  ReportBox
  -> OutputFormat
  -> OutputDestination
  -> IO ()

writeReportsBox rBox outputFormat outputDestination =
  case rBox of
    ReportBox genomeSpec problem reportSpec aMap ->
      writeReports genomeSpec problem reportSpec outputFormat outputDestination aMap


data PossibleReportSpec =
  Basic BasicReportSpec
  | Singleton SingletonProblemReportSpec
  | PairCopy PairCopyProblemReportSpec
  deriving (Read, Show, Eq, Ord, Data, Generic, Typeable)



class (Show (Signal problem),
       ToJSON (Signal problem),
       Show (Rating problem),
       ToJSON (Rating problem)
       ) =>
  BoxReportSpec problem where

  boxReportSpec ::
    forall genomeSpec.
    (GenomeSpec genomeSpec,
     ToJSON (Genome genomeSpec),
     ShortShow (Interval genomeSpec))
    =>
    genomeSpec
    -> problem
    -> PossibleReportSpec
    -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
    -> Maybe ReportBox

  -- | Default implementation is that only basic reports are possible
  boxReportSpec genomeSpec problem (Basic reportSpec) agents =
    Just (ReportBox genomeSpec problem reportSpec agents)
  boxReportSpec _genomeSpec _problem _reportSpec _agents =
    Nothing

-- | Given a 'GenomeSpec', a problem, a 'PossibleReportSpec', and
-- a population: If the report is sensible in that context, box it up.
-- Otherwise return 'Nothing'.

boxBasicOrSingletonReportSpec ::
  (ShortShow (Interval genomeSpec),
   ToJSON (Genome genomeSpec),
   ToJSON (Rating problem),
   Show (Signal problem),
   Show (Rating problem),
   Reportable genomeSpec problem SingletonProblemReportSpec)
  => genomeSpec
  -> problem
  -> PossibleReportSpec
  -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
  -> Maybe ReportBox

boxBasicOrSingletonReportSpec genomeSpec problem wrappedReportSpec aMap =
  case wrappedReportSpec of
    Basic rSpec ->  Just (ReportBox genomeSpec problem rSpec aMap)
    Singleton rSpec -> Just (ReportBox genomeSpec problem rSpec aMap)
    _ -> throw (UnsupportedReportType "This problem type only supports basic and singleton reports")

boxBasicOrPairCopy ::
  (ShortShow (Interval genomeSpec),
   ToJSON (Genome genomeSpec),
   ToJSON (Rating problem),
   Show (Signal problem),
   Show (Rating problem),
   Reportable genomeSpec problem PairCopyProblemReportSpec)
  => genomeSpec
  -> problem
  -> PossibleReportSpec
  -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
  -> Maybe ReportBox

boxBasicOrPairCopy genomeSpec problem wrappedReportSpec aMap =
  case wrappedReportSpec of
    Basic rSpec ->  Just (ReportBox genomeSpec problem rSpec aMap)
    PairCopy rSpec -> Just (ReportBox genomeSpec problem rSpec aMap)
    _ -> throw (UnsupportedReportType "This problem type only supports basic and pair Copy() reports")


instance BoxReportSpec CopyProblem where
  boxReportSpec = boxBasicOrSingletonReportSpec

instance BoxReportSpec CoincidenceDetectorProblem where
  boxReportSpec = boxBasicOrSingletonReportSpec

instance BoxReportSpec PairCopyProblem where
  boxReportSpec = boxBasicOrPairCopy

$(deriveJSON defaultOptions ''PossibleReportSpec)
