{-# LANGUAGE
FlexibleInstances,
MultiParamTypeClasses,
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
DeriveDataTypeable,
DeriveGeneric,
TemplateHaskell,
FlexibleContexts,
TypeFamilies,
OverloadedStrings,
UndecidableInstances,
QuantifiedConstraints,
DefaultSignatures,
GADTs
#-}

{-|

Module: Digiostelium.Report

Description : Creating reports from data extracted from 'Agent's

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.Report (
  Reportable(..),
  OutputFormat(..),
  OutputDestination(..),
  writeReports,
  buildJSONOrYAML
  ) where

import GHC.Generics

import Control.Exception
import Control.Monad

import Data.Aeson (ToJSON(..))
import Data.Aeson.TH
import Data.ByteString.Builder
import Data.Data
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Typeable ()

import System.Directory
import System.FilePath
import System.IO

import Text.Printf

import Digiostelium.AesonUtil
import Digiostelium.Agent (Agent)
import Digiostelium.Genome
import Digiostelium.GraphML
import Digiostelium.Logging
import Digiostelium.PandocUtil
import Digiostelium.Problem
import Digiostelium.Query
import Digiostelium.ReportException

-- | Ways of writing out information
data OutputFormat =
  JSONOutput
  | YAMLOutput
  | PlainTextOutput
  | MarkdownOutput
  | GraphMLOutput
    deriving (Read, Show, Typeable, Data, Generic)

$(deriveJSON myJSONOpts ''OutputFormat)

-- | Where to send the output
data OutputDestination =
  AllToStdout
  | AllToFile FilePath
  | ToFilePerAgentId Bool FilePath
    -- ^ must include %d for where to put the agent's identifier
    deriving (Read, Show, Typeable, Data, Generic)

$(deriveJSON myJSONOpts ''OutputDestination)



-- | Most general class for building reports that can be rendered to text as a 'Builder'.
-- These functions take a 'Map' from identifiers to agents.
-- The rationale for that type signature is that some reports only make sense if each agent is evaluated in the context of some population.
-- There are two functions, one that produces a single compound builder for one big document, and one that produces a container with one report for each agent in the given 'Map'.
-- The secone one is used when the output is supposed to go in individual files, one for each agent.
-- Since there's no way for the type system to fully represent each possible valid combination of report specification parameters, these functions are expected to throw an exception if a request is not sensible.
-- The default implementations are for reports that involve isolated 'Agent's
class
  (GenomeSpec genomeSpec)
  =>
  Reportable genomeSpec problem reportSpec where

  -- | Build a single report from all the agents.
  -- May throw 'ReportException.ReportException'
  buildEnsembleReport ::
    genomeSpec
    -> problem
    -> reportSpec
    -> OutputFormat
    -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
    -> Builder

  default buildEnsembleReport ::
    (QueryIndividual genomeSpec problem reportSpec,
     ToPandoc (genomeSpec, problem) (QueryResult genomeSpec problem reportSpec),
     ToPandoc (genomeSpec, problem) (Map Integer (QueryResult genomeSpec problem reportSpec)),
     ToJSON (QueryResult genomeSpec problem reportSpec)) =>
    genomeSpec
    -> problem
    -> reportSpec
    -> OutputFormat
    -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
    -> Builder

  buildEnsembleReport gs prob reportSpec JSONOutput agentMap =
    buildJSON $ fmap (queryIndividual gs prob reportSpec) agentMap

  buildEnsembleReport gs prob reportSpec YAMLOutput agentMap =
    buildYAML $ fmap (queryIndividual gs prob reportSpec) agentMap

  buildEnsembleReport gs prob reportSpec PlainTextOutput agentMap =
    buildPandocPlainNoWrap
    $ toPandoc (gs, prob)
    $ fmap (queryIndividual gs prob reportSpec) agentMap

  buildEnsembleReport gs prob reportSpec MarkdownOutput agentMap =
    buildPandocMarkdown
    $ toPandoc (gs, prob)
    $ fmap (queryIndividual gs prob reportSpec) agentMap

  buildEnsembleReport _gs _prob _reportSpec GraphMLOutput _agentMap =
    throw (UnsupportedFormat "GraphML output is not supported for this ensemble report.  Try a different format or use --path-template to make an individual report file for each agent.")

  -- | Build a container of reports, one for each agent.
  -- May throw 'ReportException.ReportException'
  buildManyReports ::
    genomeSpec
    -> problem
    -> reportSpec
    -> OutputFormat
    -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
    -> Map Integer Builder

  default buildManyReports ::
    (QueryIndividual genomeSpec problem reportSpec,
     ToPandoc (genomeSpec, problem) (QueryResult genomeSpec problem reportSpec),
     ToGraphML (genomeSpec, problem) (QueryResult genomeSpec problem reportSpec),
     ToJSON (QueryResult genomeSpec problem reportSpec)) =>
    genomeSpec
    -> problem
    -> reportSpec
    -> OutputFormat
    -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
    -> Map Integer Builder

  buildManyReports gs prob reportSpec JSONOutput agentMap =
    fmap (buildJSON . queryIndividual gs prob reportSpec) agentMap

  buildManyReports gs prob reportSpec YAMLOutput agentMap =
    fmap (buildYAML . queryIndividual gs prob reportSpec) agentMap

  buildManyReports gs prob reportSpec PlainTextOutput agentMap =
    fmap (buildPandocPlainNoWrap . toPandoc (gs, prob) . queryIndividual gs prob reportSpec) agentMap

  buildManyReports gs prob reportSpec MarkdownOutput agentMap =
    fmap (buildPandocMarkdown . toPandoc (gs, prob) . queryIndividual gs prob reportSpec) agentMap

  buildManyReports gs prob reportSpec GraphMLOutput agentMap =
    fmap (buildXMLLight . toGraphML (gs, prob) . queryIndividual gs prob reportSpec) agentMap




writeReports ::
  Reportable genomeSpec problem reportSpec =>
  genomeSpec
  -> problem
  -> reportSpec
  -> OutputFormat
  -> OutputDestination
  -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
  -> IO ()

writeReports
  genomeSpec problem reportSpec outputFormat
  AllToStdout
  agentMap =
  do
    hPutBuilder stdout builder
    hPutStr stdout "\n"
      where
        builder = buildEnsembleReport genomeSpec problem reportSpec outputFormat agentMap


writeReports
  genomeSpec problem reportSpec outputFormat
  (AllToFile outputPath)
  agentMap =
  do
    ensureDirectoryForFile outputPath
    let builder = buildEnsembleReport genomeSpec problem reportSpec outputFormat agentMap
    withBinaryFile outputPath WriteMode $ \h -> do
      hPutBuilder h builder
      hPutStr h "\n"

writeReports
  genomeSpec problem reportSpec outputFormat
  (ToFilePerAgentId overwrite pathTemplate)
  agentMap =
  let
    idToReportBuilder =
      buildManyReports genomeSpec problem reportSpec outputFormat agentMap
  in do
    forM_ (Map.toList idToReportBuilder) $ \(n, reportBuilder) -> do
      let
        outputPath =  printf pathTemplate n
        in do
        fileExists <- doesFileExist outputPath
        if not fileExists || overwrite
          then do
          msgInfo (printf "Writing report to %s" outputPath)
          ensureDirectoryForFile outputPath
          withBinaryFile outputPath WriteMode $ \h -> do
            hSetBuffering h (BlockBuffering Nothing)
            hPutBuilder h reportBuilder
          else do
          throwIO (CantOverwriteFile
                 (printf "File '%s' exists, won't overwrite" outputPath))


-- | Given a path to a file, create the directory it would be in
-- if necessary.
ensureDirectoryForFile :: FilePath -> IO ()
ensureDirectoryForFile filePath =
  createDirectoryIfMissing True (dropFileName filePath)



-- | For reports that can be output in JSON and YAML, this is a
-- helper function for quickly implementing 'Reportable'

buildJSONOrYAML :: ToJSON t => OutputFormat -> t -> Builder

buildJSONOrYAML JSONOutput t =
  buildJSON t

buildJSONOrYAML YAMLOutput t =
  buildYAML t

buildJSONOrYAML _ _ =
  throw (UnsupportedFormat "buildJSONOrYAML: Can only produce JSON or YAML for this report type")
