{-# LANGUAGE
TemplateHaskell,
DeriveGeneric,
DeriveDataTypeable,
StandaloneDeriving,
MultiParamTypeClasses,
UndecidableInstances,
FlexibleInstances,
GeneralizedNewtypeDeriving,
TypeFamilies,
FlexibleContexts,
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
OverloadedStrings
#-}

{-|

Module: Digiostelium.SingletonProblem

Description : Abstractions for representing a problem to be solved by an agent in isolation.

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Foundations for problems in which an agent is evaluated entirely
in isolation, no other agents involved.

The efficacious analysis report also assumes the cellular
structure is static, that is, no cells or synapses are modified
during a colony's lifetime.

-}

module Digiostelium.SingletonProblem (
  SingletonProblem(..),
  -- * Running and tracing a 'Colony'
  runColonyOnRatingSet,
  runColonyOnTestSet,
  traceColonyOn,
  traceColonyOnRatingSet,
  traceColonyOnTestSet,
  -- * Tools for implementing various instances
  singletonProblemInitialArenaState,
  singletonProblemRateEggs,
  singletonProblemRateEgg,
  -- * Reporting
  SingletonProblemReportSpec(..),
  SingletonProblemReport(..)
  ) where

import GHC.Generics
import Prelude hiding (id)

import Control.DeepSeq
import Control.Exception
-- import Control.Monad

import Data.Aeson
import Data.Aeson.TH
import Data.Data
import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
import Data.IntSet (IntSet)
import qualified Data.IntSet as IntSet
import qualified Data.List.NonEmpty as NonEmpty
-- import Data.Map (Map)
import qualified Data.Map as Map
import Data.Maybe
import Data.Semigroup
import Data.Set (Set)
-- import qualified Data.Set as Set
import Data.String
import Data.Typeable ()

-- import Text.Pandoc
import Text.Pandoc.Builder
import Text.Printf


import Data.BitString

import BaselMachine.Arena
import BaselMachine.Compiler
import BaselMachine.EfficaciousAnalysis
import BaselMachine.Instruction
import BaselMachine.Signal

import Digiostelium.AesonUtil
import Digiostelium.Agent
-- import Digiostelium.BasicReport
import Digiostelium.ColonyToStyledNetwork
import Digiostelium.Genome as G
import Digiostelium.GraphML
import Digiostelium.InputOutputProblem
import Digiostelium.PandocUtil
import Digiostelium.Problem
import Digiostelium.Query
-- import Digiostelium.Report
import Digiostelium.ReportException

import Digiostelium.StyledNetwork

import WGM.ParUtil
import WGM.SeqUtil
import WGM.ShortShow

-- | A 'SingletonProblem' is a problem that supplies a stream of
-- input to a single 'Colony', and the rating depends only on the
-- broadcast outputs and resources used during the computation.
-- It is assumed that the computation should contiune only until
-- the input runs out.

class (InputOutputProblem problem)
  =>  SingletonProblem problem where

  -- | Build the problem's initial colony state from the result
  -- of compiling the genome.
  initialColonyState ::
    problem
    -> (ProgramInformation, ReactionProgram)
    -> ColonyInitialState (Location problem) (Signal problem)

  -- | Give a (numerical) rating for yielding the output given
  -- this input.
  rateResult ::
    problem
    -> Input problem -- ^ a given input
    -> Output problem -- ^ the resulting output
    -> Int -- ^ total number of reactions performed
    -> Int -- ^ total of absolute value of all reaction deltas
    -> Rating problem -- ^ resulting rating


-- | Helper function for implementing instance 'ArenaProblem' 'initialArenaState'

singletonProblemInitialArenaState ::
  SingletonProblem problem =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> [ColonyInitialState (Location problem) (Signal problem)]

singletonProblemInitialArenaState problem compilerOutput =
    [initialColonyState problem compilerOutput]



-- | Helper function for implementing instance 'RateEggs' 'rateEggs'
singletonProblemRateEggs ::
  (
    Monad m,
    GenomeSpec gs,
    SingletonProblem problem,
    ArenaProblem problem,
    BroadcastSignal (Location problem) (Signal problem),
    Typeable (Signal problem),
    Show (Signal problem),
    Eq (Signal problem),
    Ord (Signal problem),
    Semigroup (Rating problem),
    NFData (Rating problem),
    NFData (Input problem),
    NFData (Output problem)
  )
  =>
  gs
  -> problem
  -> [Egg (Genome gs)]
  -> m [(Egg (Genome gs), Rating problem)]

singletonProblemRateEggs gs problem eggs =
  return [(egg, singletonProblemRateEgg gs problem egg)
         | egg <- eggs]


singletonProblemRateEgg ::
  (NFData (Output problem),
   NFData (Input problem),
   NFData (Rating problem),
   Semigroup (Rating problem),
   Eq (Signal problem),
   Ord (Signal problem),
   Show (Signal problem),
   Typeable (Signal problem),
   BroadcastSignal (Location problem) (Signal problem),
   SingletonProblem problem,
   ArenaProblem problem,
   GenomeSpec gs)
  =>
  gs
  -> problem
  -> Egg (Genome gs)
  -> Rating problem

singletonProblemRateEgg gs problem Egg{..} =
  snd (runColonyOnRatingSet problem progAndInfo)
  where
    instructions = interpretGenome gs genome
    progAndInfo =
      case compile (getInstructionSpec gs) instructions of
        Left errors ->
          throw (CompileErrors errors instructions)
        Right progAndInfo ->
          progAndInfo



-- | Trace a colony on a single input

traceColonyOn ::
  (SingletonProblem problem,
   ArenaProblem problem,
   BroadcastSignal (Location problem) (Signal problem),
   Typeable (Signal problem),
   Show (Signal problem),
   Eq (Signal problem),
   Ord (Signal problem))
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> Input problem
  -> (Input problem, Output problem, ArenaTraceSummary (Location problem) (Signal problem))

traceColonyOn problem progAndInfo input =
  (input, output, trSummary)
  where
    (maxTimeSteps, colonyInput) = problemInputToArenaInput problem input
    trSummary@ArenaTraceSummary{..} =
      (runArenaTrace
       [initialColonyState problem progAndInfo]
       (stopSignal problem)
       maxTimeSteps
       colonyInput)

    ArenaRunSummary{..} = trArenaRunSummary
    output = arenaOutputToProblemOutput problem arenaRunBroadcasts



-- | Trace a colony on all inputs used for rating it, including
-- all the trace mementos, which can take up a lot of memory.

traceColonyOnRatingSet ::
  (Eq (Signal problem),
   Ord (Signal problem),
   Show (Signal problem),
   Typeable (Signal problem),
   BroadcastSignal (Location problem) (Signal problem),
   SingletonProblem problem,
   ArenaProblem problem,
   NFData (Signal problem),
   NFData (Location problem),
   NFData (Input problem),
   NFData (Output problem))
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> [(Input problem, Output problem, ArenaTraceSummary (Location problem) (Signal problem))]

traceColonyOnRatingSet problem progAndInfo =
  parListDeep
  $ map (traceColonyOn problem progAndInfo)
  $ inputsForRating problem



-- | Trace a colony on all inputs used for rating it, including
-- all the trace mementos, which can take up a lot of memory.

traceColonyOnTestSet ::
  (Eq (Signal problem),
   Ord (Signal problem),
   Show (Signal problem),
   Typeable (Signal problem),
   BroadcastSignal (Location problem) (Signal problem),
   SingletonProblem problem,
   ArenaProblem problem,
   NFData (Signal problem),
   NFData (Location problem),
   NFData (Input problem),
   NFData (Output problem))
  =>
  problem
  -> String -- ^ test set name
  -> (ProgramInformation, ReactionProgram)
  -> [(Input problem, Output problem, ArenaTraceSummary (Location problem) (Signal problem))]

traceColonyOnTestSet problem testSetKey progAndInfo =
  parListDeep
  $ map (traceColonyOn problem progAndInfo)
  $ getTestSet problem testSetKey


-- | Run a colony on all inputs used for rating it
runColonyOn ::
  (SingletonProblem problem,
   ArenaProblem problem,
   BroadcastSignal (Location problem) (Signal problem),
   Typeable (Signal problem),
   Show (Signal problem),
   Eq (Signal problem),
   Ord (Signal problem))
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> Input problem
  -> (Output problem, Int, Int)
  -- ^ problem output, total reaction count, total reaction deltas

runColonyOn problem progAndInfo input =
  (problemOutput, totalReactionCount, totalReactionDeltas)
  where
    (maxTimeSteps, colonyInput) = problemInputToArenaInput problem input
    ArenaRunSummary{..} =
      (runArena
       [initialColonyState problem progAndInfo]
       (stopSignal problem)
       maxTimeSteps
       colonyInput)
    problemOutput =
      arenaOutputToProblemOutput problem arenaRunBroadcasts
    ColonyResourceSummary{..} = arenaColonyIxToResourcesUsed IntMap.! 0
    ResourceSummary{..} = colonyResourcesUsed


-- | Run a colony on the set of inputs used for rating.

runColonyOnRatingSet ::
  (Eq (Signal problem),
   Ord (Signal problem),
   Show (Signal problem),
   Typeable (Signal problem),
   BroadcastSignal (Location problem) (Signal problem),
   SingletonProblem problem,
   ArenaProblem problem,
   Semigroup (Rating problem),
   NFData (Rating problem),
   NFData (Input problem),
   NFData (Output problem))
  =>
  problem
  -> (ProgramInformation, ReactionProgram)
  -> ([(Input problem, Output problem, Int, Int, Rating problem)],
     Rating problem)
  -- ^
  -- list of problem input, generated output, total reaction
  -- count, total reaction deltas, and rating on that input;
  -- and total rating

runColonyOnRatingSet problem progAndInfo =
  (runs, sumRunRatings runs)
  where
    runs = parListDeep $
      [(input, output, totalReactionCount, totalReactionDeltas,
        rateResult problem input output totalReactionCount totalReactionDeltas)
      | input <- inputsForRating problem,
        let (output, totalReactionCount, totalReactionDeltas) =
              runColonyOn problem progAndInfo input]



-- | Run a colony on a test set of inputs.

runColonyOnTestSet ::
  (Eq (Signal problem),
   Ord (Signal problem),
   Show (Signal problem),
   Typeable (Signal problem),
   BroadcastSignal (Location problem) (Signal problem),
   SingletonProblem problem,
   ArenaProblem problem,
   Semigroup (Rating problem),
   NFData (Rating problem),
   NFData (Input problem),
   NFData (Output problem))
  =>
  problem
  -> String -- ^ test set key
  -> (ProgramInformation, ReactionProgram)
  -> ([(Input problem, Output problem, Int, Int, Rating problem)],
     Rating problem)
  -- ^
  -- list of problem input, generated output, total reaction
  -- count, total reaction deltas, and rating on that input;
  -- and total rating

runColonyOnTestSet problem testSetKey progAndInfo =
  (runs, sumRunRatings runs)
  where
    runs = parListDeep $
      [(input, output, totalReactionCount, totalReactionDeltas,
        rateResult problem input output totalReactionCount totalReactionDeltas)
      | input <- getTestSet problem testSetKey,
        let (output, totalReactionCount, totalReactionDeltas) =
              runColonyOn problem progAndInfo input]


-- | (Private) Utility function to sum ratings from 'runColonyOnRatingSet' and 'runColonyOnTestSet'
sumRunRatings :: Semigroup r => [(t1, t2, t3, t4, r)] -> r
sumRunRatings runs =
  sconcat
  $ NonEmpty.fromList
  $ [r | (_, _, _, _, r) <- runs]




-- === Reporting === --

data SingletonProblemReportSpec =
  FullDetails
  | Efficacious
  | RatingInputs
  | TestSet String
  | ReactionNetwork
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''SingletonProblemReportSpec)

-- | Various collections of reportable data that can be extracted
-- from an 'Agent' under a 'SingletonProblem'.  The various types
-- are passed in separately rather than using a 'GenomeSpec' and
-- 'Problem' because of limitations that would result when using
-- the generic and template derivations from packages @aeson@.

data SingletonProblemReport rating input output =
  FullDetailsReport {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    allInputs :: [(input, output, Int, Int, rating)],
    -- ^ table of all inputs used for rating, and the resulting output
    programInfo :: ProgramInformation,
    -- ^ instructions, reactions, and patterns referenced
    geneDescriptions ::  [String],
    -- ^ same as in 'GenomeDetails'
    instructionDetails :: [(Instruction, BitString, BitString, EncodingDetails)],
    -- ^ same as in 'GenomeDetails'
    efficaciousInstructions :: IntMap [(Int, Instruction)],
    -- ^ map from cell index to indexed efficacious instructions
    allEfficaciousInstructions :: [(Int, Instruction)],
    -- ^ all efficacious instructions
    efficaciousReactions :: IntMap (Set Reaction),
    -- ^ map from cell index to efficacious reactions
    allEfficaciousReactions :: Set Reaction,
    -- ^ all efficacious reactions
    efficaciousGeneDescriptions :: IntMap [String],
    -- ^ map from cell index to gene details of efficacious
    -- instructions
    allEfficaciousGeneDescriptions :: [String]
    -- ^ gene details as in 'GenomeDetails' but only for
    -- efficacious instructions
    }
    -- ^ summary of the agent, plus result of all inputs,
    -- analysis of efficacious genes, complete listing of the
    -- whole genome, stuff that can be computed using all inputs;
    -- requires significant computation.
  | EfficaciousReport {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    efficaciousInstructions :: IntMap [(Int, Instruction)],
    -- ^ map from cell index to indexed efficacious instructions
    allEfficaciousInstructions :: [(Int, Instruction)],
    -- ^ all efficacious instructions
    efficaciousReactions :: IntMap (Set Reaction),
    -- ^ map from cell index to efficacious reactions
    allEfficaciousReactions :: Set Reaction,
    -- ^ all efficacious reactions
    allEfficaciousGeneDescriptions :: [String]
    -- ^ gene details as in 'GenomeDetails' but only for
    -- efficacious instructions
    }
    -- ^ summary of the agent, plus analysis of efficacious
    -- genes, reduced from 'FullDetails'
  | InputsReport {
    id :: Integer,
    rating :: rating,
    allInputs :: [(input, output, Int, Int, rating)]
    }
    -- ^ table of inputs, and the resulting output.
  | ReactionNetworkReport {
      id :: Integer,
      reactionNetwork :: StyledNetwork
      }

$(deriveJSON myJSONOpts ''SingletonProblemReport)

instance
  forall genomeSpec problem.
  (GenomeSpec genomeSpec,
   ShortShow (Interval genomeSpec),
   ArenaProblem problem,
   SingletonProblem problem,
   Ord (Signal problem),
   Show (Signal problem),
   Typeable (Signal problem),
   NFData (Signal problem),
   NFData (Location problem),
   BroadcastSignal (Location problem) (Signal problem),
   Semigroup (Rating problem),
   NFData (Rating problem),
   NFData (Input problem),
   NFData (Output problem),
   SignalToNode (Signal problem)
  )
  =>
  QueryIndividual genomeSpec problem SingletonProblemReportSpec where

  type QueryResult genomeSpec problem SingletonProblemReportSpec =
    SingletonProblemReport (Rating problem) (Input problem) (Output problem)

  queryIndividual genomeSpec problem FullDetails Agent{egg=Egg{..}, ..} =
    FullDetailsReport {
      id = identifier,
      genomeLength = G.genomeLength genomeSpec genome,
      geneDescriptions = map shortShow geneDetails,
      instructionDetails =
          [ (instruction, codonBits, nucleotideBits, instructionEncodingDetails)
          | G.GeneDetails{..} <- geneDetails],
      ..}
    where
      instructions = interpretGenome genomeSpec genome

      progAndInfo@(programInfo, _) =
        case compile (getInstructionSpec genomeSpec) instructions
        of
          Right progAndInfo ->
            progAndInfo
          Left errors ->
            throw (CompileErrors errors instructions)

      instructionMap = IntMap.fromList (zip [0..] instructions)

      geneDetails = G.geneDetails genomeSpec genome
      geneDetailsMap = IntMap.fromList (zip [0..] geneDetails)

      traceResultsPlus = traceColonyOnRatingSet problem progAndInfo
      traceResults = [
        trColonyIxToSummary IntMap.! 0
        | (_input, _output, ArenaTraceSummary{..}) <- traceResultsPlus]

      allInputs = do
        (input, output, ArenaTraceSummary{..}) <- traceResultsPlus
        let ArenaRunSummary{..} = trArenaRunSummary
            colony0Resources = arenaColonyIxToResourcesUsed IntMap.! 0
            ColonyResourceSummary{..} = colony0Resources
            ResourceSummary{..} = colonyResourcesUsed
        return (
          input,
          output,
          totalReactionCount,
          totalReactionDeltas,
          rateResult problem input output totalReactionCount totalReactionDeltas)

      EfficaciousAnalysis{perCellEfficacious, allEfficaciousInstructionIxs, allEfficaciousReactions} = efficaciousAnalysis programInfo traceResults

      listInstructions :: IntSet -> [(Int, Instruction)]
      listInstructions ixSet =
        map (\j -> (j, instructionMap IntMap.! j)) (IntSet.toList ixSet)

      listGeneDescriptions :: IntSet -> [String]
      listGeneDescriptions ixSet =
        (map
         (\j -> shortShow $ geneDetailsMap IntMap.! j)
         (IntSet.toList ixSet))

      efficaciousReactions =
        IntMap.map cellEfficaciousReactionSet perCellEfficacious

      -- pullInstructions CellEfficaciousAnalysis{..} =
      --   listInstructions cellEfficaciousInstructionIxs

      efficaciousInstructions =
        IntMap.map (listInstructions . cellEfficaciousInstructionIxs) perCellEfficacious
      efficaciousGeneDescriptions =
        IntMap.map (listGeneDescriptions . cellEfficaciousInstructionIxs) perCellEfficacious
      allEfficaciousInstructions =
        listInstructions allEfficaciousInstructionIxs
      allEfficaciousGeneDescriptions =
        listGeneDescriptions allEfficaciousInstructionIxs


  -- Not the most efficient way to do this but this is easy:
  queryIndividual genomeSpec problem Efficacious a =
    EfficaciousReport {..}
    where
      FullDetailsReport{..} = queryIndividual genomeSpec problem FullDetails a


  queryIndividual genomeSpec problem RatingInputs Agent{egg=Egg{..}, ..} =
    InputsReport {
      id = identifier,
      ..}
    where
      instructions = interpretGenome genomeSpec genome

      progAndInfo =
        case compile (getInstructionSpec genomeSpec) instructions
        of
          Right progAndInfo ->
            progAndInfo
          Left errors ->
            throw (CompileErrors errors instructions)

      (allInputs, rating) = runColonyOnRatingSet problem progAndInfo


  queryIndividual genomeSpec problem (TestSet testSetKey) Agent{egg=Egg{..}, ..} =
    InputsReport {
      id = identifier,
      ..}
    where
      instructions = interpretGenome genomeSpec genome

      progAndInfo =
        case compile (getInstructionSpec genomeSpec) instructions
        of
          Right progAndInfo ->
            progAndInfo
          Left errors ->
            throw (CompileErrors errors instructions)

      (allInputs, rating) = runColonyOnTestSet problem testSetKey progAndInfo


  queryIndividual genomeSpec problem ReactionNetwork Agent{egg=Egg{..}, ..} =
    ReactionNetworkReport {
      id = identifier,
      reactionNetwork = (efficaciousAnalysisToStyledNetwork
                         effAn
                         (printf "Agent %d" identifier)
                         finalColonySpec)
      }
    where
      instructions = interpretGenome genomeSpec genome

      progAndInfo@(programInfo, _) =
        case compile (getInstructionSpec genomeSpec) instructions
        of
          Right progAndInfo ->
            progAndInfo
          Left errors ->
            throw (CompileErrors errors instructions)

      traceResultsPlus :: [(Input problem, Output problem, ArenaTraceSummary (Location problem) (Signal problem))]
      traceResultsPlus = traceColonyOnRatingSet problem progAndInfo
      -- Since this is a singleton problem with a static cellular
      -- structure, there should be one colony in the arena, with
      -- fixed cells and synapses.  That means we pull out the
      -- ColonyTraceSummary like so:
      traceResults = [ trColonyIxToSummary IntMap.! 0
                     | (_input, _output, ArenaTraceSummary{..}) <- traceResultsPlus ]

      effAn = efficaciousAnalysis programInfo traceResults

      finalColonySpec :: ColonySpec (Signal problem)
      finalColonySpec =
        colonySpec
        $ fromJust -- had better be there
        $ IntMap.lookup 0 -- only one ColonyState in map
        $ arenaColonyIxToState -- map to ColonyState
        $ fromJust -- had better be there
        $ Map.lookup AfterEmitters -- ArenaState at the end of the time step
        $ seqLast -- just the last time step
        $ trArenaSteps
        $ (\(_input, _output, arenaTraceSummary) -> arenaTraceSummary) -- fish out ArenaTraceSummary
        $ head -- just case of first input
        $ traceResultsPlus -- list of ArenaTraceSummary, one entry for each input





-- === Reports === --

instance GetTitle opt (SingletonProblemReport rating input output) where

  getTitle _ FullDetailsReport{..} =
    text "Full Details"

  getTitle _ EfficaciousReport{..} =
    text "Efficacious gene analysis"

  getTitle _ InputsReport{..} =
    text "Ratings on inputs"

  getTitle _ ReactionNetworkReport{..} =
    text "Reaction network"


instance
  (InputOutputProblem problem,
   Rating problem ~ rating,
   Input problem ~ input,
   Output problem ~ output,
   Show rating,
   ToJSON rating,
   Sniplet problem input,
   ToJSON input,
   Sniplet problem output,
   ToJSON output
  )
  =>
  Snip (genomeSpec, problem) (SingletonProblemReport rating input output) where

  toBlocks _ EfficaciousReport{..} =
    (header 1 $ text $ fromString $ printf "Agent %d" id)
    <>
    (para
     ("Rating:" <> space <> texts rating <> linebreak <>
      "Lifespan:" <> space <> texts lifespan <> linebreak <>
      "Parents:" <> space <> texts parentIds <> linebreak <>
      "Genome length:" <> space <> texts genomeLength <> linebreak))
    <>
    cellReports
    where
      cellReports =
        (mconcat
         (map cellReport (IntMap.keys efficaciousReactions)))

      cellReport cellIx =
        (header 2 $ text $ fromString $ printf "Cell %d" cellIx)
        <>
        (header 3 "Efficacious reactions")
        <>
        reactionBlock (efficaciousReactions IntMap.! cellIx)
        <>
        (header 3 "Efficacious instructions")
        <>
        instructionBlock (efficaciousInstructions IntMap.! cellIx)


  toBlocks (_genomeSpec, problem) InputsReport{..} =
    (header 1 $ text $ fromString $ printf "Agent %d" id)
    <>
    (para ("Rating:" <> space <> texts rating))
    <>
    mconcat inputReports

    where
      inputReports =
        map inputReport (zip [(0::Int) ..] allInputs)

      inputReport (j, (input, output, totalReactionCount, totalReactionDeltas, rating)) =
        (para
         ("Item #" <> space <> texts j <> ":" <> linebreak <>
          "Input :" <> space <> toInlines problem input <> linebreak <>
          "Output:" <> space <> toInlines problem output <> linebreak <>
          "Rxn ct:" <> space <> texts totalReactionCount <> linebreak <>
          "Tot dl:" <> space <> texts totalReactionDeltas <> linebreak <>
          "Rating:" <> space <> texts rating <> linebreak))

  -- If all else fails, show as YAML
  toBlocks _ r =
    toYAMLCodeBlock r


instance
  (SingletonProblem problem,
   Rating problem ~ rating,
   Input problem ~ input,
   Output problem ~ output,
   Show rating,
   ToJSON rating,
   Sniplet problem input,
   ToJSON input,
   Sniplet problem output,
   ToJSON output) =>
  ToPandoc (genomeSpec, problem) (SingletonProblemReport rating input output)


instance
  (SingletonProblem problem,
   Rating problem ~ rating,
   Input problem ~ input,
   Output problem ~ output,
   Show rating,
   ToJSON rating,
   Sniplet problem input,
   ToJSON input,
   Sniplet problem output,
   ToJSON output) =>
  ToGraphML (genomeSpec, problem) (SingletonProblemReport rating input output) where

  toGraphML (_genomeSpec, _problem) ReactionNetworkReport{..} =
    networkToGraphMLDocument reactionNetwork


  toGraphML _ _ =
    throw (UnsupportedFormat "GraphML output is not supported for this kind of SingletonProblemReport")
