module System.Posix.FileMutex
  (
    acquireLock,
    lockAcquired,
    releaseLock,
    readLockHolder,
    withLock
  ) where

import Control.Exception
import Control.Monad
import Network.HostName
import System.FilePath
import System.IO.Error
import System.Posix.Files
import System.Posix.Process
import Text.Printf

-- | (Private) Given a path to be used as a mutex, return the path to this process's target file.
lockTargetFile :: FilePath -> IO FilePath
lockTargetFile mutexPath = do
  myHostname <- getHostName
  myPID <- getProcessID
  let fileName = printf "Process-%s-%s.txt" myHostname (show myPID)
  return (takeDirectory mutexPath </> fileName)

-- | (Private) Given a path to be used as a mutex, create this process's target file and return a path to it.
makeLockTargetFile :: FilePath -> IO FilePath
makeLockTargetFile mutexPath = do
  myTargetFile <- lockTargetFile mutexPath
  myHostname <- getHostName
  myPID <- getProcessID
  let fileContent = printf "%s\n%s\n" myHostname (show myPID)
  writeFile myTargetFile fileContent
  return myTargetFile

-- | (Private) Given a path to be used as a mutex, remove this process's target file.
removeLockTargetFile :: FilePath -> IO ()
removeLockTargetFile mutexPath = do
  v <- lockAcquired mutexPath
  if v
    then do
    fail (printf "removeLockTargetFile: Lock '%s' is acquired by this process; cannot remove target file." mutexPath)
    else do
    myTargetFile <- lockTargetFile mutexPath
    removeLink myTargetFile

-- | Given a path to be used as a mutex, attempt to acquire the lock.
-- Return whether the lock was acquired.
--
-- This works by attempting to create a hard link from this process's target file to the mutex path.
-- If a file already exists at the mutex path, the attempt will fail, indicating that the mutex is held by another process.
-- If the attempt succeeds, other processes will be able to read the file at the mutex path and see which process holds the lock.
acquireLock :: FilePath -> IO Bool
acquireLock mutexPath = do
  myTargetFile <- makeLockTargetFile mutexPath
  result <- tryIOError (createLink myTargetFile mutexPath)
  case result of
    Right () -> do
      return True
    Left _ -> do
      removeLockTargetFile mutexPath
      return False

-- | Given a path to be used as a mutex, return whether this process holds the lock.
lockAcquired :: FilePath -> IO Bool
lockAcquired mutexPath = do
  result <- tryIOError $ do
    myTargetFile <- lockTargetFile mutexPath
    tStatus <- getFileStatus myTargetFile
    lStatus <- getFileStatus mutexPath
    return (deviceID tStatus == deviceID lStatus
            && fileID tStatus == fileID lStatus)
  case result of
    Right value ->
      return value
    Left _ ->
      return False

-- | Given a path to be used as a mutex, check to see whether this process holds the lock, and if so, release it.
--
-- This works by unlinking the mutex path, which will allow other processes to acquire the lock.

releaseLock :: FilePath -> IO ()
releaseLock mutexPath = do
  v <- lockAcquired mutexPath
  if v
    then removeLink mutexPath
    else return ()


{-
-- | (Private) Given a path to be used as a mutex, check to see whether any process holds the lock.
lockAvailable :: FilePath -> IO Bool
lockAvailable mutexPath = do
  status <- getFileStatus mutexPath
  return (not (isRegularFile status))
-}

-- | (Private) Same as catchIOError but the arguments are reversed, so it's a little nicer.
handleIOError :: (IOError -> IO a) -> IO a -> IO a
handleIOError handler axn =
  catchIOError axn handler


-- | Given a path to be used as a mutex, read the mutex file,
-- and return @Just (hostname, process id)@ for the process that holds the lock.
-- If no process holds the lock or an 'IOError' occurs, return @Nothing@.

readLockHolder :: FilePath -> IO (Maybe (String, Int))
readLockHolder mutexPath = do
  handleIOError (\_ -> return Nothing) $ do
    lockInfo <- readFile mutexPath
    case lines lockInfo of
      (hostname : pidStr : _) -> do
        return (Just (hostname, read pidStr))
      _ -> do
        return Nothing


-- | Given a path to be used as a mutex, attempt to acquire the lock.
-- If the attempt succeeds, perform the given 'IO' action and return 'Just' the result.
-- If the lock cannot be acquired, do not perform the action and return 'Nothing'.

withLock :: FilePath -> IO a -> IO (Maybe a)
withLock mutexPath axn =
  (bracket
   (acquireLock mutexPath)
   (\v ->
     when v $ do
       releaseLock mutexPath
       removeLockTargetFile mutexPath)
   (\v ->
     if v then do
       result <- axn
       return (Just result)
     else do
       return Nothing))
