{-# LANGUAGE NoMonomorphismRestriction, MultiParamTypeClasses, RecordWildCards, NamedFieldPuns, DeriveDataTypeable, DeriveGeneric, TemplateHaskell, FlexibleContexts, FlexibleInstances, ParallelListComp #-}

module RandomColony where

import qualified Data.IntMap as IM
import Data.Maybe
import Text.Printf

import BaselMachine.Colony
import System.Random.RandomST


data ColonyDistribution synDeltaD synDelayD =
  ColonyDistribution {
    inputTargetPattern :: Int,
    inputTargetDelta :: Int,
    synapseSourcePattern :: Int,
    synapseSourceThreshold :: Int,
    synapseActivationDelta :: Int,
    synapseTargetPattern :: Int,
    numLayers :: Int,
    numCellsPerLayer :: Int,
    layerCellIndexOffset :: Int, -- ^ cells in layer n are numbered starting from n * layerCellIndexOffset
    numDigits :: Int, -- ^ digits used in cell names
    synapseTargetDeltaDist :: synDeltaD,
    synapseDelayDist :: synDelayD
    }

makeEmitterSpec ColonyDistribution{..} =
  EmitterSpec{
    sourcePattern = synapseSourcePattern,
    sourceThreshold = synapseSourceThreshold,
    activationDelta = synapseActivationDelta
    }

makeInputSpec ColonyDistribution{..} nDigits j =
  (printf "In%0*d" nDigits j, j, r)
  where
    r = ReceptorSpec{
      targetPattern = inputTargetPattern,
      targetDelta = inputTargetDelta
      }

-- Monitor only!
makeOutputSpec ColonyDistribution{..} =
  EmitterSpec{
    sourcePattern = synapseSourcePattern,
    sourceThreshold = synapseSourceThreshold,
    activationDelta = 0
    }



-- Apparently there's a benefit to making reservoir networks
-- sparse, so I'm going to put in some efficiency: If the target
-- delta comes out 0, that means there's not really a synapse.
-- So this sampler generates Maybe SynapseSpec, so it's then easy
-- to filter out the non-existant synapses.

instance
  (Sample synDeltaD Int,
   Sample synDelayD Int) =>
  Sample (ColonyDistribution synDeltaD synDelayD)
  (Maybe SynapseSpec) where

  pullSample cd@ColonyDistribution{..} = do
    delay <- pullSample synapseDelayDist
    targetDelta <- pullSample synapseTargetDeltaDist
    let receptorSpec = ReceptorSpec{
          targetPattern = synapseTargetPattern,
          targetDelta }
    return
      (if targetDelta /= 0
       then Just SynapseSpec{
         emitterSpec = makeEmitterSpec cd,
         ..}
       else
         Nothing)



instance
  (Sample synDeltaD Int,
   Sample synDelayD Int) =>
  Sample (ColonyDistribution synDeltaD synDelayD) ColonySpec where

    pullSample cd@ColonyDistribution{..} =
      let
        cellNames =
          IM.fromList [
            (j, printf "Cell%0*d" numDigits j)
            | m <- [0 .. numCellsPerLayer - 1],
              layer <- [0 .. numLayers - 1],
              let j = layer * layerCellIndexOffset + m]

        inputSpecs =
          IM.fromList [
            (j, makeInputSpec cd numDigits j)
            | j <- [0 .. numCellsPerLayer - 1]   ]

        e = makeOutputSpec cd
        emitterSpecs =
          [(printf "Out%0*d" numDigits j, j, e, j)
          | m <- [0 .. numCellsPerLayer - 1],
              layer <- [0 .. numLayers - 1],
              let j = layer * layerCellIndexOffset + m]


      in do
        withinLayers <- concatMapM (randomWithinLayer cd) [0 .. numLayers - 1]
        betweenLayers <- concatMapM (randomBetweenLayers cd) [0 .. numLayers - 2]
        return ColonySpec {
          -- Leaves lots of disconnected cells, but that's okay...
          numCells = numLayers * layerCellIndexOffset + numCellsPerLayer,
          synapseSpecs = withinLayers ++ betweenLayers,
          ..}




randomWithinLayer cd@ColonyDistribution{..} layer = do
  let numEdgesInLayer = (numCellsPerLayer - 1) * (numCellsPerLayer - 1)
      numEdgesBetween = numCellsPerLayer * numCellsPerLayer
  mWithinLayerSyns <- pullSampleList numEdgesInLayer cd
  let withinLayerSyns =
        catMaybes $ [
          do
            s <- ms
            return (j, k, s)
          | m <- [0 .. numCellsPerLayer - 1],
            n <- [0 .. numCellsPerLayer - 1],
            m /= n,
            let j = layer * layerCellIndexOffset + m,
            let k = layer * layerCellIndexOffset + n
          | ms <- mWithinLayerSyns]
  return withinLayerSyns

randomBetweenLayers cd@ColonyDistribution{..} layer = do
  let numEdgesInLayer = (numCellsPerLayer - 1) * (numCellsPerLayer - 1)
      numEdgesBetween = numCellsPerLayer * numCellsPerLayer
  mForwardSyns <- pullSampleList numEdgesBetween cd
  mBackwardSyns <- pullSampleList numEdgesBetween cd
  let forwardSyns =
        catMaybes $ [
          do
            s <- ms
            return (j, k, s)
          | m <- [0 .. numCellsPerLayer - 1],
            n <- [0 .. numCellsPerLayer - 1],
            m /= n,
            let j = layer * layerCellIndexOffset + m,
            let k = (layer + 1) * layerCellIndexOffset + n
          | ms <- mForwardSyns]
      backwardSyns =
        catMaybes $ [
          do
            s <- ms
            return (j, k, s)
          | m <- [0 .. numCellsPerLayer - 1],
            n <- [0 .. numCellsPerLayer - 1],
            m /= n,
            let j = layer * layerCellIndexOffset + m,
            let k = (layer + 1) * layerCellIndexOffset + n
          | ms <- mBackwardSyns]
  return (forwardSyns ++ backwardSyns)


-- There's probably a better way, but...
concatMapM axn xs = do
  ys <- mapM axn xs
  return (concat ys)
