{-# LANGUAGE BangPatterns, RecordWildCards, NamedFieldPuns, FlexibleContexts,
MultiParamTypeClasses, UndecidableInstances, FlexibleInstances,
NoMonomorphismRestriction, ScopedTypeVariables,
TypeSynonymInstances, ForeignFunctionInterface,
DeriveDataTypeable, DeriveGeneric #-}

{-|

Module: Data.BitString

Description: A bit string type and associated functions

Copyright: © W. Garrett Mitchener 2017

License: GPL-3

Maintainer: garrett.mitchener@gmail.com

Stability: experimental

Portability: POSIX

The range and indexing convention here is more like C++ STL and
boost than Haskell arrays: For example, you pass in a @beginIx@
and @endIx@, and get back the bits with indices @beginIx@,
@beginIx@ + 1, ..., @endIx@ - 1, that is
@beginIx@ ≤ @j@ \< @endIx@.

-}

module Data.BitString (
  BitString,
  W,
  bsTo01String,
  bsFrom01String,
  bsFrom01StringChecked,
  bsEmpty,
  bsNull,
  bsSingleton,
  bsNumBits,
  bsSubstring,
  bsGetBit,
  bsGetBits,
  bsGetBitsAsWord,
  bsSetBit,
  bsToggleBit,
  bsToBits,
  bsToBytes,
  bsFromBytes,
  bsFromWord32s,
  bsToHexString,
  bsFromHexString,
  bsUnpack,
  bsSplitAt,
  bsTake,
  bsDrop,
  bsFirstRest,
  bsJoin,
  bsConcat,
  bsFromBits,
  bsIsPacked,
  bsRepack,
  bsChunks,
  bsChunksAsWords,
  bsChunkMap,
  bsHammingDistance,
  bsDelete,
  bsInsert,
  bsSplice,
  bsMajorityDecode,
--  BitStringRewrite(..),
--  bsApplyRewrites,
  ) where

import GHC.Exts
import GHC.Generics

import Control.Exception
import Control.Monad
import Control.DeepSeq

import Data.Binary
import Data.Bits
import Data.Char
import Data.Data
import Data.Aeson
import Data.List.Split
import Data.Typeable

-- import Debug.Trace

import Numeric
import Text.Printf
import Text.Regex.PCRE

import qualified Data.Vector.Storable as VS

import Foreign.C
import Foreign hiding (unsafePerformIO)
import System.IO.Unsafe (unsafePerformIO)



-- I originally wanted this to be flexible, but the ghc 64 bit
-- code generator is too slow, so I have to tune this for 32 bit
-- and run that instead.  When possible I'm going to abstract
-- away from that so that I could more easily switch to Word64,
-- but for database compatibility across platforms, W has to
-- be a specific size of word, not platform-dependent Word.

type W = Word32

--type WordArray = UArray Int W
type WordArray = VS.Vector W

instance  Binary WordArray where
  put v =
    put (VS.toList v)

  get = do
    ws <- get
    return (VS.fromList ws)

wBits :: Int -- how many bits are in a 'W', which had better be 32
wBits = finiteBitSize (0 :: W)

newtype WordBlock = WordBlock WordArray
                  deriving (Eq, Ord, Read, Show, Typeable, Data, Generic)

wbNil :: WordBlock
wbNil = WordBlock (VS.fromList [])


instance NFData WordBlock

wordToBits :: W -> [Bool]

wordToBits wd =
  map (testBit wd) [0 .. wBits - 1]


foreign import ccall unsafe "unpack_w32" c_unpack_w ::
  Word32 -> Ptr Word8 -> IO ()

foreign import ccall unsafe "unpack_w32_array" c_unpack_w_array ::
  Word -> Ptr Word32 -> Ptr Word8 -> IO ()

foreign import ccall unsafe "majority_decode" c_majority_decode ::
  Word -> Ptr Word32 -> Word -> Word -> Word -> Word -> Ptr Word32 -> IO ()

{-
foreign import ccall unsafe "get_bits_as_word" c_get_bits_as_word ::
  Ptr Word32 -> CUInt -> CUInt -> IO W
-}

{- Ended up not using this
wordToBitV :: W -> VS.Vector Word8

wordToBitV wd = unsafePerformIO $ do
  fp <- mallocForeignPtrBytes wBits
  withForeignPtr fp $ c_unpack_w wd
  return (VS.unsafeFromForeignPtr0 fp wBits)
-}

bitsToWord :: [Bool] -> W
bitsToWord bits =
  foldr (.|.) 0 [bit j
                | (j, True) <- zip [0 .. wBits - 1] bits]

wSetBit n v w =
  if v then setBit w n else clearBit w n

wbGetBit :: Int -> WordBlock -> Bool

wbGetBit n (WordBlock a) =
  assert (n >= 0) $
  let
    --(wdIx, btIx) = n `divMod` wBits
    wdIx = n `div` wBits
    btIx = n `mod` wBits
    w = a VS.! wdIx
  in
   testBit w btIx

wbGetBitsAsWord :: Int -> Int -> WordBlock -> W

-- This turns out to be faster than calling into C.
wbGetBitsAsWord = wbGetBitsAsWordRef

{-
wbGetBitsAsWordC nBegin nEnd wb@(WordBlock a) =
  unsafePerformIO $ VS.unsafeWith a $ \p ->
  c_get_bits_as_word p (toEnum nBegin) (toEnum nEnd)

wbGetBitsAsWordChecked nBegin nEnd wb@(WordBlock a) =
  let
    r = wbGetBitsAsWordRef nBegin nEnd wb
    v = unsafePerformIO $ VS.unsafeWith a $ \p ->
      c_get_bits_as_word p (toEnum nBegin) (toEnum nEnd)
  in
   if r == v
   then r
   else
     error (
       printf "wbGetBitsAsWord: mistake nBegin: %d nEnd: %d\nr: %08x v: %08x\n"
       nBegin nEnd r v
       ++
       concatMap (printf "%08x\n") (VS.toList a))
-}


wbGetBitsAsWordRef !nBegin !nEnd wb@(WordBlock a) =
  let
    !beginWdIx = nBegin `quot` wBits
    !beginBtIx = nBegin `rem` wBits
    !endWdIx  = nEnd `quot` wBits
    !endBtIx  = nEnd `rem` wBits
    !nBits = nEnd - nBegin
    !nWords = VS.length a
  in
   assert (nEnd > nBegin) $
   assert (0 <= beginWdIx) $
   assert (beginWdIx < nWords) $
   assert (0 <= endWdIx) $
   assert (endWdIx < nWords) $
   assert (nBits <= wBits) $
   if endWdIx == beginWdIx
   then -- all bits are well inside one word
     let
       !loWd = a VS.! beginWdIx
       -- clear upper bits and shift back down to bottom
       !bitsAsWord = shiftR (shiftL loWd (wBits - endBtIx)) (wBits - nBits)
     in
      assert (beginBtIx < endBtIx) $
      assert (wbGetBits nBegin nEnd wb == take nBits (wordToBits bitsAsWord)) $
      bitsAsWord
   else -- bits span two words
   -- if endWdIx == beginWdIx + 1 && endBtIx == 0
   if endBtIx == 0
   then -- goes all the way through the end of a word
     let
       !loWd = a VS.! beginWdIx
       -- shift back down to bottom
       !bitsAsWord = shiftR loWd (wBits - nBits)
     in
      assert (wbGetBits nBegin nEnd wb == take nBits (wordToBits bitsAsWord)) $
      bitsAsWord
   else
     assert (beginWdIx + 1 == endWdIx) $
     let
       !loWd = a VS.! beginWdIx
       !hiWd = a VS.! endWdIx
       !loBits = shiftR loWd beginBtIx
       !nLoBits = wBits - beginBtIx
       !nHiBits = nBits - nLoBits
       !hiBits = shiftR (shiftL hiWd (wBits - nHiBits)) (wBits - nHiBits - nLoBits)
       !bitsAsWord = hiBits .|. loBits
     in
      assert (wbGetBits nBegin nEnd wb == take nBits (wordToBits bitsAsWord)) $
      bitsAsWord


wbGetBits :: Int -> Int -> WordBlock -> [Bool]

wbGetBits nBegin nEnd (WordBlock a)
  | (nBegin == nEnd) =
    []
  | (nEnd >= nBegin) =
    let
      (beginWdIx, beginBtIx) = nBegin `quotRem` wBits
      (endWdIx, endBtIx) = nEnd `quotRem` wBits
      wFirst = a VS.! beginWdIx
      bitsFirst = drop beginBtIx (wordToBits wFirst)
    in
      if endWdIx == beginWdIx
      then take (endBtIx - beginBtIx) bitsFirst -- only one word involved
      else let
        wLast = a VS.! endWdIx
        bitsLast = take endBtIx (wordToBits wLast)
        in
         if endWdIx == beginWdIx + 1 -- no middle
         then bitsFirst ++ bitsLast
         else let -- at least one word in the middle
           wsMiddle = map (a VS.!) [beginWdIx + 1 .. endWdIx - 1]
           bitssMiddle = map wordToBits wsMiddle
           in
            concat (concat [[bitsFirst], bitssMiddle, [bitsLast]])
  | otherwise =
      error ("wbGetBits: inconsistent bit range " ++ show (nBegin, nEnd))

-- | Store each bit in words [begin,end) as a byte

wbToBitV :: Int -> Int -> WordBlock -> VS.Vector Word8

wbToBitV !begin !end (WordBlock a) =
  let
    nWords = end - begin
    nBits = wBits * nWords

  in
   assert (0 <= begin) $
   assert (begin < end) $
   unsafePerformIO $ do
     fp <- mallocForeignPtrBytes nBits
     withForeignPtr fp $ \rPtr ->
       VS.unsafeWith a $ \aPtr ->
       c_unpack_w_array (fromIntegral nWords) (plusPtr aPtr (4*begin)) rPtr
     return (VS.unsafeFromForeignPtr0 fp nBits)

-- | Group the input bits [nBegin, nEnd) into chunks of size
-- (2*k+1) and construct a new word block whose jth bit is 1 if
-- more than half of the jth input chunk are 1s and 0 otherwise.

wbMajorityDecode :: Int -- ^ k
                    -> Int -- ^ nBegin
                    -> Int -- ^ nEnd
                    -> WordBlock
                    -> WordBlock
wbMajorityDecode k nBegin nEnd (WordBlock a) =
  let
    beginWdIx :: Int
    beginBtIx :: Int
    !beginWdIx = nBegin `div` wBits
    !beginBtIx = nBegin `mod` wBits
    endWdIx :: Int
    endBtIx :: Int
    !endWdIx  = nEnd `div` wBits
    !endBtIx  = nEnd `mod` wBits
    nBits :: Int
    !nBits = nEnd - nBegin
    nWords :: Int
    !nWords = VS.length a
    nResultWords :: Int
    !nResultWords = 1 + nBits `quot` (wBits * (2*k + 1))
  in
   assert (k > 0) $
   assert (nBegin < nEnd) $
   assert (0 <= beginWdIx) $
   assert (beginWdIx < nWords) $
   assert (0 <= endWdIx) $
   assert (endWdIx < nWords) $
   assert (beginWdIx <= endWdIx) $
   unsafePerformIO $ do
     fp <- mallocForeignPtrArray nResultWords
     withForeignPtr fp $ \rPtr ->
       VS.unsafeWith a $ \aPtr ->
       (c_majority_decode
        (fromIntegral k)
        aPtr
        (fromIntegral beginWdIx)
        (fromIntegral beginBtIx)
        (fromIntegral endWdIx)
        (fromIntegral endBtIx)
        rPtr)
     return (WordBlock (VS.unsafeFromForeignPtr0 fp nResultWords))


data BitBlock = BitBlock {
  beginBitIx :: !Int,
  endBitIx :: !Int,
  wordBlock :: !WordBlock }
              deriving (Eq, Read, Show, Ord, Typeable, Data, Generic)

bbNil = BitBlock 0 0 wbNil

instance NFData BitBlock where

bbGetBit :: Int -> BitBlock -> Bool

bbGetBit n bb@(BitBlock {..}) =
  wbGetBit (beginBitIx + n) wordBlock

bbGetBits :: Int -> Int -> BitBlock -> [Bool]

bbGetBits nBegin nEnd bb@(BitBlock {..}) =
  wbGetBits (beginBitIx + nBegin) (beginBitIx + nEnd) wordBlock

bbGetBitsAsWord :: Int -> Int -> BitBlock -> W

bbGetBitsAsWord !nBegin !nEnd bb@(BitBlock {..}) =
  assert (0 <= nBegin) $
  assert (nBegin < nEnd) $
  assert (nEnd <= bbNumBits bb) $
  let
    bitsAsWord =
      wbGetBitsAsWord (beginBitIx + nBegin) (beginBitIx + nEnd) wordBlock
  in
   assert (bbGetBits nBegin nEnd bb == take (nEnd - nBegin) (wordToBits bitsAsWord))
     bitsAsWord

bbNumBits :: BitBlock -> Int

bbNumBits (BitBlock {..}) =
  endBitIx - beginBitIx

bbSplitAt :: Int -> BitBlock -> (BitBlock, BitBlock)

bbSplitAt n bb@(BitBlock {..})
  | n < 0 =
    (bbNil, bb)
  | n >= bbNumBits bb =
      (bb, bbNil)
  | otherwise =
        (BitBlock {endBitIx = beginBitIx + n, ..},
         BitBlock {beginBitIx = beginBitIx + n, ..})


bbToBits :: BitBlock -> [Bool]

bbToBits bb@(BitBlock {..}) =
   bbGetBits 0 (endBitIx - beginBitIx) bb

bbToBitV :: BitBlock -> VS.Vector Word8

bbToBitV bb@(BitBlock {..}) =
  let
    beginWdIx = beginBitIx `quot` wBits
    beginRemIx = beginBitIx `rem` wBits
    endRemIx = endBitIx `rem` wBits
    endWdIx = endBitIx `quot` wBits + (if endRemIx > 0 then 1 else 0)
    wdsAsBytes = wbToBitV beginWdIx endWdIx wordBlock
  in
   assert (0 <= beginBitIx) $
   assert (beginBitIx < endBitIx) $
   VS.unsafeSlice beginRemIx (endBitIx - beginBitIx) wdsAsBytes



bbTake :: Int -> BitBlock -> BitBlock

bbTake n bb =
  let
    (front, _) = bbSplitAt n bb
  in
   front

bbDrop :: Int -> BitBlock -> BitBlock

bbDrop n bb =
  let
    (_, back) = bbSplitAt n bb
  in
   back

-- | Group the input bits [nBegin, nEnd) into chunks of size
-- (2*k+1) and construct a new word block whose jth bit is 1 if
-- more than half of the jth input chunk are 1s and 0 otherwise.

bbMajorityDecode :: Int -- ^ k
                    -> Int -- ^ nBegin
                    -> Int -- ^ nEnd
                    -> BitBlock
                    -> BitBlock
bbMajorityDecode k nBegin nEnd bb@(BitBlock {..})
  | (nBegin == nEnd) = bbNil
  | otherwise =
    let
      adjBegin = beginBitIx + nBegin
      adjEnd = beginBitIx + nEnd
      decodedWordBlock = wbMajorityDecode k adjBegin adjEnd wordBlock
    in
     assert (0 < k) $
     assert (0 <= nBegin) $
     assert (nBegin < nEnd) $
     assert (nEnd <= bbNumBits bb) $
     BitBlock {
       beginBitIx = 0,
       endBitIx = (nEnd - nBegin) `quot` (2*k + 1),
       wordBlock = decodedWordBlock
       }


newtype BitString = BitString [BitBlock]
                  deriving (Typeable, Data)



-- | Representation of this 'BitString' as a string of characters
-- @0@ and @1@.  The bits are output in little-endian order, so
-- the first bit is output first, etc.

bsTo01String :: BitString -> String

bsTo01String bs =
  [if b then '1' else '0' | b <- bsToBits bs]

-- {- ## INLINABLE bsTo01String ## -}

-- | Reverse of 'bsTo01String'.  Given a string of binary digits
-- as characters @0@ and @1@, store the first binary digit as the
-- first bit of the string, etc.  That is, the input is assumed
-- to be in little-endian order.  Space characters are simply
-- skipped, but other characters result in a call to 'error'.

bsFrom01String :: String -> BitString

bsFrom01String cs =
  (bsFromBits
   (concat
    [case c of
        '0' -> [False]
        '1' -> [True]
        c ->
          if isSpace c
          then []
          else
            error ("bsFrom01String: string must only contain '0' and '1', found " ++ show c)
    | c <- cs]))

-- {- ## INLINABLE bsFrom01String ## -}



-- | Reverse of 'bsTo01String'.  Given a string of binary digits
-- as characters @0@ and @1@, store the first binary digit as the
-- first bit of the string, etc.  That is, the input is assumed
-- to be in little-endian order.  Space characters are simply
-- skipped, but other characters result in 'Left' with an error
-- message.  Success results in 'Right' with the 'BitString'.

bsFrom01StringChecked :: String -> Either String BitString

bsFrom01StringChecked cs =
  let
    nonBinaryRX :: Regex
    nonBinaryRX = makeRegex "[^01[:space:]]"
    ms :: [[String]]
    ms = match nonBinaryRX cs
  in
   if null ms
   then Right (bsFrom01String cs)
   else Left (
     "bsFrom01StringChecked: string must only contain '0' and '1', found "
     ++ show (concat ms))

-- {- ## INLINABLE bsFrom01StringChecked ## -}

instance Show BitString where
  showsPrec p bs =
    showsPrec p (bsTo01String bs)

instance Read BitString where
  readsPrec p inputString =
    let
      strParses :: [(String, String)]
      strParses = readsPrec p inputString
    in
     [(bsFrom01String cs, remainder)
      | (cs, remainder) <- strParses]

instance NFData BitString where
  rnf (BitString blocks) =
    rnf blocks

instance Eq BitString where
  -- This has to be done bit by bit because you might get
  -- identical bitstrings that are stored differently
  left == right =
    (bsToBits left == bsToBits right)

instance Ord BitString where
  compare x y =
    compare (bsToBits x) (bsToBits y)


bsEmpty :: BitString

bsEmpty = BitString []

bsSingleton :: Bool -> BitString

bsSingleton b = bsFromBits [b]

bsNull :: BitString -> Bool

bsNull (BitString []) = True

bsNull (BitString (bb:bbs)) =
  if bbNumBits bb == 0
  then bsNull (BitString bbs)
  else False


bsNumBits :: BitString -> Int

bsNumBits (BitString bs) =
  let
    bitCounts = map bbNumBits bs
  in
   sum bitCounts

bsGetBit :: Int -> BitString -> Bool

bsGetBit n s@(BitString (b:bs)) =
  assert (0 <= n) $
  assert (n < bsNumBits s) $
  let
    m = bbNumBits b
  in if n < m
     then bbGetBit n b
     else bsGetBit (n - m) (BitString bs)

-- {- ## INLINABLE bsGetBit ## -}

bsSetBit ::
  Int -> Bool -> BitString -> BitString

bsSetBit n v bs =
  let
    (front, back0) = bsSplitAt n bs
    back1 = bsDrop 1 back0
  in
   (bsJoin
    front
    (bsJoin
     (bsFromBits [v])
     back1))

-- {- ## INLINABLE bsSetBit ## -}


bsToggleBit ::
  Int -> BitString -> BitString

bsToggleBit n bs =
  assert (0 <= n) $
  assert (n < bsNumBits bs) $
  let
    v = bsGetBit n bs
  in
   bsSetBit n (not v) bs

-- {- ## INLINABLE bsToggleBit ## -}


bsSubstring :: Int -> Int -> BitString -> BitString

bsSubstring beginBtIx endBtIx bs =
  assert (0 <= beginBtIx) $
  assert (beginBtIx <= endBtIx) $
  assert (endBtIx <= bsNumBits bs) $
  bsTake (endBtIx - beginBtIx) (bsDrop beginBtIx bs)

-- {- ## INLINABLE bsSubstring ## -}


bsGetBits :: Int -> Int -> BitString -> [Bool]

bsGetBits beginBtIx endBtIx bs@(BitString (bb:bbs)) =
  assert (beginBtIx >= 0) $
  assert (beginBtIx <= endBtIx) $
  assert (endBtIx <= bsNumBits bs) $
  if beginBtIx == endBtIx
  then []
  else
    let
      lengthBB = bbNumBits bb
    in
     if beginBtIx >= lengthBB
     then bsGetBits (beginBtIx - lengthBB) (endBtIx - lengthBB) (BitString bbs)
     else
       if endBtIx <= lengthBB
       then bbGetBits beginBtIx endBtIx bb
       else
         let
           firstChunk = bbGetBits beginBtIx lengthBB bb
           rest = bsGetBits 0 (endBtIx - lengthBB) (BitString bbs)
         in
          firstChunk ++ rest

-- {- ## INLINABLE bsGetBits ## -}


bsGetBitsAsWord :: Int -> Int -> BitString -> W

bsGetBitsAsWord beginBtIx endBtIx bs@(BitString (bb:bbs)) =
  assert (beginBtIx >= 0) $
  assert (beginBtIx <= endBtIx) $
  assert (endBtIx <= bsNumBits bs) $
  if beginBtIx == endBtIx -- no bits requested
  then 0 -- not sure what else to do here...
  else
    let
      lengthBB = bbNumBits bb
      nBits = endBtIx - beginBtIx
    in
     if beginBtIx >= lengthBB -- skip past the first bit block
     then
       let
         bitsAsWord = bsGetBitsAsWord (beginBtIx - lengthBB) (endBtIx - lengthBB) (BitString bbs)
       in
        assert (bsGetBits beginBtIx endBtIx bs == take nBits (wordToBits bitsAsWord)) $
        bitsAsWord

   else
     if endBtIx <= lengthBB -- span begins and ends within first bit block
     then
       let
         bitsAsWord = bbGetBitsAsWord beginBtIx endBtIx bb
       in
        assert (bsGetBits beginBtIx endBtIx bs == take nBits (wordToBits bitsAsWord)) $
        bitsAsWord

     else -- span begins in this bit block but ends farther down
       let
         nLoBits = lengthBB - beginBtIx
         -- get some bits from this block
         loBits = bbGetBitsAsWord beginBtIx lengthBB bb
         -- get the rest recursively
         hiBits = bsGetBitsAsWord 0 (endBtIx - lengthBB) (BitString bbs)
         bitsAsWord = shiftL hiBits nLoBits .|. loBits
       in
        assert (bsGetBits beginBtIx endBtIx bs == take nBits (wordToBits bitsAsWord)) $
        bitsAsWord

bsGetBitsAsWord beginBtIx endBtIx bs =
  error (unwords
         ["bsGetBitsAsWord: *shrug* ",
          show beginBtIx,
          show endBtIx,
          show bs])


-- {- ## INLINABLE bsGetBitsAsWord ## -}


bsToBits :: BitString -> [Bool]

bsToBits (BitString bs) =
  concatMap bbToBits bs

-- {- ## INLINABLE bsToBits ## -}


-- | Representation of the string in big-endian hexadecimal
-- notation. That is, the string in little-endian (left-to-right)
-- order
--
-- @<0000 1000  0011 0101>@
--
-- is encoded to hex digits
--
-- @0@, @1@, @c@, and @a@
--
-- and the returned string is the big-endian (right-to-left) order
--
-- @ac10@
--
-- The point of doing this is that most software that reads
-- integers from text expects big-endian order.  So in C
-- notation, @0xac10@ represents a 16-bit integer whose bits from
-- the 1s place on up are @000010000110101@.

bsToHexString :: BitString -> String

bsToHexString bs =
  concatMap (printf "%02x") (reverse (bsToBytes bs))

-- {- ## INLINABLE bsToHexString ## -}


-- | Reverse of 'bsToHexString'. Given a string of hexadecimal
-- digits in big-endian (right-to-left) order, make the
-- corresponding bit string, storing the left-most bit of the
-- left-most hex digit as the first bit in the string, etc.

bsFromHexString :: String -> BitString

bsFromHexString chars =
  (bsFromBytes
   (reverse [b
            | chunk <- chunksOf 2 chars,
              let [(b :: Word8, _)] = readHex chunk]))

-- {- ## INLINABLE bsFromHexString ## -}


-- | This is probably hideously slow but it will work for now
bsFromBytes :: [Word8] -> BitString

bsFromBytes bs =
  bsConcat [
    BitString [BitBlock 0 8
               (WordBlock (VS.fromList [fromIntegral b]))
              | b <- bs]
    ]

-- {- ## INLINABLE bsFromBytes ## -}


bsToBytes :: BitString -> [Word8]

bsToBytes bs =
  -- have to add some extra zeros
  let
    n = bsNumBits bs
    (nBytes, nLoose) = n `quotRem` 8
    nExtras = if nLoose /= 0 then 8 - nLoose else 0
    zeros = bsFromBits (replicate nExtras False)
  in
   map fromIntegral (bsChunksAsWords 8 (bsJoin bs zeros))

-- {- ## INLINABLE bsToBytes ## -}


bsUnpack :: BitString -> VS.Vector Bool

bsUnpack bs =
  VS.fromList (bsToBits bs)

-- {- ## INLINABLE bsUnpack ## -}


bsSplitAt :: Int -> BitString -> (BitString, BitString)

bsSplitAt _ (BitString []) =
  (bsEmpty, bsEmpty)

bsSplitAt n bs
  -- Another potential case is:
  -- where n > bsNumBits bs  = (bs, bsEmpty) -- this is too slow to check for
  | n <= 0  = (bsEmpty, bs)
  | otherwise =
    let
      BitString (bb:bbs) = bs
      lengthBB = bbNumBits bb
    in
     if n > lengthBB
     then -- the entire first bit block goes on the left
       let
         (BitString front, BitString back) =
           bsSplitAt (n - lengthBB) (BitString bbs)
         splitFront = BitString (bb:front)
         splitBack = BitString back
       in
        assert (let m = bsNumBits splitFront in
                 m == n || (m < n && bsNull splitBack)) $
        assert (bsNumBits splitFront + bsNumBits splitBack ==
                bsNumBits bs) $
        (splitFront, splitBack)
     else
       if n == lengthBB
       then -- the split point is right at the end of the first block
         let
           splitFront = BitString [bb]
           splitBack = BitString bbs
         in
          assert (bsNumBits splitFront == n) $
          assert (bsNumBits splitFront + bsNumBits splitBack ==
                  bsNumBits bs) $
          (splitFront, splitBack)
       else -- the split point is inside the first bit block
         let
           (bbFront, bbBack) = bbSplitAt n bb
           splitFront = BitString [bbFront]
           splitBack = BitString (bbBack:bbs)
         in
          assert (bsNumBits splitFront == n) $
          assert (bsNumBits splitFront + bsNumBits splitBack ==
                  bsNumBits bs) $
          (splitFront, splitBack)

-- {- ## INLINABLE bsSplitAt ## -}


-- | Return the first @n@ bits from the front of the string.

bsTake :: Int -> BitString -> BitString

bsTake n bs =
  let
    (front, _) = bsSplitAt n bs
  in front

-- {- ## INLINABLE bsTake ## -}


-- | Drop the first @n@ bits from the front of the string

bsDrop :: Int -> BitString -> BitString

bsDrop n bs =
  let
    (_, back) = bsSplitAt n bs
  in back

-- {- ## INLINABLE bsDrop ## -}


-- | Return the first bit and the remainder of the bit string, or
-- 'Nothing' if the string is empty.

bsFirstRest :: BitString -> Maybe (Bool, BitString)

bsFirstRest bs =
  if bsNumBits bs > 0
  then let
    firstBit = bsGetBit 0 bs
    restBS = bsDrop 1 bs
    in
     Just (firstBit, restBS)
  else
    Nothing

-- {- ## INLINABLE bsFirstRest ## -}


bsJoin :: BitString -> BitString -> BitString

bsJoin (BitString bs1) (BitString bs2) =
  BitString (bs1 ++ bs2)

-- {- ## INLINABLE bsJoin ## -}


bsConcat :: [BitString] -> BitString

bsConcat bitStrings =
  BitString (concat [bs | BitString bs <- bitStrings])

-- {- ## INLINABLE bsConcat ## -}


bsFromBits :: [Bool] -> BitString

bsFromBits bits =
  let
    wordSizeChunks = chunksOf wBits bits
    wds = map bitsToWord wordSizeChunks
    nBits = length bits
    a = VS.fromList wds
  in
   BitString [BitBlock 0 nBits (WordBlock a)]

-- {- ## INLINABLE bsFromBits ## -}


-- | Return whether this bit string is in canonical form,
-- consisting of just a single bit block, possibly with an offset,
-- or an empty bit string.

bsIsPacked :: BitString -> Bool

bsIsPacked (BitString []) = True

bsIsPacked (BitString [bb]) = True

bsIsPacked (BitString (_:_:_)) = False

-- {- ## INLINABLE bsIsPacked ## -}


-- | Repack the whole bit string into a single contiguous bit
-- block

bsRepack :: BitString -> BitString

-- no need to repack if it's already in canonical form
bsRepack bs@(BitString [BitBlock 0 _ _]) =
  bs

bsRepack bs =
  let
    nBits = bsNumBits bs
    (nWds, nXBts) = nBits `quotRem` wBits
    words = [bsGetBitsAsWord (j*wBits) ((j+1)*wBits) bs
            | j <- [0 .. nWds - 1] ]
    lastWord =
      if nXBts == 0
      then 0
      else bsGetBitsAsWord (nWds*wBits) (nWds*wBits + nXBts) bs
    a = VS.fromList (words ++ [lastWord])
    newBS = BitString [BitBlock 0 nBits (WordBlock a)]
  in
   assert (newBS == bs) $
   newBS

-- {- ## INLINABLE bsRepack ## -}


bsToBitV :: BitString -> VS.Vector Word8

bsToBitV bs@(BitString [bb]) =
  let
    resultV = bbToBitV bb
  in
   -- assert (map fromEnum (VS.toList resultV) == map fromEnum (bsToBits bs)) $
   resultV

bsToBitV bs@(BitString bbs) =
  VS.concat (map bbToBitV bbs)

-- {- ## INLINABLE bsToBitV ## -}


bsChunks :: Int -> BitString -> [BitString]

bsChunks chunkSize bs =
  if bsNull bs
  then []
  else
    let
      (chunk, rest) = bsSplitAt chunkSize bs
    in
     chunk : bsChunks chunkSize rest

-- {- ## INLINABLE bsChunks ## -}


bsChunksAsWords :: Int -> BitString -> [W]

bsChunksAsWords chunkSize bs =
  assert (0 < chunkSize) $
  assert (chunkSize <= wBits) $
  loop 0
  where
    endIx = bsNumBits bs
    loop startIx =
      let
        chunkEndIx = startIx + chunkSize
      in
       if startIx >= endIx
       then []
       else
         (bsGetBitsAsWord startIx chunkEndIx bs) : loop chunkEndIx

-- {- ## INLINABLE bsChunksAsWords ## -}


bsToWords :: BitString -> [W]

bsToWords bs =
  loop 0
  where
    !endIx = bsNumBits bs
    loop !startIx =
      if startIx < endIx
      then
        let
          chunkEndIx = endIx `min` (startIx + wBits)
        in
         (bsGetBitsAsWord startIx chunkEndIx bs) : loop chunkEndIx
      else
        []

-- {- ## INLINABLE bsToWords ## -}


bsChunkMap :: Int -> (W -> Bool) -> BitString -> BitString
bsChunkMap chunkSize f bs =
  let
    nthChunk chunkIx =
      let
        begin = chunkIx * chunkSize
        end = begin + chunkSize
      in
       bsGetBitsAsWord begin end bs

    nChunks = bsNumBits bs `div` chunkSize
    r = bsNumBits bs `mod` chunkSize
    chunkResults = VS.generate nChunks (f . nthChunk)
  in
   assert (0 < chunkSize) $
   assert (chunkSize <= wBits) $
   assert (0 == r) $
   BitString [bbPackBitV chunkResults]

-- {- ## INLINABLE bsChunkMap ## -}


bbPackBitV :: VS.Vector Bool -> BitBlock
bbPackBitV explBitV =
  let
    nBits = VS.length explBitV
    nMainWords = nBits `div` wBits
    resBits = nBits `mod` wBits
    nWords = nMainWords + (if resBits > 0 then 1 else 0)
    wordAt j = wPackBitV (VS.slice (j*wBits) (min wBits (nBits-j*wBits)) explBitV)
    words = VS.generate nWords wordAt
    wb = WordBlock words
  in
   BitBlock 0 nBits wb

-- | Group the input bits [nBegin, nEnd) into chunks of size
-- (2*k+1) and construct a new word block whose jth bit is 1 if
-- more than half of the jth input chunk are 1s and 0 otherwise.
-- Note: I've only implemented this for packed bit strings so
-- far.
bsMajorityDecode :: Int -- ^ k
                    -> Int -- ^ nBegin
                    -> Int -- ^ nEnd
                    -> BitString
                    -> BitString
bsMajorityDecode _ _ _ (BitString []) = bsEmpty

bsMajorityDecode k nBegin nEnd (BitString [bb])
  | (nBegin == nEnd) = bsEmpty
  | otherwise =
    assert ((nEnd - nBegin) `mod` (2*k+1) == 0) $
    BitString [bbMajorityDecode k nBegin nEnd bb]

bsMajorityDecode k nBegin nEnd bs@(BitString bbs) =
  error (printf
         "bsMajorityDecode: this case is not implemented yet\nk: %d\nnBegin: %d\nnEnd: %d\nbs: (%d blocks) %s\n"
        k nBegin nEnd (length bbs) (show bbs))

-- {- ## INLINABLE bsMajorityDecode ## -}


-- bools at low indices in the vector end up at low indices in
-- the resulting word
wPackBitV :: VS.Vector Bool -> W
wPackBitV explBitV =
  let
    v :: W
    v = VS.foldr' (\bitBool a -> toEnum (fromEnum bitBool) + shiftL a 1) 0 explBitV
  in
  -- trace ("packed " ++ show explBitV ++ " into " ++ printf "%08x" v) $
  v


instance Binary BitString where
  put bs =
    let
      BitString [BitBlock 0 nBits (WordBlock a)] = bsRepack bs
    in do
      put nBits
      put a

  get = do
    nBits <- get
    a <- get
    return (BitString [BitBlock 0 nBits (WordBlock a)])


-- | Convert a list of 'Word32' directly into a 'BitString'.  The
-- supplied words are interpreted little-endian, so the lowest
-- order bit in the first word is the first bit in the resulting
-- 'BitString'.

bsFromWord32s :: [Word32] -> BitString
bsFromWord32s words =
  let
    arr = VS.fromList words

  in
   BitString [BitBlock 0 (32 * VS.length arr) (WordBlock arr)]

-- {- ## INLINABLE bsFromWord32s ## -}


-- | Return the Hamming distance of the two bit strings, that is,
-- the number of indices at which they are different.  If the
-- strings are of different lengths, the number of additional
-- bits set in the longer string is included.

bsHammingDistance ::
  BitString -> BitString -> Int

bsHammingDistance = bsHammingDistanceViaWDirect

-- {- ## INLINABLE bsHammingDistance ## -}


bsHammingDistanceViaWDirect x y =
  let
    !nBitsX = bsNumBits x
    !nBitsY = bsNumBits y
    !nBitsBoth = min nBitsX nBitsY

    loopBoth :: Int -> Int -> Int
    loopBoth !j !n =
      if j + wBits < nBitsBoth
      then
        (loopBoth
         (j+wBits)
         (n
          + popCount ((bsGetBitsAsWord j (j+wBits) x)
                      `xor`
                      (bsGetBitsAsWord j (j+wBits) y))))
      else
        -- might be some residue
        if j < nBitsBoth
        then
          (n
           + popCount ((bsGetBitsAsWord j nBitsBoth x)
                      `xor`
                      (bsGetBitsAsWord j nBitsBoth y)))
        else
          n

    loopRest !z !jEnd !j !n =
      if j + wBits < jEnd
      then
        (loopRest
         z
         jEnd
         (j+wBits)
         (n + popCount (bsGetBitsAsWord j (j+wBits) z)))
      else
        -- might be some residue
        if j < jEnd
        then
          (n
           + popCount (bsGetBitsAsWord j jEnd z))
        else
          n

    !bothSum = loopBoth 0 0
    !restSum =
      if nBitsX > nBitsY
      then
        loopRest x nBitsX nBitsBoth bothSum
      else
        if nBitsY > nBitsX
        then
          loopRest y nBitsY nBitsBoth bothSum
        else
          bothSum
  in
   restSum

-- {- ## INLINE bsHammingDistanceViaWDirect ## -}



-- Not quite as fast as the above, because it allocates lists
bsHammingDistanceViaW x y =
  let
    xw = bsToWords x
    yw = bsToWords y
    loop (x:xs) (y:ys) !n =
      loop xs ys (n + popCount (x `xor` y))
    loop [] ys !n =
      n + sum (map popCount ys)
    loop xs [] !n =
      n + sum (map popCount xs)
    !m = loop xw yw 0
  in
   m

-- I'm keeping this one just for reference
bsHammingDistanceViaBV x y =
  let
    xb = bsToBitV x
    yb = bsToBitV y
    nx = VS.length xb
    ny = VS.length yb
    d = abs (nx - ny)
    n = min nx ny
    ks :: VS.Vector Int
    ks = VS.zipWith (\ p q -> if p == q then 0 else 1) xb yb
  in
   VS.sum ks + d

bsDelete ::
  Int -> Int -> BitString
  -> BitString

bsDelete beginBtIx endBtIx bs =
  assert (0 <= beginBtIx) $
  assert (beginBtIx <= endBtIx) $
  assert (endBtIx <= bsNumBits bs) $
  let
    front = bsTake beginBtIx bs
    back = bsDrop endBtIx bs
  in
   bsJoin front back

-- {- ## INLINABLE bsDelete ## -}


bsInsert ::
  Int -> BitString -> BitString
  -> BitString

bsInsert beginBtIx editInsert bs =
  assert (0 <= beginBtIx) $
  assert (beginBtIx <= bsNumBits bs) $
  let
     (front, back) = bsSplitAt beginBtIx bs
  in
   bsJoin front (bsJoin editInsert back)

-- {- ## INLINABLE bsInsert ## -}


bsSplice ::
  Int -> Int -> BitString -> BitString
  -> BitString

bsSplice beginBtIx endBtIx editInsert bs =
  assert (0 <= beginBtIx) $
  assert (beginBtIx <= endBtIx) $
  assert (endBtIx <= bsNumBits bs) $
  let
    front = bsTake beginBtIx bs
    back = bsDrop endBtIx bs
  in
   bsJoin front (bsJoin editInsert back)

-- {- ## INLINABLE bsSplice ## -}

instance ToJSON BitString where
  toJSON bs = toJSON (bsNumBits bs, bsToHexString bs)

instance FromJSON BitString where
  parseJSON v = do
    (n, hexStr) <- parseJSON v
    return (bsTake n (bsFromHexString hexStr))


--- UNTESTED FROM HERE ---
{-

data SubstringSource =
  Literal !BitString
  | InOriginal !Int !Int
    deriving (Read, Show, Eq, Ord)

-- | This type represents stream oriented rewriting of a
-- bitstring instead of indexed rewriting: Traverse the string
-- from the beginning and apply transformations as you go.

data BitStringRewrite =
  ToggleNextBit
  | TakeBits !Int -- ^ how many bits to keep
  | DropBits !Int -- ^ how many bits to drop
  | PushBits !BitString -- ^ bits to insert here
  | DuplicateBits !Int -- ^ how many bits to duplicate
    deriving (Read, Show, Eq)


bsApplyRewrites :: [BitStringRewrite] -> BitString -> BitString

bsApplyRewrites [] bs = bs

bsApplyRewrites (ToggleNextBit : rewrites) bs =
  case bsFirstRest bs of
    Just (b, rest) ->
      bsJoin (bsSingleton (not b)) (bsApplyRewrites rewrites rest)
    Nothing  ->
      error "bsApplyRewrites ToggleNextBit : bitstring is null"


bsApplyRewrites (TakeBits n : rewrites) bs =
  let
    (front, back) = bsSplitAt n bs
  in
   bsJoin front (bsApplyRewrites rewrites back)

bsApplyRewrites (DropBits n : rewrites) bs =
  bsApplyRewrites rewrites (bsDrop n bs)

bsApplyRewrites (DuplicateBits n : rewrites) bs =
  let
    front = bsTake n bs
  in
   bsJoin front (bsApplyRewrites rewrites bs)

bsApplyRewrites (PushBits toInsert : rewrites) bs =
  bsJoin toInsert (bsApplyRewrites rewrites bs)

-}
