{-|

Module: WGM.ShowAsMathematica

Description: A class for converting objects into a string
representation that can be read and interpreted by Mathematica.

Copyright: © W. Garrett Mitchener 2017

License: GPL-3

Maintainer: garrett.mitchener@gmail.com

Stability: experimental

Portability: Portable

-}


module WGM.ShowAsMathematica where

import Numeric
import Data.List
import qualified Data.Map as Map
import qualified Data.IntMap as IntMap


-- | Represent an object in Mathematica syntax

class Show t => ShowAsMathematica t where
  -- | The default implementation is the standard 'show'
  showAsMathematica :: t -> String
  showAsMathematica = show

instance ShowAsMathematica Int
instance ShowAsMathematica Integer
instance ShowAsMathematica Bool

-- Notice: this doesn't work :-(
--instance ShowAsMathematica String

instance ShowAsMathematica Double where
    showAsMathematica x =
	let
	  (digits, power) = floatToDigits 10 x
	in
	  if abs power < 3 then
	    showFFloat Nothing x ""
	  else
	    "0." ++ concat (map show digits) ++ "*^" ++ show power

instance ShowAsMathematica t => ShowAsMathematica [t] where
    showAsMathematica xs =
	("{" ++ concat (intersperse ", " (map showAsMathematica xs)) ++ "}" )

instance (ShowAsMathematica k, ShowAsMathematica v) =>
    ShowAsMathematica (Map.Map k v) where
    showAsMathematica fm =
	("{" ++
	 concat (intersperse ", " (map helper (Map.assocs fm)))
	 ++ "}")
	where
	  helper (k, v) =
	      showAsMathematica k ++ " -> " ++ showAsMathematica v

instance ShowAsMathematica v =>
    ShowAsMathematica (IntMap.IntMap v) where
    showAsMathematica fm =
	("{" ++
	 concat (intersperse ", " (map helper (IntMap.assocs fm)))
	 ++ "}")
	where
	  helper (k, v) =
	      showAsMathematica k ++ " -> " ++ showAsMathematica v

instance (ShowAsMathematica t1, ShowAsMathematica t2) =>
    ShowAsMathematica (t1, t2) where
    showAsMathematica (x1, x2) =
        concat ["{",
                showAsMathematica x1,
		", ", showAsMathematica x2,
                "}"]

instance (ShowAsMathematica t1, ShowAsMathematica t2,
          ShowAsMathematica t3) =>
    ShowAsMathematica (t1, t2, t3) where
    showAsMathematica (x1, x2, x3) =
        concat ["{",
                showAsMathematica x1,
		", ", showAsMathematica x2,
		", ", showAsMathematica x3,
                "}"]

instance (ShowAsMathematica t1, ShowAsMathematica t2,
          ShowAsMathematica t3, ShowAsMathematica t4) =>
    ShowAsMathematica (t1, t2, t3, t4) where
    showAsMathematica (x1, x2, x3, x4) =
        concat ["{",
                showAsMathematica x1,
		", ", showAsMathematica x2,
		", ", showAsMathematica x3,
		", ", showAsMathematica x4,
                "}"]

instance (ShowAsMathematica t1, ShowAsMathematica t2,
          ShowAsMathematica t3, ShowAsMathematica t4,
          ShowAsMathematica t5) =>
    ShowAsMathematica (t1, t2, t3, t4, t5) where
    showAsMathematica (x1, x2, x3, x4, x5) =
        concat ["{",
                showAsMathematica x1,
		", ", showAsMathematica x2,
		", ", showAsMathematica x3,
		", ", showAsMathematica x4,
		", ", showAsMathematica x5,
                "}"]

-- | A string that should be shown as itself with no quotation
-- marks, that is, as a symbol, for output to Mathematica.
data Symbol = Symbol String
	    deriving (Eq, Ord, Read, Show)

instance ShowAsMathematica Symbol where
    showAsMathematica (Symbol s) = s
