{-# LANGUAGE FlexibleInstances #-}

{-|

Module: WGM.MapUtil

Description: Tools for working with sets and maps

Copyright: © 2017 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : portable


-}


module WGM.MapUtil where

import Data.Map (Map)
import qualified Data.Map as M
import Data.Set (Set)
import qualified Data.Set as S
import Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import Data.IntSet (IntSet)
import qualified Data.IntSet as IS
import qualified Data.Sequence as Seq
import Data.Foldable


-- | Something that can be converted to a 'Map'
class AsMap c where

  -- | Convert an object to a 'Map'
  asMap :: Ord k => c k -> Map k ()

-- | Interpret a list @[x1, x2, ...]@ as a map @{ x1: (), x2: (), ... }@
instance AsMap [] where
  asMap sList = M.fromList [(k, ()) | k <- sList]

-- | Interpret a set @{x1, x2, ...}@ as a map @{ x1: (), x2: (), ...}@
instance AsMap Set where
  asMap sSet = M.fromAscList [(k, ()) | k <- S.toAscList sSet ]

-- | Convert the second argument to a 'Map' and compute the
-- difference from the first map
mapDifference :: (AsMap c, Ord k) => Map k a -> c k -> Map k a
mapDifference m s =
   M.difference m (asMap s)

-- | Convert the second argument to a 'Map' and compute the
-- intersection with the first map
mapIntersection :: (AsMap c, Ord k) => Map k a -> c k -> Map k a
mapIntersection m s =
  M.intersection m (asMap s)


-- | Something that can be converted to an 'IntMap'
class AsIntMap c where

  -- | Convert an object to an 'IntMap'
  asIntMap :: c -> IntMap ()


-- | Interpret a list @[x1, x2, ...]@ as a map @{ x1: (), x2: (), ... }@
instance AsIntMap [Int] where
  asIntMap sList = IM.fromList [ (k, ()) | k <- sList]

-- | Interpret a set @{x1, x2, ...}@ as a map @{ x1: (), x2: (), ...}@
instance AsIntMap IntSet where
  asIntMap sSet = IM.fromAscList [ (k, ()) | k <- IS.toAscList sSet]

-- | Interpret a sequence @[|x1, x2, ...|]@ as a map @{ x1: (), x2: (), ...}@
instance AsIntMap (Seq.Seq Int) where
  asIntMap sSeq = IM.fromList [ (k, ()) | k <- toList sSeq ]


-- | Convert the second argument to an 'IntMap' and compute the
-- difference from the first map
intMapDifference :: AsIntMap c => IntMap a -> c -> IntMap a
intMapDifference m s =
   IM.difference m (asIntMap s)


-- | Convert the second argument to an 'IntMap' and compute the
-- intersection with the first map
intMapIntersection :: AsIntMap c => IntMap a -> c -> IntMap a
intMapIntersection m s =
   IM.intersection m (asIntMap s)


-- | Map a function over a 'Set' of keys and return a 'Map' of
-- keys to values
setMap ::
  Ord key =>
  (key -> value) -> Set key -> Map key value

setMap f keySet =
  M.fromDistinctAscList [(key, f key) | key <- S.toAscList keySet]


-- | Map a function over an 'IntSet' of keys and return a 'IntMap' of
-- keys to values
intSetMap ::
  (Int -> value) -> IntSet -> IntMap value

intSetMap f iSet =
  IM.fromDistinctAscList [(key, f key) | key <- IS.toAscList iSet]
