{-# LANGUAGE NoMonomorphismRestriction, ScopedTypeVariables #-}

{-|

Module: WGM.MonadUtil

Description: A few utilities for use with monads.

Copyright: © W. Garrett Mitchener 2017

License: GPL-3

Maintainer: garrett.mitchener@gmail.com

Stability: experimental

Portability: Portable

-}

module WGM.MonadUtil where

import Control.Monad


-- | If the condition is true, do the action
whenM :: (Monad m) => m Bool -> m () -> m ()
whenM cond axn = do
  b <- cond
  when b axn

{-# INLINABLE whenM #-}


-- | If the condition is true, do the "true" action, else do the
-- "false" action

ifM :: (Monad m) => m Bool -> m b -> m b -> m b
ifM cond tAxn fAxn = do
  b <- cond
  if b then tAxn else fAxn

{-# INLINABLE ifM #-}


-- | Short-circuit monadic "or": If the first action returns
-- 'True', stop, otherwise perform the second action.

orM :: (Monad m) => m Bool -> m Bool -> m Bool

orM cond fAxn = do
  b <- cond
  if b then return True else fAxn

{-# INLINABLE orM #-}


-- | Short-circuit monadic "and": If the first action returns
-- 'False', stop, otherwise perform the second action.

andM :: (Monad m) => m Bool -> m Bool -> m Bool

andM cond tAxn = do
  b <- cond
  if b then tAxn else return False

{-# INLINABLE andM #-}


-- | Fold 'orM' over a list of actions

orLM :: (Monad m) => [m Bool] -> m Bool
orLM = foldr1 orM

{-# INLINABLE orLM #-}


-- | Fold 'andM' over a list of actions
andLM :: (Monad m) => [m Bool] -> m Bool
andLM = foldr1 andM

{-# INLINABLE andLM #-}


-- | Form of 'concatMap' that works monadically
concatMapMM :: (Monad m, MonadPlus list) =>
              (m a -> m (list b)) -> [m a] -> m (list b)
concatMapMM f axns = do
  ys :: [list b] <- mapM f axns
  return (msum ys)

{-# INLINABLE concatMapMM #-}


-- | Repeat the action until it returns 'True'
untilTrue :: Monad m => m Bool -> m ()
untilTrue axn = do
  v <- axn
  unless v $ untilTrue axn
